/*
 *	spButton_Gtk.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spFrameP.h>
#include <sp/spMenuP.h>
#include <sp/spMenuItemP.h>
#include <sp/spButtonP.h>

int spGetMnemonicGtk(const char *label, char *text)
{
    int i, j;
    int len;
    int c = NUL;
#ifndef SP_SUPPORT_MNEMONIC
    int prev_c = NUL;
#endif

    len = strlen(label);
    spDebug(40, "spGetMnemonicGtk", "strlen = %d\n", len);
    
    for (i = 0, j = 0; i < len; i++) {
	if (label[i] == '&' && label[i + 1] == '&') {
	    i++;
	    text[j] = label[i];
	    j++;
	} else if (label[i] == '&') {
	    c = label[i + 1];
#ifndef SP_SUPPORT_MNEMONIC
	    if (prev_c == '(' && c != NUL && label[i + 2] == ')') {
		j--;
		i += 2;
	    }
#else
	    text[j] = '_';
	    j++;
#endif
	} else {
#ifdef SP_SUPPORT_MNEMONIC
	    if (label[i] == '_') {
		text[j] = '_';
		j++;
	    }
#endif
	    text[j] = label[i];
	    j++;
	}
#ifndef SP_SUPPORT_MNEMONIC
	prev_c = label[i];
#endif
    }
    text[j] = NUL;

#ifdef MNEMONIC_LOWER
    if (isupper(c)) {
	c = tolower(c);
    }
#endif
    
    spDebug(30, NULL, "label = %s, mnemonic = %c, text = %s\n", label, c, text);
    
    return c;
}

static void createRadioButton(spComponent component, char *label)
{
    spComponent prev_radio;
    GSList *group = NULL;

    spAddRadioGroup(component, SpButtonPart(component).radio_end);

    if ((prev_radio = spGetRadioGroupPrev(component)) != NULL) {
#if GTK_CHECK_VERSION(3,0,0)
	group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(SpPrimitiveArch(prev_radio).widget));
#else
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(SpPrimitiveArch(prev_radio).widget));
#endif
    }
    
#if GTK_CHECK_VERSION(2,0,0)
    SpPrimitiveArch(component).widget = gtk_radio_button_new_with_mnemonic(group, label);
#else
    SpPrimitiveArch(component).widget = gtk_radio_button_new_with_label(group, label);
#endif
    
#if GTK_CHECK_VERSION(3,0,0)
    gtk_radio_button_get_group(GTK_RADIO_BUTTON(SpPrimitiveArch(component).widget));
#else
    gtk_radio_button_group(GTK_RADIO_BUTTON(SpPrimitiveArch(component).widget));
#endif
    
    return;
}

void spButtonCreateArch(spComponent component)
{
    static char label[SP_MAX_BUTTON_LABEL] = "";

    strcpy(label, "");
    if (!strnone(SpComponentPart(component).title)) {
	spGetMnemonicGtk(SpComponentPart(component).title, label);
    } else {
	strcpy(label, (!strnone(SpGetName(component)) ? SpGetName(component) : ""));
    }
    
    if (spIsSubClass(component, SpCheckBox)) {
#if GTK_CHECK_VERSION(2,0,0)
	SpPrimitiveArch(component).widget = gtk_check_button_new_with_mnemonic(label);
#else
	SpPrimitiveArch(component).widget = gtk_check_button_new_with_label(label);
#endif
    } else if (spIsSubClass(component, SpRadioButton)) {
	createRadioButton(component, label);
    } else {
#if GTK_CHECK_VERSION(2,0,0)
	SpPrimitiveArch(component).widget = gtk_button_new_with_mnemonic(label);
#else
	SpPrimitiveArch(component).widget = gtk_button_new_with_label(label);
#endif
	if (SpButtonPart(component).default_flag == SP_TRUE) {
#if GTK_CHECK_VERSION(3,0,0)
	    gtk_widget_set_can_default(SpPrimitiveArch(component).widget, TRUE);
#else
	    GTK_WIDGET_SET_FLAGS(SpPrimitiveArch(component).widget, GTK_CAN_DEFAULT);
#endif
	    gtk_window_set_default(GTK_WINDOW(SpFrameArch(SpGetWindow(component)).toplevel),
				   SpPrimitiveArch(component).widget);
	}
    }

    spPrimitiveContainerAddGtk(component);
    
    return;
}

void spUpdateWidgetTitleGtk(GtkWidget *widget, const char *label)
{
    GList *children;

#if GTK_CHECK_VERSION(3,0,0)
    children = gtk_container_get_children(GTK_CONTAINER(widget));
#else
    children = gtk_container_children(GTK_CONTAINER(widget));
#endif

    while (children != NULL) {
	if (children->data != NULL && GTK_IS_LABEL(children->data) == TRUE) {
#if GTK_CHECK_VERSION(2,0,0)
	    gtk_label_set_text_with_mnemonic(GTK_LABEL(children->data), label);
#else
	    gtk_label_set(GTK_LABEL(children->data), label);
#endif
	    break;
	}

	children = children->next;
    }
    
    return;
}

void spButtonSetParamsArch(spComponent component)
{
    static char label[SP_MAX_BUTTON_LABEL] = "";

    if (SpComponentPart(SpOldObject(component)).title != SpComponentPart(component).title) {    
	strcpy(label, "");
	if (!strnone(SpComponentPart(component).title)) {
	    spGetMnemonicGtk(SpComponentPart(component).title, label);
	}
	
	if (!strnone(label)) {
	    spUpdateWidgetTitleGtk(SpPrimitiveArch(component).widget, label);
	}
    }

    return;
}

void spSetToggleStateArch(spComponent component)
{
    if (spIsSubClass(component, SpMenuItem) == SP_TRUE) {
#if GTK_CHECK_VERSION(3,0,0)
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(SpPrimitiveArch(component).widget),
				       (SpButtonPart(component).set == SP_ON ? TRUE : FALSE));
#else
	gtk_check_menu_item_set_state(GTK_CHECK_MENU_ITEM(SpPrimitiveArch(component).widget),
				      (SpButtonPart(component).set == SP_ON ? TRUE : FALSE));
#endif
    } else if (SpPrimitiveArch(component).widget != NULL) {
#if GTK_CHECK_VERSION(3,0,0)
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(SpPrimitiveArch(component).widget),
				     (SpButtonPart(component).set == SP_ON ? TRUE : FALSE));
#else
	gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(SpPrimitiveArch(component).widget),
				    (SpButtonPart(component).set == SP_ON ? TRUE : FALSE));
#endif
    }
    
    return;
}

gboolean spGetToggleStateGtk(spComponent component)
{
    gboolean active = FALSE;
    
    if (spIsSubClass(component, SpMenuItem) == SP_TRUE) {
#if GTK_CHECK_VERSION(3,0,0)
	active = gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(SpPrimitiveArch(component).widget));
#else
	active = GTK_CHECK_MENU_ITEM(SpPrimitiveArch(component).widget)->active;
#endif
    } else if (SpPrimitiveArch(component).widget != NULL) {
#if GTK_CHECK_VERSION(3,0,0)
	active = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(SpPrimitiveArch(component).widget));
#else
	active = GTK_TOGGLE_BUTTON(SpPrimitiveArch(component).widget)->active;
#endif
    }

    return active;
}
    
void spGetToggleStateArch(spComponent component)
{
    spDebug(30, "spGetToggleStateArch", "in\n");

    if (spGetToggleStateGtk(component)) {
	SpButtonPart(component).set = SP_TRUE;
    } else {
	SpButtonPart(component).set = SP_FALSE;
    }
    
    return;
}
