/*
 *        spCanvas_Mac.c
 */

#include <stdio.h>
#include <stdlib.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spTopLevelP.h>
#include <sp/spFrameP.h>
#include <sp/spDrawP.h>
#include <sp/spGraphicsP.h>
#include <sp/spSliderP.h>
#include <sp/spImageP.h>
#include <sp/spCanvasP.h>

void spDrawCanvasMac(spComponent component)
{
    Rect rect;
    
#ifdef SP_DRAW_CANVAS_BORDER
    if (SpCanvasPart(component).border_on == SP_TRUE) {
        if (spGetAppearanceVersionMac() >= 0x00000101) {
            GrafPtr save_port;

            spLockDrawMutexMac();
            
            spLockWindowPort(SpPrimitiveArch(SpGetWindow(component)).window, &save_port);
        
            rect = SpPrimitiveArch(component).rect;
            InsetRect(&rect, -SP_CANVAS_BORDER, -SP_CANVAS_BORDER);
            
            spDebug(20, "spDrawCanvasMac", "draw border\n");

            if (SpPrimitiveArch(component).sensitive_flag == SP_FALSE
                || SpFrameArch(SpGetWindow(component)).activate_flag == SP_FALSE) {
                DrawThemeSecondaryGroup(&rect, kThemeStateInactive);
            } else {
                DrawThemeSecondaryGroup(&rect, kThemeStateActive);
            }

            spUnlockWindowPort(SpPrimitiveArch(SpGetWindow(component)).window, save_port);
            
            spUnlockDrawMutexMac();
        }
    }
#endif

    return;
}

spMouseButton spGetPressedMouseButtonMac(short modifiers)
{
    spMouseButton button = SP_LBUTTON;
    
    if ((controlKey & modifiers) != 0
        || (rightControlKey & modifiers) != 0) {        /* control key down */
        spDebug(80, "spGetPressedMouseButtonMac", "SP_RBUTTON (control key)\n");
        button = SP_RBUTTON;
    } else if ((optionKey & modifiers) != 0
               || (rightOptionKey & modifiers) != 0) {        /* option key down */
        spDebug(80, "spGetPressedMouseButtonMac", "SP_MBUTTON (option key)\n");
        button = SP_MBUTTON;        
    }

    return button;
}

spBool spHandleCanvasMouseMac(spComponent component, int button_id, Point point,
                              short modifiers, spBool pressed)
{
    spCallbackType call_type;
    spCallbackReason call_reason;
    spMouseButton button;
    GrafPtr save_port;
    Point gpoint;
    spComponent window;
    static spComponent pressed_canvas = NULL;
    
    if (spIsCanvas(component) == SP_FALSE) return SP_FALSE;
        
    if (pressed == SP_TRUE) {
        window = SpGetWindow(component);
        if (/*SpFramePart(window).window_type == SP_TRANSIENT_WINDOW*/
            SpCanvasPart(component).capture_flag == SP_TRUE) {
            spSelectMenuBarMac(window, NULL);
            
            gpoint = point;
            spLockWindowPort(SpPrimitiveArch(window).window, &save_port);
            LocalToGlobal(&gpoint);
            spUnlockWindowPort(SpPrimitiveArch(window).window, save_port);
            spDragWindowMac(SpPrimitiveArch(window).window, gpoint);

            return /*SP_TRUE*/SP_FALSE;
        }
        
        spFocusComponentMac(window, component);
            
        pressed_canvas = component;
    } else {
        if (component != pressed_canvas) {
            return SP_FALSE;
        }
        pressed_canvas = NULL;
    }
        
    if (pressed == SP_TRUE) {
        SpFrameArch(SpGetWindow(component)).key_send_component = component;        
    }

    if (button_id > 1) {
        button = button_id - 1;
    } else {
        button = spGetPressedMouseButtonMac(modifiers);
    }
    spDebug(40, "spHandleCanvasMouseMac", "button_id = %d, button = %d\n", button_id, button);
    
    if (button == SP_RBUTTON) {
        call_reason = (pressed == SP_TRUE ? SP_CR_RBUTTON_PRESS : SP_CR_RBUTTON_RELEASE);
    } else if (button == SP_MBUTTON) {
        call_reason = (pressed == SP_TRUE ? SP_CR_MBUTTON_PRESS : SP_CR_MBUTTON_RELEASE);
    } else {
        call_reason = (pressed == SP_TRUE ? SP_CR_LBUTTON_PRESS : SP_CR_LBUTTON_RELEASE);
    }
    call_type = (pressed == SP_TRUE ? SP_BUTTON_PRESS_CALLBACK : SP_BUTTON_RELEASE_CALLBACK);
    spDebug(40, "spHandleCanvasMouseMac", "pressed = %d, button = %d, call_reason = %d\n",
            pressed, button, call_reason);
    
    return spPostMessageMac(component, call_type, call_reason);
}

spBool spHandleCanvasKeyDownMac(spComponent component, EventRecord *event)
{
    spBool flag = SP_FALSE;
    
    if (spIsCanvas(component) == SP_TRUE) {
        if (spIsVisible(component) == SP_TRUE
            && SpPrimitiveArch(component).sensitive_flag == SP_TRUE
            && SpPrimitiveArch(component).map_flag == SP_TRUE) {
            SpPrimitiveArch(component).event = *event;
            
#if !TARGET_API_MAC_CARBON
            if (SpCanvasPart(component).accept_wheel_event == SP_TRUE
                || spIsScrollCanvas(component) == SP_TRUE) {
                char key;
                key = charCodeMask & event->message;
                if (key == kPageUpCharCode || key == kPageDownCharCode) {
                    double delta = 0.0;
                    
                    if (key == kPageUpCharCode) {
                        delta = 1.0;
                    } else {
                        delta = -1.0;
                    }
                    
                    if ((cmdKey & event->modifiers) || (optionKey & event->modifiers)) {
                        double relative_factor;
                        
                        relative_factor = pow(2.0, delta);
                        if (spIsScrollCanvas(component) == SP_TRUE) {
                            flag = spZoomScrollCanvas(component, relative_factor);
                        } else {
                            flag = spUpdateCanvasContentFactor(component, relative_factor, SP_TRUE);
                            if (flag == SP_TRUE) {
                                flag = spPostMessageMac(component, SP_ZOOM_CALLBACK, SP_CR_ZOOM);
                            }
                        }
                    } else {
                        if (spIsScrollCanvas(component) == SP_TRUE) {
                            flag = spHandleScrollCanvasWheelMac(component, delta, spIsTrue(shiftKey & event->modifiers));
                        } else {
                            flag = spPostMessageMac(component, SP_WHEEL_CALLBACK, SP_CR_WHEEL);
                        }
                    }
                    if (flag == SP_TRUE) {
                        return SP_TRUE;
                    }
                }
            }
#endif
            
            flag = spPostMessageMac(component, SP_KEY_PRESS_CALLBACK, SP_CR_KEY_PRESS);
        }
    }

    return flag;
}

spBool spHandleCanvasKeyUpMac(spComponent component, EventRecord *event)
{
    spBool flag = SP_FALSE;
    
    if (spIsCanvas(component) == SP_TRUE) {
        if (spIsVisible(component) == SP_TRUE
            && SpPrimitiveArch(component).sensitive_flag == SP_TRUE
            && SpPrimitiveArch(component).map_flag == SP_TRUE) {
            SpPrimitiveArch(component).event = *event;
            
#if !TARGET_API_MAC_CARBON
            if (SpCanvasPart(component).accept_wheel_event == SP_TRUE) {
                char key;
                key = charCodeMask & event->message;
                if (key == kPageUpCharCode || key == kPageDownCharCode) {
                    return SP_TRUE;
                }
            }
#endif
            
            flag = spPostMessageMac(component, SP_KEY_RELEASE_CALLBACK, SP_CR_KEY_RELEASE);
        }
    }

    return flag;
}

spBool spHandleScrollCanvasWheelMac(spComponent canvas, double delta, spBool horizontal_flag)
{
    int value;
    spBool flag = SP_FALSE;

    if (horizontal_flag == SP_TRUE 
        || (horizontal_flag == SP_FALSE && SpCanvasPart(canvas).wheel_target_horizontal == SP_TRUE)) {
        if (SpCanvasArch(canvas).hscroll != NULL) {
            if (horizontal_flag == SP_FALSE && SpCanvasPart(canvas).wheel_target_horizontal == SP_TRUE) {
                if (SpCanvasPart(canvas).scroll_left_by_wheel_down == SP_TRUE) {
                    delta *= -1.0;
                }
            }
            value = SpCanvasPart(canvas).h_value;
            spMoveCanvasScrollBarValue(&value, -delta,
                                       SpCanvasPart(canvas).h_minimum,
                                       SpCanvasPart(canvas).h_maximum,
                                       SpCanvasPart(canvas).h_page_size,
                                       SpCanvasPart(canvas).h_wheel_increment);
                        
            SetControlValue(SpPrimitiveArch(SpCanvasArch(canvas).hscroll).control, value);
            spPostSliderValueChangeMac(SpCanvasArch(canvas).hscroll, value);
            flag = SP_TRUE;
        }
    } else {
        if (SpCanvasArch(canvas).vscroll != NULL) {
            value = SpCanvasPart(canvas).v_value;
            spMoveCanvasScrollBarValue(&value, -delta,
                                       SpCanvasPart(canvas).v_minimum,
                                       SpCanvasPart(canvas).v_maximum,
                                       SpCanvasPart(canvas).v_page_size,
                                       SpCanvasPart(canvas).v_wheel_increment);
                        
            SetControlValue(SpPrimitiveArch(SpCanvasArch(canvas).vscroll).control, value);
            spPostSliderValueChangeMac(SpCanvasArch(canvas).vscroll, value);
            flag = SP_TRUE;
        }
    }

    return flag;
}

spBool spIsCanvasExposedArch(spComponent component)
{
    if (/*SpComponentPart(component).client_width >= 0
          && SpComponentPart(component).client_height >= 0*/SpCanvasPart(component).exposed) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

void spExposeCanvasCBArch(spComponent component)
{
    return;
}

spBool spUpdateCanvasScrollBarArch(spComponent component)
{
    spBool flag = SP_FALSE;
    
    if (SpCanvasArch(component).hscroll != NULL) {
        spSetParams(SpCanvasArch(component).hscroll,
                    SppMinimum, SpCanvasPart(component).h_minimum,
                    SppMaximum, SpCanvasPart(component).h_maximum,
                    SppIncrement, SpCanvasPart(component).h_increment,
                    SppPageIncrement, SpCanvasPart(component).h_page_increment,
                    SppPageSize, SpCanvasPart(component).h_page_size,
                    SppValue, SpCanvasPart(component).h_value,
                    NULL);
        flag = SP_TRUE;
    }

    if (SpCanvasArch(component).vscroll != NULL) {
        spSetParams(SpCanvasArch(component).vscroll,
                    SppMinimum, SpCanvasPart(component).v_minimum,
                    SppMaximum, SpCanvasPart(component).v_maximum,
                    SppIncrement, SpCanvasPart(component).v_increment,
                    SppPageIncrement, SpCanvasPart(component).v_page_increment,
                    SppPageSize, SpCanvasPart(component).v_page_size,
                    SppValue, SpCanvasPart(component).v_value,
                    NULL);
        flag = SP_TRUE;
    }
    
    return flag;
}

static void canvasScrollCB(spComponent component, void *data)
{
    spComponent canvas = (spComponent)data;

    if (SpComponentPart(component).orientation == SP_VERTICAL) {
        spGetSliderValue(component, &SpCanvasPart(canvas).v_value);
    } else {
        spGetSliderValue(component, &SpCanvasPart(canvas).h_value);
    }
    
    spPostMessageMac(canvas, SP_SCROLL_CALLBACK, SP_CR_SCROLL);

    return;
}

void spCanvasCreateArch(spComponent component)
{
    SpCanvasArch(component).hscroll = NULL;
    SpCanvasArch(component).vscroll = NULL;
#if TARGET_API_MAC_CARBON
    SpCanvasArch(component).axis = kEventMouseWheelAxisY;
    SpCanvasArch(component).wheelDelta = 0;
#endif
    
    spImageCreateArch(component);
    spSetNeedUpdateMac(component);
    spSetNeedMoveCallMac(component);

    if (SpCanvasPart(component).focusable == SP_TRUE) {
        spSetKeySendComponentMac(component, SP_TRUE);
    }
    
    if (SpCanvasPart(component).h_scroll == SP_TRUE) {
        spDebug(50, "spCanvasCreateArch", "create horizontal scroll bar\n");
        
        /* create horizontal scroll bar */
        SpCanvasArch(component).hscroll = spCreateScrollBar(SpGetParent(component), "_CanvasScrollBarHorizontal",
                                                            SppOrientation, SP_HORIZONTAL,
                                                            SppMinimum, SpCanvasPart(component).h_minimum,
                                                            SppMaximum, SpCanvasPart(component).h_maximum,
                                                            SppIncrement, SpCanvasPart(component).h_increment,
                                                            SppPageIncrement, SpCanvasPart(component).h_page_increment,
                                                            SppPageSize, SpCanvasPart(component).h_page_size,
                                                            SppValue, SpCanvasPart(component).h_value,
                                                            SppTrackCallbackOn, SP_TRUE,
                                                            SppCallbackFunc, canvasScrollCB,
                                                            SppCallbackData, component,
                                                            NULL);
    }
    if (SpCanvasPart(component).v_scroll == SP_TRUE) {
        spDebug(50, "spCanvasCreateArch", "create vertical scroll bar\n");
        
        /* create vertical scroll bar */
        SpCanvasArch(component).vscroll = spCreateScrollBar(SpGetParent(component), "_CanvasScrollBarVertical",
                                                            SppOrientation, SP_VERTICAL,
                                                            SppMinimum, SpCanvasPart(component).v_minimum,
                                                            SppMaximum, SpCanvasPart(component).v_maximum,
                                                            SppIncrement, SpCanvasPart(component).v_increment,
                                                            SppPageIncrement, SpCanvasPart(component).v_page_increment,
                                                            SppPageSize, SpCanvasPart(component).v_page_size,
                                                            SppValue, SpCanvasPart(component).v_value,
                                                            SppTrackCallbackOn, SP_TRUE,
                                                            SppCallbackFunc, canvasScrollCB,
                                                            SppCallbackData, component,
                                                            NULL);
    }
    
    if (SpPrimitivePart(component).draw_func != NULL
        && (SpCanvasArch(component).hscroll != NULL || SpCanvasArch(component).vscroll != NULL)) {
        spAddCallback(component, SP_SCROLL_CALLBACK,
                      SpPrimitivePart(component).draw_func, SpPrimitivePart(component).draw_data);
    }
    
    return;
}

void spCanvasSetParamsArch(spComponent component)
{
    spUpdateCanvasScrollBarArch(component);
    
    return;
}
    
void spCanvasDestroyArch(spComponent component)
{
    return;
}
    
spBool spRefreshCanvasArch(spComponent component)
{
    Rect src_rect;
    Rect dest_rect;
    WindowPtr window;
    PixMapHandle pixmap;
    BitMap *bitmap;
    RgnHandle orig_rgn;
    /*GrafPtr save_port;*/
    CGrafPtr window_port;

    spLockControlMutexMac();
    spLockDrawMutexMac();
    
    window = SpPrimitiveArch(SpGetWindow(component)).window;

    window_port = GetWindowPort(window);
#if TARGET_API_MAC_CARBON
    LockPortBits(window_port);
#endif
    /*GetPort(&save_port);*/
    
    if ((pixmap = spGetLockedPixmapMac(component)) != NULL) {
#if 1
        GWorldPtr save_gwp;
        GDHandle save_gdh;
        
        spGetOriginalGWorldMac(&save_gwp, &save_gdh);
        SetGWorld(window_port, NULL);
#else
        SetPortWindowPort(window);
#endif
    
        bitmap = spGetPortBitMapMac(window);
        
        SetRect(&src_rect, 0, 0, SpComponentPart(component).client_width,
                SpComponentPart(component).client_height);
        dest_rect = SpPrimitiveArch(component).rect;

        spDebug(50, "spRefreshCanvasArch",
                "src: left = %d, right = %d, top = %d, bottom = %d\n",
                src_rect.left, src_rect.right, src_rect.top, src_rect.bottom);
        spDebug(50, "spRefreshCanvasArch",
                "dest: left = %d, right = %d, top = %d, bottom = %d\n",
                dest_rect.left, dest_rect.right, dest_rect.top, dest_rect.bottom);

        /* store original region */
        orig_rgn = NewRgn();
        GetClip(orig_rgn);
        
        /* set clip region */
        SetClip(spGetVisibleRegionMac(window));

        /* set default RGB */
        spGetOriginalRGBMac();
        spSetNormalRGBMac();

        /* copy pixmap to window */
        CopyBits((BitMap *)*pixmap, bitmap, &src_rect, &dest_rect, srcCopy, NULL);
        spUnlockPixmapMac(component, pixmap);

        /* restore RGB */
        spSetOriginalRGBMac();
        
        /* restore original region */
        SetClip(orig_rgn);
        DisposeRgn(orig_rgn);

#if 1
        SetGWorld(save_gwp, save_gdh);
#endif
    }
    
    /*SetPort(save_port);*/
#if TARGET_API_MAC_CARBON
    UnlockPortBits(window_port);
#endif
        
#if TARGET_API_MAC_CARBON
    if (QDIsPortBuffered(window_port)) {
        QDFlushPortBuffer(window_port, NULL);
    }
#endif
    
    spUnlockDrawMutexMac();
    spUnlockControlMutexMac();

    spDebug(80, "spRefreshCanvasArch",        "done\n");
    
    return SP_TRUE;
}

spBool spRedrawCanvasArch(spComponent component)
{
    spRedrawImageArch(component);
    return SP_TRUE;
}

spBool spSetCanvasCaptureArch(spComponent window)
{
    return SP_TRUE;
}

void spSetCanvasCursorMac(spComponent component)
{
    CursHandle cursor;
    
    if (SpPrimitiveArch(component).cursor_type != SP_CURSOR_UNKNOWN
        && (cursor = spGetCursorMac(SpPrimitiveArch(component).cursor_type)) != NULL) {
        spDebug(20, "spSetCanvasCursorMac", "in\n");
        spSetCursorMac(cursor);
    } else {
        spUnsetCursorMac();
    }
    
    return;
}

void spUnsetCanvasCursorMac(spComponent component)
{
    spDebug(20, "spUnsetCanvasCursorMac", "in\n");
    spUnsetCursorMac();
    return;
}

spBool spSetCanvasCursorArch(spComponent component, spCursor cursor)
{
    SpPrimitiveArch(component).cursor_type = cursor->type;
    return SP_TRUE;
}

spBool spUnsetCanvasCursorArch(spComponent component)
{
    SpPrimitiveArch(component).cursor_type = SP_CURSOR_UNKNOWN;
    return SP_TRUE;
}

spBool spIsCanvasFocusedArch(spComponent component)
{
    if (SpFrameArch(SpGetWindow(component)).key_send_component == component) {
        return SP_TRUE;
    }
    
    return SP_FALSE;
}

spBool spZoomScrollCanvasArch(spComponent component)
{
    return SP_TRUE;
}
