/*
 *	spDialogBox.c
 */

#include <stdio.h>
#include <stdlib.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spOption.h>
#include <sp/spMemory.h>

#include <sp/spComponent.h>
#include <sp/spFrame.h>
#include <sp/spContainer.h>
#include <sp/spButton.h>
#include <sp/spText.h>
#include <sp/spParamField.h>
#include <sp/spCanvas.h>
#include <sp/spDraw.h>

#include <sp/spDialogP.h>
#include <sp/spGraphicsP.h>
#include <sp/spDialogBoxP.h>

static spParamTable sp_dialog_box_param_tables[] = {
    {SppPopupStyle, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spDialogBox, dialog_box.popup_style), SP_MODAL_POPUP_STRING},
    {SppCloseStyle, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spDialogBox, dialog_box.close_style), /*SP_NO_CLOSE_STRING*/NULL},
    {SppDialogBoxButtonType, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spDialogBox, dialog_box.button_type), SP_DB_OK_CANCEL_STRING},
    {SppResizable, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spDialogBox, dialog_box.resize_flag), SP_FALSE_STRING},
    {SppIconfiable, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spDialogBox, dialog_box.iconfy_flag), SP_FALSE_STRING},
    {SppSimplifiable, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spDialogBox, dialog_box.simplify_flag), SP_FALSE_STRING},
    {SppHelpButtonVisible, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spDialogBox, dialog_box.help_button_visible), SP_FALSE_STRING},
    {SppParentWindow, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spDialogBox, dialog_box.parent_window), NULL},
    {SppSystemWindow, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spDialogBox, dialog_box.system_window), NULL},
};

spDialogBoxClassRec SpDialogBoxClassRec = {
    /* spObjectClassPart */
    {
	SpDialogBox,
	(spObjectClass)&SpComponentClassRec,
	sizeof(spDialogBoxRec),
	spArraySize(sp_dialog_box_param_tables),
	sp_dialog_box_param_tables,
	spDialogBoxPartInit,
	spDialogBoxPartFree,
	SP_FALSE,
	NULL,
	NULL,
	spDialogBoxCreate,
	spDialogBoxDestroy,
	spDialogBoxSetParams,
	NULL,
    },
    /* spComponentClassPart */
    {
	SP_TRUE,
	SP_TRUE,
	SP_TRUE,
	spDialogBoxMap,
	spDialogBoxUnmap,
	spDialogBoxSetSize,
	spDialogBoxGetSize,
	spDialogBoxGetClientSize,
	spDialogBoxSetSensitive,
	spDialogBoxIsComponentType,
	
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
    },
    /* spDialogBoxClassPart */
    {
	0,
    },
};

spComponentClass SpDialogBoxClass = (spComponentClass)&SpDialogBoxClassRec;

void spDialogBoxPartInit(spObject object)
{
    spComponent component = (spComponent)object;
    
    SpDialogBoxPart(component).window = NULL;
    SpDialogBoxPart(component).box = NULL;
    SpDialogBoxPart(component).separator = NULL;
    SpDialogBoxPart(component).button_box = NULL;
    SpDialogBoxPart(component).ok_button = NULL;
    SpDialogBoxPart(component).cancel_button = NULL;
    SpDialogBoxPart(component).apply_button = NULL;
    SpDialogBoxPart(component).close_style = SP_NO_CLOSE;
    SpDialogBoxPart(component).parent_window = NULL;
    SpDialogBoxPart(component).system_window = NULL;
    
    SpComponentPart(component).orientation = SP_VERTICAL;
    SpComponentPart(component).margin_width = SP_DEFAULT_MARGIN;
    SpComponentPart(component).margin_height = SP_DEFAULT_MARGIN;
    SpComponentPart(component).spacing = SP_DEFAULT_SPACING;

    return;
}

void spDialogBoxPartFree(spObject object)
{
    return;
}

spBool spDialogBoxCreate(spObject object)
{
    int i;
    char *name;
    char *title;
    int box_size;
    int num_button;
    spComponent component = (spComponent)object;

    if (SpComponentPart(component).title == NULL) {
	title = SpGetName(component);
    } else {
	title = SpComponentPart(component).title;
    }

#if defined(ANDROID) || defined(IPHONE)
    name = "_DialogBox_Window";
#else
    name = SpGetName(component);
#endif

    /* create dialog window */
    SpDialogBoxPart(component).window =
	spCreateFrame(name,
		      SppWindowType, SP_DIALOG_WINDOW,
		      SppTitle, title,
		      SppPopupStyle, SpDialogBoxPart(component).popup_style,
		      SppCloseStyle, SpDialogBoxPart(component).close_style,
		      SppResizable, SpDialogBoxPart(component).resize_flag,
		      SppIconfiable, SpDialogBoxPart(component).iconfy_flag,
		      SppSimplifiable, SpDialogBoxPart(component).simplify_flag,
		      SppParentWindow, SpDialogBoxPart(component).parent_window,
		      SppSystemWindow, SpDialogBoxPart(component).system_window,
		      SppGroupId, SpComponentPart(component).group_id,
		      SppUserData, SpComponentPart(component).user_data,
		      SppMarginWidth, SP_DEFAULT_DIALOG_MARGIN,
		      SppMarginHeight, SP_DEFAULT_DIALOG_MARGIN,
		      SppSpacing, 0,
		      SppHelpButtonVisible, SpDialogBoxPart(component).help_button_visible,
		      SppHelpPath, SpComponentPart(component).help_path,
#if defined(ANDROID) || defined(IPHONE)
		      SppDialogBoxButtonType, SpDialogBoxPart(component).button_type,
		      SppCallbackFunc, SpComponentPart(component).call_func,
		      SppCallbackData, SpComponentPart(component).call_data,
#endif
		      NULL);
    if (SpDialogBoxPart(component).close_style == SP_CALLBACK_CLOSE
	&& SpComponentPart(component).call_func != NULL) {
	spAddCallback(SpDialogBoxPart(component).window, SP_CLOSE_CALLBACK,
		      SpComponentPart(component).call_func,
		      SpComponentPart(component).call_data);
    }

    /* create main box */
    box_size = -(SP_DEFAULT_BUTTON_HEIGHT + 2 * SP_DEFAULT_DIALOG_BUTTON_MARGIN);
    SpDialogBoxPart(component).box =
	spCreateBox(SpDialogBoxPart(component).window, "_DialogBox_MainBox", box_size,
		    SppOrientation, SpComponentPart(component).orientation,
		    SppMarginWidth, SpComponentPart(component).margin_width,
		    SppMarginHeight, SpComponentPart(component).margin_height,
		    SppSpacingOn, SpComponentPart(component).spacing_flag,
		    SppSpacing, SpComponentPart(component).spacing,
		    NULL);

    SpComponentPart(component).component = SpDialogBoxPart(component).box;
    SpGetWindow(component) = SpDialogBoxPart(component).window;

#if defined(IPHONE)
    if (1) {
	return SP_TRUE;
    }
#elif defined(ANDROID)
    if (SpDialogBoxPart(component).popup_style == SP_MODAL_POPUP
	|| SpDialogBoxPart(component).popup_style == SP_SYSTEM_MODAL_POPUP) {
	return SP_TRUE;
    }
#endif
    
    /* create box for buttons */
    SpDialogBoxPart(component).button_box =
	spCreateBox(SpDialogBoxPart(component).window, "_DialogBox_ButtonBox", 0,
		    SppOrientation, SP_HORIZONTAL,
		    SppMarginWidth, SP_DEFAULT_DIALOG_BUTTON_MARGIN,
		    SppMarginHeight, SP_DEFAULT_DIALOG_BUTTON_MARGIN,
		    SppSpacing, SP_DEFAULT_DIALOG_BUTTON_SPACING,
		    NULL);

    num_button = 0;
    for (i = 0; i < 3; i++) {
	switch (i) {
	  case SP_DIALOG_OK_BUTTON_ORDER:
	    switch (SpDialogBoxPart(component).button_type) {
	      case SP_DB_YES_NO:
	      case SP_DB_YES_NO_CANCEL:
		/* create yes button */
		SpDialogBoxPart(component).ok_button =
		    spCreatePushButton(SpDialogBoxPart(component).button_box, "_YesButton",
				       SppTitle, SP_YES_LABEL,
				       SppDefaultButton, SP_TRUE,
				       SppCallbackFunc, SpComponentPart(component).call_func,
				       SppCallbackData, SpComponentPart(component).call_data,
				       NULL);
		num_button++;
		break;
	      case SP_DB_RETRY_CANCEL:
		/* create retry button */
		SpDialogBoxPart(component).ok_button =
		    spCreatePushButton(SpDialogBoxPart(component).button_box, "_RetryButton",
				       SppTitle, SP_RETRY_LABEL,
				       SppDefaultButton, SP_TRUE,
				       SppCallbackFunc, SpComponentPart(component).call_func,
				       SppCallbackData, SpComponentPart(component).call_data,
				       NULL);
		num_button++;
		break;
	      default:
		/* create OK button */
		SpDialogBoxPart(component).ok_button =
		    spCreatePushButton(SpDialogBoxPart(component).button_box, "_OkButton",
				       SppTitle, SP_OK_LABEL,
				       SppDefaultButton, SP_TRUE,
				       SppCallbackFunc, SpComponentPart(component).call_func,
				       SppCallbackData, SpComponentPart(component).call_data,
				       NULL);
		num_button++;
		break;
	    }
	    break;
	    
	  case SP_DIALOG_CANCEL_BUTTON_ORDER:
	    switch (SpDialogBoxPart(component).button_type) {
	      case SP_DB_OK_CANCEL:
	      case SP_DB_RETRY_CANCEL:
	      case SP_DB_OK_CANCEL_APPLY:
		/* create cancel button */
		SpDialogBoxPart(component).cancel_button =
		    spCreatePushButton(SpDialogBoxPart(component).button_box, "_CancelButton",
				       SppTitle, SP_CANCEL_LABEL,
				       SppCancelButton, SP_TRUE,
				       SppCallbackFunc, SpComponentPart(component).call_func,
				       SppCallbackData, SpComponentPart(component).call_data,
				       NULL);
		num_button++;
		break;
	      case SP_DB_YES_NO:
	      case SP_DB_YES_NO_CANCEL:
		/* create no button */
		SpDialogBoxPart(component).cancel_button =
		    spCreatePushButton(SpDialogBoxPart(component).button_box, "_NoButton",
				       SppTitle, SP_NO_LABEL,
				       SppCallbackFunc, SpComponentPart(component).call_func,
				       SppCallbackData, SpComponentPart(component).call_data,
				       NULL);
		num_button++;
		break;
	      default:
		break;
	    }
	    break;
	    
	  case SP_DIALOG_APPLY_BUTTON_ORDER:
	    switch (SpDialogBoxPart(component).button_type) {
	      case SP_DB_OK_CANCEL_APPLY:
		/* create apply button */
		SpDialogBoxPart(component).apply_button =
		    spCreatePushButton(SpDialogBoxPart(component).button_box, "_ApplyButton",
				       SppTitle, SP_APPLY_LABEL,
				       SppCallbackFunc, SpComponentPart(component).call_func,
				       SppCallbackData, SpComponentPart(component).call_data,
				       NULL);
		num_button++;
		break;
	      case SP_DB_YES_NO_CANCEL:
		/* create cancel button */
		SpDialogBoxPart(component).apply_button =
		    spCreatePushButton(SpDialogBoxPart(component).button_box, "_CancelButton",
				       SppTitle, SP_CANCEL_LABEL,
				       SppCancelButton, SP_TRUE,
				       SppCallbackFunc, SpComponentPart(component).call_func,
				       SppCallbackData, SpComponentPart(component).call_data,
				       NULL);
		num_button++;
		break;
	      default:
		break;
	    }
	    break;
	}
    }
    
    return SP_TRUE;
}

spBool spDialogBoxDestroy(spObject object)
{
    spComponent component = (spComponent)object;
    
    spDestroyWindow(SpGetWindow(component));

    return SP_TRUE;
}

spBool spDialogBoxSetParams(spObject object)
{
    spComponent component = (spComponent)object;
    
    spSetParams(SpDialogBoxPart(component).window,
		SppTitle, SpComponentPart(component).title,
		NULL);
    
    return SP_TRUE;
}

void spDialogBoxMap(spComponent component)
{
    spPopupWindow(SpGetWindow(component));
    
    return;
}

void spDialogBoxUnmap(spComponent component)
{
    spPopdownWindow(SpGetWindow(component));
    
    return;
}

spBool spDialogBoxSetSize(spComponent component, int width, int height)
{
    return spSetSize(SpGetWindow(component), width, height);
}

spBool spDialogBoxGetSize(spComponent component, int *width, int *height)
{
    return spGetSize(SpGetWindow(component), width, height);
}

spBool spDialogBoxGetClientSize(spComponent component, int *width, int *height)
{
    return spGetClientSize(SpGetWindow(component), width, height);
}

spBool spDialogBoxSetSensitive(spComponent component, spBool flag)
{
    return spSetSensitive(SpGetWindow(component), flag);
}

spBool spDialogBoxIsComponentType(spComponent component, const char *class_name)
{
    if (streq(class_name, SpDialog)) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spBool spIsDialogBox(spComponent component)
{
    return spIsSubClass(component, SpDialogBox);
}

spComponent spCreateDialogBox(const char *name, ...)
{
    int num_arg = 0;
    spArg args[SP_MAX_NUM_ARG];
    va_list argp;
    
    if (spIsTopLevelCreated() == SP_FALSE) spError(1, "spInitialize should be called.\n");
    
    va_start(argp, name);
    spGetArgs(argp, args, num_arg);
    va_end(argp);

    spDebug(60, "spCreateDialogBox", "spGetArgs done\n");
    
    return spCreateComponentArg(SpDialogBoxClass, NULL, name, NULL, args, num_arg);
}

void spGetOptionParamFieldsData(spComponent parent)
{
    spBool set;
    char *string;
    spComponent child;
    spOption *option;

    if (parent == NULL) return;
    
    child = parent;
    while ((child = spGetChild(child)) != NULL) {
	if ((option = (spOption *)spGetUserData(child)) != NULL) {
	    break;
	}
    }
    
    while (child != NULL) {
	spDebug(60, "spGetOptionParamFieldsData", "class_name = %s\n", SpGetClassName(child));
    
	if ((option = (spOption *)spGetUserData(child)) != NULL) {
	    if (spEqType(option->type, SP_TYPE_BOOL)) {
		if (spGetToggleState(child, &set) == SP_TRUE) {
		    if (set == SP_TRUE) {
			spConvertOptionValue(option, "True");
		    } else {
			spConvertOptionValue(option, "False");
		    }
		}
	    } else if ((string = xspGetTextString(child)) != NULL) {
		spDebug(60, "spGetOptionParamFieldsData", "string = %s\n", string);
	    
		/* set option value */
		spConvertOptionValue(option, string);
	    
		xfree(string);
	    }
	}

	child = spGetNextComponent(child);
    }
    
    spDebug(60, "spGetOptionParamFieldsData", "done\n");
	
    return;
}

void spGetOptionDialogData(spComponent component)
{
    if (component == NULL) return;
    
    spGetOptionParamFieldsData(spGetWindow(component));
    
    return;
}

void spResetOptionParamFieldsData(spComponent parent)
{
    spBool set;
    spComponent child;
    spOption *option;
    char buf[SP_MAX_LINE];

    if (parent == NULL) return;
    
    child = parent;
    while ((child = spGetChild(child)) != NULL) {
	if ((option = (spOption *)spGetUserData(child)) != NULL) {
	    break;
	}
    }
    
    while (child != NULL) {
	spDebug(60, "spResetOptionParamFieldsData", "class_name = %s\n", SpGetClassName(child));
    
	if ((option = (spOption *)spGetUserData(child)) != NULL) {
	    if (spGetOptionValueString(option, buf, sizeof(buf))) {
		if (spEqType(option->type, SP_TYPE_BOOL)) {
		    if (strcaseeq(buf, "True")) {
			set = SP_TRUE;
		    } else {
			set = SP_FALSE;
		    }
		    spSetToggleState(child, set);
		} else {
		    if ((spEqType(option->type, SP_TYPE_COMBO) && spEqType(option->type, SP_TYPE_EDITABLE))
			|| spEqType(option->type, SP_TYPE_LIST)) {
			spSelectListItem(child, buf);
		    } else {
			spSetTextString(child, buf);
		    }
		}
	    }
	}

	child = spGetNextComponent(child);
    }
    
    return;
}

void spResetOptionDialogData(spComponent component)
{
    if (component == NULL) return;
    
    spResetOptionParamFieldsData(spGetWindow(component));
    return;
}

void spPopdownOptionDialogCB(spComponent component, void *data)
{
    spCallbackReason reason = SP_CR_NONE;
    
    reason = spGetCallbackReason(component);

    if (reason == SP_CR_OK || reason == SP_CR_APPLY) {
	spGetOptionDialogData(component);
    }
    
    if (reason == SP_CR_OK || reason == SP_CR_CANCEL) {
	spPopdownWindow(component);
    }

    return;
}

spBool spCreateOptionParamFields(spComponent parent, spOptions options)
{
    int num_option;
    spOption *option_list;
    int i, j;
    const char *desc;
    char *string;
    char title[SP_MAX_LINE];
    char dimention[SP_MAX_LINE];
    char buf[SP_MAX_LINE];
    spBool flag;
    spBool edit_flag;
    spBool select_only;
    int field_height;
    int field_offset, field_size;
    spComponent component;

    if (spIsCreated(parent) == SP_FALSE || options == NULL) return SP_FALSE;

    num_option = spGetNumOption(options);
    option_list = spGetOptionList(options);
    
    spDebug(80, "spCreateOptionParamFields", "num_option = %d\n", num_option);

    field_height = 0;
    for (i = 0; i < num_option; i++) {
	if (spEqType(option_list[i].type, SP_TYPE_NO_GUI)) {
	    continue;
	}
		
	if (spEqType(option_list[i].type, SP_TYPE_EDITABLE)) {
	    edit_flag = SP_TRUE;
	    select_only = SP_FALSE;
	} else {
	    edit_flag = SP_FALSE;
	    select_only = SP_TRUE;
	}

	if ((string = xspGetOptionLabel(&option_list[i], 1)) != NULL
	    || (string = xspGetOptionLabel(&option_list[i], 0)) != NULL) {
	    strcpy(title, string);
	    xfree(string);
	} else {
	    strcpy(title, "");
	}
	if ((string = xspGetOptionLabel(&option_list[i], 4)) != NULL) {
	    strcpy(dimention, string);
	    xfree(string);
	} else {
	    strcpy(dimention, "");
	}

	if ((string = xspGetOptionLabel(&option_list[i], 2)) != NULL) {
	    field_offset = atoi(string);
	    xfree(string);
	} else {
	    field_offset = 0;
	}
	if ((string = xspGetOptionLabel(&option_list[i], 3)) != NULL) {
	    field_size = atoi(string);
	    xfree(string);
	} else {
	    field_size = 0;
	}

	spDebug(60, "spCreateOptionParamFields", "title = %s\n", title);

	if (spEqType(option_list[i].type, SP_TYPE_SHOW_TIP)) {
	    desc = option_list[i].desc;
	} else {
	    desc = NULL;
	}
	
	if (spEqType(option_list[i].type, SP_TYPE_BOOL)) {
	    flag = SP_FALSE;
	    if (spGetOptionValueString(&option_list[i], buf, sizeof(buf))) {
		spDebug(60, "spCreateOptionParamFields", "BOOL: spGetOptionValueString result: buf = %s\n", buf);
		if (strcaseeq(buf, "True")) {
		    flag = SP_TRUE;
		}
	    } else if (option_list[i].def_value != NULL) {
		flag = spStrToBool(option_list[i].def_value);
	    }
	    
	    /* create check box */
	    component = spCreateCheckBox(parent, "checkBox",
					 SppTitle, title,
					 SppSet, flag,
					 SppDescription, desc,
					 SppUserData, &option_list[i],
					 NULL);
	} else if (spEqType(option_list[i].type, SP_TYPE_COMBO)
		   || spEqType(option_list[i].type, SP_TYPE_LIST)) {
	    spDebug(60, "spCreateOptionParamFields", "field_offset = %d, field_size = %d\n",
		    field_offset, field_size);
	    
	    /* create combo box */
	    component = spCreateParamField(parent, "comboBox", field_height,
					   SppTitle, title,
					   SppDimention, dimention,
					   SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
					   /*SppFieldType, SP_FIELD_TYPE_TEXT,*/
					   SppEditable, edit_flag,
					   SppSelectOnly, select_only,
					   SppFieldOffset, field_offset,
					   SppFieldSize, field_size,
					   SppDescription, desc,
					   SppUserData, &option_list[i],
					   NULL);

	    if (option_list[i].def_value != NULL) {
		spDebug(80, "spCreateOptionParamFields", "option_list[%d].def_value = %s\n",
			i, option_list[i].def_value);
	    
		for (j = 0;; j++) {
		    if ((string = xspCutOptionValue(option_list[i].def_value, j)) == NULL) {
			break;
		    }
		    spDebug(100, "spCreateOptionParamFields", "j = %d, string = %s\n", j, string);

		    spAddListItem(component, string);
		    xfree(string);
		}
	    }
	    
	    if (option_list[i].def_value == NULL && spGetOptionValueString(&option_list[i], buf, sizeof(buf))) {
		spDebug(1, "spCreateOptionParamFields", "buf = %s\n", buf);
		spSelectListItem(component, buf);
	    } else {
		if ((string = xspCutOptionValue(option_list[i].def_value, -1)) != NULL) {
		    spDebug(1, "spCreateOptionParamFields", "default string = %s\n", string);
		    spSelectListItem(component, string);
		    xfree(string);
		} else {
		    spDebug(1, "spCreateOptionParamFields", "select first item as default\n");
		    spSelectListIndex(component, 0);
		}
	    }
	} else if (spEqType(option_list[i].type, SP_TYPE_STRING)
		   || spEqType(option_list[i].type, SP_TYPE_INT)
		   || spEqType(option_list[i].type, SP_TYPE_SHORT)
		   || spEqType(option_list[i].type, SP_TYPE_LONG)
		   || spEqType(option_list[i].type, SP_TYPE_FLOAT)
		   || spEqType(option_list[i].type, SP_TYPE_DOUBLE)
		   || spEqType(option_list[i].type, SP_TYPE_CHAR)
		   || spEqType(option_list[i].type, SP_TYPE_ENUM)
		   || spEqType(option_list[i].type, SP_TYPE_STRING_A)) {
	    component = spCreateParamField(parent, "textField", field_height,
					   SppTitle, title,
					   SppDimention, dimention,
					   SppFieldType, SP_FIELD_TYPE_TEXT,
					   SppEditable, SP_TRUE,
					   SppFieldOffset, field_offset,
					   SppFieldSize, field_size,
					   SppDescription, desc,
					   SppUserData, &option_list[i],
					   NULL);
	    if (option_list[i].def_value == NULL && spGetOptionValueString(&option_list[i], buf, sizeof(buf))) {
                spDebug(80, "spCreateOptionParamFields", "normal value: buf = %s\n", buf);
		spSetTextString(component, buf);
	    } else if (option_list[i].def_value != NULL) {
                spDebug(80, "spCreateOptionParamFields", "normal value: def_value = %s\n", option_list[i].def_value);
		spSetTextString(component, option_list[i].def_value);
	    }
	}
	
	spDebug(80, "spCreateOptionParamFields", "done: i = %d\n", i);
    }

    return SP_TRUE;
}

spComponent spFindOptionParamField(spComponent parent, spOption *option)
{
    spComponent child;
    spOption *option2;
    
    if (parent == NULL || option == NULL) return NULL;
    
    child = parent;
    while ((child = spGetChild(child)) != NULL) {
	if ((option2 = (spOption *)spGetUserData(child)) != NULL) {
	    break;
	}
    }
    
    while (child != NULL) {
	spDebug(60, "spFindOptionParamField", "class_name = %s\n", SpGetClassName(child));
    
	if ((option2 = (spOption *)spGetUserData(child)) != NULL
	    && option2 == option) {
	    return child;
	}
	
	child = spGetNextComponent(child);
    }

    return NULL;
}

spComponent spCreateOptionDialog(const char *name, spOptions options, ...)
{
    int num_arg = 0;
    spArg args[SP_MAX_NUM_ARG];
    va_list argp;
    spComponent dialog;

    if (spIsTopLevelCreated() == SP_FALSE) spError(1, "spInitialize should be called.\n");
    if (options == NULL) return NULL;

    va_start(argp, options);
    spGetArgs(argp, args, num_arg);
    va_end(argp);

    /* create dialog box */
    dialog = spCreateComponentArg(SpDialogBoxClass, NULL, name, NULL, args, num_arg);

    spCreateOptionParamFields(dialog, options);
    
    return dialog;
}

typedef struct _spFontSelectionStyleTable {
    const char *style_name;
    unsigned long style_mask;
} spFontSelectionStyleTable;
    
typedef struct _spFontSelectionDialog {
    spComponent dialog;
    
    spComponent family_field;
    spComponent style_field;
    spComponent size_field;
    
    spGraphics graphics;
    spComponent sample_box;
    spComponent sample_canvas;
    
    int num_style_table;
    spFontSelectionStyleTable *style_table;

    spBool lock;
    
    spCallbackReason reason;
    char font_name[SP_MAX_FONTNAME];
} *spFontSelectionDialog;

static void updateFontName(spFontSelectionDialog dialog, int *pfamily, unsigned long *pstyle, int *psize)
{
    int j;
    int size = 0;
    unsigned long style = 0L;
    char *string;
    char family_string[SP_MAX_FONTNAME];

    if ((string = xspGetTextString(dialog->family_field)) != NULL) {
	strcpy(family_string, string);
	xfree(string);
    } else {
	strcpy(family_string, "*");
    }
	
    if ((string = xspGetTextString(dialog->style_field)) != NULL) {
	for (j = 0; j < dialog->num_style_table; j++) {
	    if (streq(string, dialog->style_table[j].style_name)) {
		style = dialog->style_table[j].style_mask;
		break;
	    }
	}
	xfree(string);
    }
	
    if ((string = xspGetTextString(dialog->size_field)) != NULL) {
	size = atoi(string);
	xfree(string);
    }

    spGetFontName(dialog->font_name, family_string, style, size);

    if (pfamily != NULL) {
	*pfamily = MAX(spGetSelectedListIndex(dialog->family_field), 0);
    }
    if (pstyle != NULL) *pstyle = style;
    if (psize != NULL) *psize = size;
    
    return;
}

static void popdownFontDialogCB(spComponent component, void *data)
{
    spFontSelectionDialog dialog = (spFontSelectionDialog)data;

    dialog->reason = spGetCallbackReason(component);
    spDebug(50, "popdownFontDialogCB", "reason = %d\n", dialog->reason);
    
    if (dialog->reason == SP_CR_OK) {
	updateFontName(dialog, NULL, NULL, NULL);
    } else {
	strcpy(dialog->font_name, "");
    }
    
    spPopdownWindow(component);
    
    spDebug(50, "popdownFontDialogCB", "done\n");
    
    return;
}

static void drawFontSample(spComponent canvas, spFontSelectionDialog dialog)
{
    int width, height;
    int sx, sy, swidth, sheight;
    
    spDebug(50, "drawFontSample", "dialog->lock = %d\n", dialog->lock);
    
    if (dialog->lock == SP_TRUE) return;
    
    if (!strnone(dialog->font_name)) {
	spSetFont(dialog->graphics, dialog->font_name);
	spDebug(50, "drawFontSample", "font_name = %s\n", dialog->font_name);
    }
    if (spGetSize(canvas, &width, &height) == SP_TRUE) {
	spSetGraphicsParams(dialog->graphics, SppForeground, "white", NULL);
	spFillRectangle(canvas, dialog->graphics, 0, 0, width, height);
	spSetGraphicsParams(dialog->graphics, SppForeground, "black", NULL);

	if (spGetStringExtent(canvas, dialog->graphics, SP_FONT_SAMPLE_CONTENTS,
			      &sx, &sy, &swidth, &sheight, NULL) == SP_TRUE) {
	    spDebug(50, "drawFontSample", "width = %d, height = %d, swidth = %d, sheight = %d\n",
		    width, height, swidth, sheight);
	    spDrawString(canvas, dialog->graphics, 5, 5 + MIN(sheight, height), SP_FONT_SAMPLE_CONTENTS);
	}
    }
    
    return;
}

static void updateFontStyle(spFontSelectionDialog dialog, int family, int size)
{
    int j;
    spBool supported;
    char *selected_item;
    
    spDebug(50, "updateFontStyle", "family = %d, size = %d\n", family, size);
    
    selected_item = xspGetSelectedListItem(dialog->style_field);

    spClearList(dialog->style_field);
    spAddListItem(dialog->style_field, "*");

    for (j = 0; j < dialog->num_style_table; j++) {
	supported = spIsFontSizeSupportedArch(family,
					      dialog->style_table[j].style_mask, size);
	if (supported == SP_TRUE) {
	    spAddListItem(dialog->style_field, dialog->style_table[j].style_name);
	}
        spDebug(100, "updateFontStyle", "style %s: %d\n", dialog->style_table[j].style_name, supported);
    }

    if (selected_item == NULL
	|| spSelectListItem(dialog->style_field, selected_item) < 0) {
	spSelectListIndex(dialog->style_field, 0);
    } 
	
    if (selected_item != NULL) {
	spDebug(50, "updateFontStyle", "selected_item = %s\n", selected_item);
	xfree(selected_item);
    }

    return;
}

static void updateFontSize(spFontSelectionDialog dialog, int family, unsigned long style)
{
    int j;
    spBool supported;
    char *selected_item;
    char buf[SP_MAX_FONTNAME];
    static int font_sizes[] = {
	8, 9, 10, 11, 12, 13, 14, 16, 18, 20, 22, 24, 26, 28, 32, 36, 40, 48, 56, 64, 72, 96, 144, 288, -1,
    };

    spDebug(50, "updateFontSize", "family = %d, style = %lx\n", family, style);
    
    selected_item = xspGetSelectedListItem(dialog->size_field);
    
    spClearList(dialog->size_field);
    spAddListItem(dialog->size_field, "*");

    for (j = 0; font_sizes[j] > 0; j++) {
	supported = spIsFontSizeSupportedArch(family, style, font_sizes[j]);
	if (supported == SP_TRUE) {
	    sprintf(buf, "%d", font_sizes[j]);
	    spAddListItem(dialog->size_field, buf);
	}
	spDebug(100, "updateFontSize", "size %d: %d\n", font_sizes[j], supported);
    }
    
    if (selected_item == NULL
	|| spSelectListItem(dialog->size_field, selected_item) < 0) {
	spSelectListIndex(dialog->size_field, 0);
    }
	
    if (selected_item != NULL) {
	spDebug(50, "updateFontSize", "selected_item = %s\n", selected_item);
	xfree(selected_item);
    }
    
    return;
}

static void changeFontStyle(spComponent component, spFontSelectionDialog dialog)
{
    int family = 0, size = 0;
    unsigned long style = 0;

    spDebug(50, "changeFontStyle", "dialog->lock = %d\n", dialog->lock);
    
    if (dialog->lock == SP_TRUE) return;
    
    updateFontName(dialog, &family, &style, &size);
    updateFontSize(dialog, family, style);
    spRedrawCanvas(dialog->sample_canvas);

    spDebug(50, "changeFontStyle", "family = %d, size = %d\n", family, size);
    
    return;
}

static void changeFontSize(spComponent component, spFontSelectionDialog dialog)
{
    int family = 0, size = 0;
    unsigned long style = 0;
    
    spDebug(50, "changeFontSize", "dialog->lock = %d\n", dialog->lock);
    
    if (dialog->lock == SP_TRUE) return;
    
    updateFontName(dialog, &family, &style, &size);
    spRedrawCanvas(dialog->sample_canvas);

    return;
}

static void selectFontFamily(spComponent component, spFontSelectionDialog dialog)
{
    int family;
	    
    if ((family = spGetSelectedListIndex(dialog->family_field)) >= 0
	&& spIsCreated(dialog->size_field) == SP_TRUE) {
	spDebug(50, "selectFontFamily", "family = %d\n", family);

	dialog->lock = SP_TRUE;

	updateFontStyle(dialog, family, 0);
	updateFontSize(dialog, family, 0);

	updateFontName(dialog, NULL, NULL, NULL);
	
	dialog->lock = SP_FALSE;
	
	spDebug(50, "selectFontFamily", "call spRedrawCanvas\n");
	spRedrawCanvas(dialog->sample_canvas);
    }
    
    return;
}

static void destroyFontSelectionDialog(spFontSelectionDialog dialog)
{
    if (dialog != NULL) {
	xfree(dialog->style_table);
	xfree(dialog);
    }
    
    return;
}

static void destroyFontSelectionDialogCB(void *data)
{
    spFontSelectionDialog *pdialog = (spFontSelectionDialog *)data;

    destroyFontSelectionDialog(*pdialog);
    *pdialog = NULL;

    return;
}

char *xspChooseFontDefault(spComponent component)
{
    const char **family_list;
    static spFontSelectionDialog dialog = NULL;

    if (dialog == NULL) {
	dialog = xalloc(1, struct _spFontSelectionDialog);
	/*dialog->component = component;*/
	dialog->family_field = NULL;
	dialog->style_field = NULL;
	dialog->size_field = NULL;
	dialog->lock = SP_FALSE;
	
	dialog->num_style_table = 12;
	dialog->style_table = NULL;
	
	dialog->style_table = xalloc(dialog->num_style_table, spFontSelectionStyleTable);
	dialog->style_table[0].style_name = SP_FONT_STYLE_MEDIUM_LABEL;
	dialog->style_table[0].style_mask = SP_FONT_STYLE_MEDIUM;
	dialog->style_table[1].style_name = SP_FONT_STYLE_ITALIC_LABEL;
	dialog->style_table[1].style_mask = SP_FONT_STYLE_ITALIC;
	dialog->style_table[2].style_name = SP_FONT_STYLE_OBLIQUE_LABEL;
	dialog->style_table[2].style_mask = SP_FONT_STYLE_OBLIQUE;
	dialog->style_table[3].style_name = SP_FONT_STYLE_BOLD_LABEL;
	dialog->style_table[3].style_mask = SP_FONT_STYLE_BOLD;
	dialog->style_table[4].style_name = SP_FONT_STYLE_BOLD_ITALIC_LABEL;
	dialog->style_table[4].style_mask = SP_FONT_STYLE_BOLD_ITALIC;
	dialog->style_table[5].style_name = SP_FONT_STYLE_BOLD_OBLIQUE_LABEL;
	dialog->style_table[5].style_mask = SP_FONT_STYLE_BOLD_OBLIQUE;
	dialog->style_table[6].style_name = SP_FONT_STYLE_LIGHT_LABEL;
	dialog->style_table[6].style_mask = SP_FONT_STYLE_LIGHT;
	dialog->style_table[7].style_name = SP_FONT_STYLE_LIGHT_ITALIC_LABEL;
	dialog->style_table[7].style_mask = SP_FONT_STYLE_LIGHT_ITALIC;
	dialog->style_table[8].style_name = SP_FONT_STYLE_LIGHT_OBLIQUE_LABEL;
	dialog->style_table[8].style_mask = SP_FONT_STYLE_LIGHT_OBLIQUE;
	dialog->style_table[9].style_name = SP_FONT_STYLE_BLACK_LABEL;
	dialog->style_table[9].style_mask = SP_FONT_STYLE_BLACK;
	dialog->style_table[10].style_name = SP_FONT_STYLE_BLACK_ITALIC_LABEL;
	dialog->style_table[10].style_mask = SP_FONT_STYLE_BLACK_ITALIC;
	dialog->style_table[11].style_name = SP_FONT_STYLE_BLACK_OBLIQUE_LABEL;
	dialog->style_table[11].style_mask = SP_FONT_STYLE_BLACK_OBLIQUE;
	
	dialog->graphics = spCreateGraphics("_FontSelectionGraphics",
					    SppLineWidth, 1,
					    NULL);
	
	dialog->dialog = spCreateDialogBox("_FontSelectionDialog",
					   SppTitle, SP_FONT_SELECTION_TITLE,
					   SppCloseStyle, SP_UNMAP_CLOSE,
					   SppCallbackFunc, popdownFontDialogCB,
					   SppCallbackData, dialog,
					   NULL);
	
	family_list = spGetFontFamilyListArch();

	dialog->family_field = spCreateParamField(dialog->dialog, "_FontSelectionFamilyField", 0,
						  SppTitle, SP_FONT_FAMILY_LABEL,
						  SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
						  SppFieldOffset, 120,
						  SppFieldSize, 180,
						  SppFieldStrings, family_list,
						  SppEditable, SP_FALSE,
						  SppCallbackFunc, selectFontFamily,
						  SppCallbackData, dialog,
						  SppSelectOnly, SP_TRUE,
						  NULL);
	spSelectListIndex(dialog->family_field, 0);
	
	dialog->style_field = spCreateParamField(dialog->dialog, "_FontSelectionStyleField", 0,
						 SppTitle, SP_FONT_STYLE_LABEL,
						 SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
						 SppFieldOffset, 120,
						 SppFieldSize, 180,
						 SppEditable, SP_FALSE,
						 SppCallbackFunc, changeFontStyle,
						 SppCallbackData, dialog,
						 SppSelectOnly, SP_TRUE,
						 NULL);
	dialog->size_field = spCreateParamField(dialog->dialog, "_FontSelectionSizeField", 0,
						SppTitle, SP_FONT_SIZE_LABEL,
						SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
						SppFieldOffset, 120,
						SppFieldSize, 180,
						SppEditable, SP_FALSE,
						SppCallbackFunc, changeFontSize,
						SppCallbackData, dialog,
						SppSelectOnly, SP_TRUE,
						NULL);

	dialog->sample_box = spCreateBox(dialog->dialog, "_FontSelectionSampleBox", 80,
					 SppTitle, SP_FONT_SAMPLE_LABEL,
					 SppSpacingOn, SP_TRUE,
					 SppBorderOn, SP_TRUE,
					 SppTitleOn, SP_TRUE,
					 NULL);
	dialog->sample_canvas = spCreateCanvas(dialog->sample_box, "_FontSelectionSampleCanvas",
					       300, 60,
					       SppSpacingOn, SP_TRUE,
					       /*SppBorderOn, SP_TRUE,*/
					       SppDrawBackground, SP_FALSE,
                                               SppFocusable, SP_FALSE,
					       SppCallbackFunc, drawFontSample,
					       SppCallbackData, dialog,
					       NULL);

	selectFontFamily(dialog->family_field, dialog);
	
	spAddExitCallback(destroyFontSelectionDialogCB, &dialog);
    }

    dialog->reason = SP_CR_NONE;
    strcpy(dialog->font_name, "");
    
    spPopupWindow(dialog->dialog);

    spDebug(50, "xspChooseFontDefault", "popup done: reason = %d\n", dialog->reason);
    
    if (dialog->reason == SP_CR_OK && !strnone(dialog->font_name)) {
	return strclone(dialog->font_name);
    }
    
    return NULL;
}

typedef struct _spColorSelectionDialog {
    spComponent component;
    spComponent dialog;
    
    spGraphics graphics;
    spGraphics cursor_graphics;

    spComponent canvas_box;
    int hs_cursor_x, hs_cursor_y;
    int hs_width, hs_height;
    spComponent hs_canvas;
    int l_cursor_y;
    int l_width, l_height;
    spComponent l_canvas;

    double h_value;
    double s_value;

    spComponent field_box;
    
    spComponent preview_canvas;
    
    spComponent hsl_box;
    spComponent h_field;
    spComponent s_field;
    spComponent l_field;
    
    spComponent rgb_box;
    spComponent r_field;
    spComponent g_field;
    spComponent b_field;

    spBool lock;
    spPixel pixel;
    spPixel prev_pixel;
    
    spCallbackReason reason;
} *spColorSelectionDialog;

static void popdownColorDialogCB(spComponent component, void *data)
{
    spColorSelectionDialog dialog = (spColorSelectionDialog)data;

    dialog->reason = spGetCallbackReason(component);
    
    spPopdownWindow(component);
    
    spDebug(50, "popdownColorDialogCB", "done\n");
    
    return;
}

#define convRGBToPixel(r, g, b) spRGB(spRound(2.55 * (r)), spRound(2.55 * (g)), spRound(2.55 * (b)))

static void getRGB(spPixel pixel, double *r, double *g, double *b)
{
    *r = (double)spGetRValue(pixel)/2.55;
    *g = (double)spGetGValue(pixel)/2.55;
    *b = (double)spGetBValue(pixel)/2.55;
    return;
}

static void getHSL(spPixel pixel, double *h, double *s, double *l)
{
    double r, g, b;
    
    getRGB(pixel, &r, &g, &b);
    spRgbToHsl(r/100.0, g/100.0, b/100.0, h, s, l);
    *s *= 100.0; *l *= 100.0;
    
    return;
}

static void convHSLToRGB(double h, double s, double l,
			 double *r, double *g, double *b)
{
    spHslToRgb(h, s/100.0, l/100.0, r, g, b);
    *r *= 100.0; *g *= 100.0; *b *= 100.0;
    
    return;
}

static spPixel convHSLToPixel(double h, double s, double l)
{
    double r, g, b;

    convHSLToRGB(h, s, l, &r, &g, &b);
    
    return convRGBToPixel(r, g, b);
}

static void updateHSL(spColorSelectionDialog dialog, double h, double s, double l)
{
    char buf[SP_MAX_LINE];
    
    sprintf(buf, "%.0f", h); spSetTextString(dialog->h_field, buf);
    sprintf(buf, "%.0f", s); spSetTextString(dialog->s_field, buf);
    sprintf(buf, "%.0f", l); spSetTextString(dialog->l_field, buf);
    dialog->h_value = h;
    dialog->s_value = s;

    return;
}

static void updateRGB(spColorSelectionDialog dialog, double r, double g, double b)
{
    char buf[SP_MAX_LINE];
    
    sprintf(buf, "%.0f", r); spSetTextString(dialog->r_field, buf);
    sprintf(buf, "%.0f", g); spSetTextString(dialog->g_field, buf);
    sprintf(buf, "%.0f", b); spSetTextString(dialog->b_field, buf);

    return;
}

static void drawHSCursor(spColorSelectionDialog dialog, int x, int y)
{
    spDrawLine(dialog->hs_canvas, dialog->cursor_graphics,
	       x-10, y, x-5, y);
    spDrawLine(dialog->hs_canvas, dialog->cursor_graphics,
	       x+5, y, x+10, y);
    spDrawLine(dialog->hs_canvas, dialog->cursor_graphics,
	       x, y-10, x, y-5);
    spDrawLine(dialog->hs_canvas, dialog->cursor_graphics,
	       x, y+5, x, y+10);
    
    return;
}

static void drawLCursor(spColorSelectionDialog dialog, int y)
{
    spDrawLine(dialog->l_canvas, dialog->cursor_graphics,
	       0, y, 5, y);
    spDrawLine(dialog->l_canvas, dialog->cursor_graphics,
	       dialog->l_width, y, dialog->l_width - 5, y);
    
    return;
}

static void updateHSCursor(spColorSelectionDialog dialog, int x, int y)
{
    if (dialog->hs_cursor_x >= 0) {
	spDebug(80, "updateHSCursor", "erase previous cursor: x = %d, y = %d\n",
		dialog->hs_cursor_x, dialog->hs_cursor_y);
	drawHSCursor(dialog, dialog->hs_cursor_x, dialog->hs_cursor_y);
    }
	
    dialog->hs_cursor_x = x;
    dialog->hs_cursor_y = y;
    spDebug(80, "updateHSCursor", "draw new cursor: x = %d, y = %d\n",
	    dialog->hs_cursor_x, dialog->hs_cursor_y);
    drawHSCursor(dialog, dialog->hs_cursor_x, dialog->hs_cursor_y);
    
    spRefreshCanvas(dialog->hs_canvas);
    
    return;
}

static void updateLCursor(spColorSelectionDialog dialog, int y)
{
    if (dialog->l_cursor_y >= 0) {
	drawLCursor(dialog, dialog->l_cursor_y);
    }
	
    dialog->l_cursor_y = y;
    drawLCursor(dialog, dialog->l_cursor_y);
	
    spRefreshCanvas(dialog->l_canvas);
    
    return;
}

static void updateCursor(spColorSelectionDialog dialog)
{
    int x, y;
    double h, s, l;

    getHSL(dialog->pixel, &h, &s, &l);
    
    x = (int)spRound(h * (double)dialog->hs_width / 360.0);
    y = (int)spRound((100.0 - s) * (double)dialog->hs_height / 100.0);
    updateHSCursor(dialog, x, y);
    
    y = (int)spRound((100.0 - l) * (double)dialog->l_height / 100.0);
    updateLCursor(dialog, y);
    
    return;
}

static void drawHSCanvasCB(spComponent canvas, void *data)
{
    spColorSelectionDialog dialog = (spColorSelectionDialog)data;
    
    spDebug(50, "drawHSCanvasCB", "dialog->lock = %d\n", dialog->lock);
    
    if (dialog->lock == SP_TRUE) return;
    
    dialog->lock = SP_TRUE;
    
    spDebug(80, "drawHSCanvasCB", "hs_width = %d, hs_height = %d\n",
	    dialog->hs_width, dialog->hs_height);
    
    if (dialog->hs_width <= 0) {
	int x, y;
	int px;
	double h, s, l;
	double hf, sf;
	spPixel pixel;
	spPixel prev_pixel;
	spCursor cursor;
	
	if ((cursor = spGetCursor(SP_CURSOR_WAIT)) != NULL) {
	    spSetCanvasCursor(canvas, cursor);
	}
	
	spGetSize(canvas, &dialog->hs_width, &dialog->hs_height);
	spDebug(50, "drawHSCanvasCB", "Size updated: hs_width = %d, hs_height = %d\n",
		dialog->hs_width, dialog->hs_height);

	l = 50.0;
	hf = 360.0 / (double)dialog->hs_width;
	sf = 100.0 / (double)dialog->hs_height;
	spDebug(80, "drawHSCanvasCB", "hf = %f, sf = %f\n", hf, sf);
	
	for (y = 0; y <= dialog->hs_height; y++) {
	    s = 100.0 - sf * (double)y;

	    px = 0;
	    prev_pixel = 0;
	    for (x = 1; x <= dialog->hs_width; x++) {
		h = hf * (double)x;
		/*spDebug(100, "drawHSCanvasCB", "y = %d, x = %d, s = %f, h = %f\n", y, x, s, h);*/
		
		pixel = convHSLToPixel(h, s, l);
		/*spDebug(100, "drawHSCanvasCB", "prev_pixel = %ld, pixel = %ld\n", prev_pixel, pixel);*/

#if 0
		spSetForegroundPixel(dialog->graphics, pixel);
		spDrawPoint(canvas, dialog->graphics, x, y);
#else
		if (pixel != prev_pixel || x == dialog->hs_width) {
		    /*spDebug(100, "drawHSCanvasCB",
		      "px = %d, x = %d, prev_pixel = %ld, pixel = %ld\n", px, x, prev_pixel, pixel);*/
		    spSetForegroundPixel(dialog->graphics, pixel);
		    spDrawLine(canvas, dialog->graphics, px, y, x, y);
		    prev_pixel = pixel;
		    px = x + 1;
		}
#endif
	    }
	}
	
	getHSL(dialog->pixel, &h, &s, &l);
	dialog->hs_cursor_x = (int)spRound(h * (double)dialog->hs_width / 360.0);
	dialog->hs_cursor_y = (int)spRound((100.0 - s) * (double)dialog->hs_height / 100.0);
	spDebug(50, "drawHSCanvasCB", "dialog->hs_cursor_x = %d, dialog->hs_cursor_y = %d\n",
		dialog->hs_cursor_x, dialog->hs_cursor_y);
    
	if (dialog->hs_cursor_x >= 0) {
	    drawHSCursor(dialog, dialog->hs_cursor_x, dialog->hs_cursor_y);
	}

	if (cursor != NULL) {
	    spUnsetCanvasCursor(canvas);
	    spDestroyCursor(cursor);
	}
    }
    
    dialog->lock = SP_FALSE;
	
    spDebug(50, "drawHSCanvasCB", "done\n");
    
    return;
}

static void drawLCanvasCB(spComponent canvas, void *data)
{
    int y;
    double h, s, l;
    double lf;
    spPixel pixel;
    spColorSelectionDialog dialog = (spColorSelectionDialog)data;
    
    spDebug(50, "drawLCanvasCB", "dialog->lock = %d\n", dialog->lock);
    
    if (dialog->lock == SP_TRUE) return;
    
    dialog->lock = SP_TRUE;
    
    getHSL(dialog->pixel, &h, &s, &l);
    
    if (dialog->l_width <= 0) {
	spGetSize(canvas, &dialog->l_width, &dialog->l_height);
	dialog->l_cursor_y = (int)spRound((100.0 - l) * (double)dialog->l_height / 100.0);
    }

    lf = 100.0 / (double)dialog->l_height;
    
    for (y = 0; y <= dialog->l_height; y++) {
	l = 100.0 - lf * (double)y;
	pixel = convHSLToPixel(h, s, l);
	spSetForegroundPixel(dialog->graphics, pixel);
	spDrawLine(canvas, dialog->graphics, 0, y, dialog->l_width, y);
    }

    if (dialog->l_cursor_y >= 0) {
	drawLCursor(dialog, dialog->l_cursor_y);
    }
    
    dialog->lock = SP_FALSE;
    
    spDebug(50, "drawLCanvasCB", "done\n");
    
    return;
}

static void drawPreviewCanvasCB(spComponent canvas, void *data)
{
    int width, height;
    spColorSelectionDialog dialog = (spColorSelectionDialog)data;

    spDebug(50, "drawPreviewCanvasCB", "dialog->lock = %d\n", dialog->lock);
    
    if (dialog->lock == SP_TRUE) return;

    if (spGetSize(canvas, &width, &height) == SP_TRUE) {
	spSetForegroundPixel(dialog->graphics, dialog->pixel);
	spFillRectangle(canvas, dialog->graphics, 0, 0, width, height/2);

	spSetForegroundPixel(dialog->graphics, dialog->prev_pixel);
	spFillRectangle(canvas, dialog->graphics, 0, height/2, width, height/2);
    }
    
    spDebug(50, "drawPreviewCanvasCB", "done\n");
    
    return;
}

static void changeColorCB(spComponent component, void *data)
{
    double r, g, b;
    double h, s, l;
    char *string;
    spColorSelectionDialog dialog = (spColorSelectionDialog)data;

    spDebug(50, "changeColorCB", "dialog->lock = %d, dialog->hs_width = %d, dialog->l_width = %d\n",
	    dialog->lock, dialog->hs_width, dialog->l_width);
    
    if (dialog->lock == SP_TRUE || dialog->hs_width <= 0 || dialog->l_width <= 0) return;

    dialog->lock = SP_TRUE;

    h = s = l = 0.0;
    if ((string = xspGetTextString(dialog->h_field)) != NULL) {
	h = atof(string);
	xfree(string);
    }
    if ((string = xspGetTextString(dialog->s_field)) != NULL) {
	s = atof(string);
	xfree(string);
    }
    if ((string = xspGetTextString(dialog->l_field)) != NULL) {
	l = atof(string);
	xfree(string);
    }

    h = MAX(MIN(h, 360.0), 0.0);
    s = MAX(MIN(s, 100.0), 0.0);
    l = MAX(MIN(l, 100.0), 0.0);
    
    spDebug(50, "changeColorCB", "h = %f, s = %f, l = %f\n", h, s, l);

    spHslToRgb(h, s/100.0, l/100.0, &r, &g, &b);
    r *= 100.0; g *= 100.0; b *= 100.0;
    
    updateRGB(dialog, r, g, b);
    dialog->pixel = convRGBToPixel(r, g, b);
    
    dialog->lock = SP_FALSE;

    updateCursor(dialog);
    
    spRedrawCanvas(dialog->preview_canvas);
    spRedrawCanvas(dialog->l_canvas);
    
    spDebug(50, "changeColorCB", "done\n");
    
    return;
}

static void changeRGBColorCB(spComponent component, void *data)
{
    double r, g, b;
    char *string;
    spColorSelectionDialog dialog = (spColorSelectionDialog)data;

    spDebug(50, "changeRGBColorCB", "dialog->lock = %d, dialog->hs_width = %d, dialog->l_width = %d\n",
	    dialog->lock, dialog->hs_width, dialog->l_width);
    
    if (dialog->lock == SP_TRUE || dialog->hs_width <= 0 || dialog->l_width <= 0) return;

    dialog->lock = SP_TRUE;

    r = g = b = 0.0;
    if ((string = xspGetTextString(dialog->r_field)) != NULL) {
	r = atof(string);
	xfree(string);
    }
    if ((string = xspGetTextString(dialog->g_field)) != NULL) {
	g = atof(string);
	xfree(string);
    }
    if ((string = xspGetTextString(dialog->b_field)) != NULL) {
	b = atof(string);
	xfree(string);
    }

    r = MAX(MIN(r, 100.0), 0.0);
    g = MAX(MIN(g, 100.0), 0.0);
    b = MAX(MIN(b, 100.0), 0.0);
    
    spDebug(50, "changeRGBColorCB", "r = %f, g = %f, b = %f\n", r, g, b);

    dialog->pixel = convRGBToPixel(r, g, b);
    
    dialog->lock = SP_FALSE;

    updateCursor(dialog);
    
    spRedrawCanvas(dialog->preview_canvas);
    spRedrawCanvas(dialog->l_canvas);
    
    spDebug(50, "changeRGBColorCB", "done\n");
    
    return;
}

static void HSCanvasMotionCB(spComponent component, void *data)
{
    int x, y;
    double h, s, l;
    double r, g, b;
    spColorSelectionDialog dialog = (spColorSelectionDialog)data;

    spDebug(50, "HSCanvasMotionCB", "dialog->lock = %d, dialog->hs_width = %d\n",
	    dialog->lock, dialog->hs_width);
    
    if (dialog->lock == SP_TRUE || dialog->hs_width <= 0) return;
    
    if (spGetCallbackMousePosition(component, &x, &y) == SP_TRUE) {
	dialog->lock = SP_TRUE;

	x = MAX(MIN(x, dialog->hs_width), 0);
	y = MAX(MIN(y, dialog->hs_height), 0);

	spDebug(50, "HSCanvasMotionCB", "x = %d, y = %d\n", x, y);
    
	getHSL(dialog->pixel, &h, &s, &l);
	
	h = 360.0 * (double)x / (double)dialog->hs_width;
	s = 100.0 * (1.0 - (double)y / (double)dialog->hs_height);

	updateHSL(dialog, h, s, l);

	convHSLToRGB(h, s, l, &r, &g, &b);
	updateRGB(dialog, r, g, b);
	dialog->pixel = convRGBToPixel(r, g, b);
	
	dialog->lock = SP_FALSE;
	
	spRedrawCanvas(dialog->preview_canvas);
	spRedrawCanvas(dialog->l_canvas);

	updateHSCursor(dialog, x, y);
    }
    
    spDebug(50, "HSCanvasMotionCB", "done\n");
    
    return;
}

static void LCanvasMotionCB(spComponent component, void *data)
{
    int x, y;
    double h, s, l;
    double r, g, b;
    spColorSelectionDialog dialog = (spColorSelectionDialog)data;

    spDebug(50, "LCanvasMotionCB", "dialog->lock = %d, dialog->l_width = %d\n",
	    dialog->lock, dialog->l_width);
    
    if (dialog->lock == SP_TRUE || dialog->l_width <= 0) return;
    
    if (spGetCallbackMousePosition(component, &x, &y) == SP_TRUE) {
	dialog->lock = SP_TRUE;

	x = MAX(MIN(x, dialog->l_width), 0);
	y = MAX(MIN(y, dialog->l_height), 0);
	
	getHSL(dialog->pixel, &h, &s, &l);
	h = dialog->h_value;
	s = dialog->s_value;

	l = 100.0 * (1.0 - (double)y / (double)dialog->l_height);
	
	updateHSL(dialog, h, s, l);

	convHSLToRGB(h, s, l, &r, &g, &b);
	updateRGB(dialog, r, g, b);
	dialog->pixel = convRGBToPixel(r, g, b);
	
	dialog->lock = SP_FALSE;
	
	spRedrawCanvas(dialog->preview_canvas);
	updateLCursor(dialog, y);
    }
    
    spDebug(50, "LCanvasMotionCB", "done\n");
    
    return;
}

static void destroyColorSelectionDialog(spColorSelectionDialog dialog)
{
    if (dialog != NULL) {
	xfree(dialog);
    }
    
    return;
}

static void destroyColorSelectionDialogCB(void *data)
{
    spColorSelectionDialog *pdialog = (spColorSelectionDialog *)data;

    destroyColorSelectionDialog(*pdialog);
    *pdialog = NULL;

    return;
}

#define SP_HS_CANVAS_WIDTH 280
#define SP_HS_CANVAS_HEIGHT 200
#define SP_L_CANVAS_WIDTH 20

#define SP_PREVIEW_CANVAS_WIDTH 40
#define SP_PREVIEW_CANVAS_HEIGHT 80

char *xspChooseColorDefault(spComponent component)
{
    double r, g, b;
    double h, s, l;
    static spColorSelectionDialog dialog = NULL;

    if (dialog == NULL) {
	dialog = xalloc(1, struct _spColorSelectionDialog);
	memset(dialog, 0, sizeof(struct _spColorSelectionDialog));
	dialog->lock = SP_FALSE;
	dialog->pixel = 0L;
	dialog->prev_pixel = 0L;
	dialog->h_value = 0.0;
	dialog->s_value = 0.0;
	/*dialog->component = component;*/
	
	dialog->graphics = spCreateGraphics("_ColorSelectionGraphics",
					    SppLineWidth, 1,
					    NULL);

	dialog->cursor_graphics = spCreateGraphics("_ColorSelectionCursorGraphics",
						   SppForeground, "max",
						   SppGraphicsMode, SP_GM_XOR,
						   SppLineWidth, 3,
						   NULL);
	
	dialog->dialog = spCreateDialogBox("_ColorSelectionDialog",
					   SppTitle, SP_COLOR_SELECTION_TITLE,
					   SppCloseStyle, SP_UNMAP_CLOSE,
					   SppCallbackFunc, popdownColorDialogCB,
					   SppCallbackData, dialog,
					   NULL);
	
	dialog->canvas_box = spCreateBox(dialog->dialog, "_ColorSelectionSampleBox", 210,
					 SppOrientation, SP_HORIZONTAL,
					 NULL);
	dialog->hs_cursor_x = dialog->hs_cursor_y = -1;
	dialog->hs_width = dialog->hs_height = 0;
	dialog->hs_canvas = spCreateCanvas(dialog->canvas_box, "_ColorSelectionHSCanvas",
					   SP_HS_CANVAS_WIDTH, SP_HS_CANVAS_HEIGHT,
					   SppWidth, SP_HS_CANVAS_WIDTH,
					   SppSpacingOn, SP_TRUE,
					   SppBorderOn, SP_TRUE,
                                           SppFocusable, SP_FALSE,
					   SppCallbackFunc, drawHSCanvasCB,
					   SppCallbackData, dialog,
					   NULL);
	spAddCallback(dialog->hs_canvas, SP_BUTTON_MOTION_CALLBACK
		      | SP_BUTTON_PRESS_CALLBACK | SP_BUTTON_RELEASE_CALLBACK,
		      HSCanvasMotionCB, dialog);
	
	dialog->l_cursor_y = -1;
	dialog->l_width = dialog->l_height = 0;
	dialog->l_canvas = spCreateCanvas(dialog->canvas_box, "_ColorSelectionLCanvas",
					  SP_L_CANVAS_WIDTH, SP_HS_CANVAS_HEIGHT,
					  SppWidth, SP_L_CANVAS_WIDTH,
					  SppSpacingOn, SP_TRUE,
					  SppBorderOn, SP_TRUE,
                                          SppFocusable, SP_FALSE,
					  SppCallbackFunc, drawLCanvasCB,
					  SppCallbackData, dialog,
					  NULL);
	spAddCallback(dialog->l_canvas, SP_BUTTON_MOTION_CALLBACK
		      | SP_BUTTON_PRESS_CALLBACK | SP_BUTTON_RELEASE_CALLBACK,
		      LCanvasMotionCB, dialog);
	
	dialog->field_box = spCreateBox(dialog->dialog, "_ColorSelectionFieldBox", 0,
					SppOrientation, SP_HORIZONTAL,
					NULL);
	dialog->preview_canvas = spCreateCanvas(dialog->field_box, "_ColorSelectionPreviewCanvas",
						SP_PREVIEW_CANVAS_WIDTH, SP_PREVIEW_CANVAS_HEIGHT,
						SppWidth, SP_PREVIEW_CANVAS_WIDTH,
						SppSpacingOn, SP_TRUE,
						SppBorderOn, SP_TRUE,
                                                SppFocusable, SP_FALSE,
						SppCallbackFunc, drawPreviewCanvasCB,
						SppCallbackData, dialog,
						NULL);
	
	dialog->hsl_box = spCreateBox(dialog->field_box, "_ColorSelectionHSLox", 130,
				      SppSpacingOn, SP_TRUE,
				      NULL);
	dialog->h_field = spCreateParamField(dialog->hsl_box, "_ColorSelectionHField", 0,
					     SppTitle, "Hue",
					     SppFieldOffset, 70,
					     SppFieldSize, 50,
					     SppSpacingOn, SP_FALSE,
					     SppEditable, SP_TRUE,
					     SppCallbackFunc, changeColorCB,
					     SppCallbackData, dialog,
					     NULL);
	dialog->s_field = spCreateParamField(dialog->hsl_box, "_ColorSelectionSField", 0,
					     SppTitle, "Saturation",
					     SppFieldOffset, 70,
					     SppFieldSize, 50,
					     SppSpacingOn, SP_FALSE,
					     SppEditable, SP_TRUE,
					     SppCallbackFunc, changeColorCB,
					     SppCallbackData, dialog,
					     NULL);
	dialog->l_field = spCreateParamField(dialog->hsl_box, "_ColorSelectionLField", 0,
					     SppTitle, "Lightness",
					     SppFieldOffset, 70,
					     SppFieldSize, 50,
					     SppSpacingOn, SP_FALSE,
					     SppEditable, SP_TRUE,
					     SppCallbackFunc, changeColorCB,
					     SppCallbackData, dialog,
					     NULL);
	
	dialog->rgb_box = spCreateBox(dialog->field_box, "_ColorSelectionHSLBox", 0,
				      SppSpacingOn, SP_TRUE,
				      NULL);
	dialog->r_field = spCreateParamField(dialog->rgb_box, "_ColorSelectionRField", 0,
					     SppTitle, "Red",
					     SppFieldOffset, 70,
					     SppFieldSize, 50,
					     SppSpacingOn, SP_FALSE,
					     SppEditable, SP_TRUE,
					     SppCallbackFunc, changeRGBColorCB,
					     SppCallbackData, dialog,
					     NULL);
	dialog->g_field = spCreateParamField(dialog->rgb_box, "_ColorSelectionGField", 0,
					     SppTitle, "Green",
					     SppFieldOffset, 70,
					     SppFieldSize, 50,
					     SppSpacingOn, SP_FALSE,
					     SppEditable, SP_TRUE,
					     SppCallbackFunc, changeRGBColorCB,
					     SppCallbackData, dialog,
					     NULL);
	dialog->b_field = spCreateParamField(dialog->rgb_box, "_ColorSelectionBField", 0,
					     SppTitle, "Blue",
					     SppFieldOffset, 70,
					     SppFieldSize, 50,
					     SppSpacingOn, SP_FALSE,
					     SppEditable, SP_TRUE,
					     SppCallbackFunc, changeRGBColorCB,
					     SppCallbackData, dialog,
					     NULL);
	
	spAddExitCallback(destroyColorSelectionDialogCB, &dialog);
    }

    if (!strnone(SpDialogPart(component).initial_color)) {
	dialog->pixel = spGetColorPixel(SpDialogPart(component).initial_color);
    }
    dialog->prev_pixel = dialog->pixel;
    dialog->reason = SP_CR_NONE;

    getRGB(dialog->pixel, &r, &g, &b);
    updateRGB(dialog, r, g, b);
    spDebug(50, "xspChooseColorDefault", "r = %f, g = %f, b = %f\n", r, g, b);

    getHSL(dialog->pixel, &h, &s, &l);
    updateHSL(dialog, h, s, l);
    spDebug(50, "xspChooseColorDefault", "h = %f, s = %f, l = %f\n", h, s, l);

    spPopupWindow(dialog->dialog);

    if (dialog->reason == SP_CR_OK) {
	char buf[SP_MAX_LINE];
	spGetColorName(dialog->pixel, buf);
	return strclone(buf);
    }
    
    return NULL;
}

typedef struct _spPageSetupDialog {
    spComponent dialog;
    
    spComponent main_box;
    
    spComponent paper_box;
    spComponent paper_type_field;
    spComponent paper_dimension_box;
    spComponent paper_size_field;
    spComponent paper_dimension_combo;

    spComponent bottom_box;
    
    spComponent orientation_box;
    spComponent orientation_portrait_button;
    spComponent orientation_landscape_button;
    
    spComponent margin_box;
#if defined(IPHONE) || defined(ANDROID)
    spComponent margin_left_box;
    spComponent margin_right_box;
    spComponent margin_top_box;
    spComponent margin_bottom_box;
#endif
    
    spComponent margin_horizontal_box;
    spComponent margin_left_field;
    spComponent margin_right_field;
    
    spComponent margin_vertical_box;
    spComponent margin_top_field;
    spComponent margin_bottom_field;
    
    spComponent preview_box;
    spComponent preview_canvas;
    spGraphics preview_graphics;
    
    spCallbackReason reason;

    spBool landscape_flag;
    spPaperSize paper;
    double left_margin_mm;
    double right_margin_mm;
    double top_margin_mm;
    double bottom_margin_mm;
    
    spBool inch_flag;
    spBool landscape_flag_temp;
    spPaperSize paper_temp;
    
    int num_paper_list;
    char **paper_list_strings;
    int paper_orig_index;

    char paper_size_string[SP_MAX_LINE];

    char left_margin_string[SP_MAX_LINE];
    char right_margin_string[SP_MAX_LINE];
    char top_margin_string[SP_MAX_LINE];
    char bottom_margin_string[SP_MAX_LINE];
    
} *spPageSetupDialog;

static char *sp_paper_dimension_strings[] = {
    "mm",
    "inch",
    NULL,
};

spPaperSize sp_paper_size_list[] = {
    SP_PAPER_SIZE_A0,
    SP_PAPER_SIZE_A1,
    SP_PAPER_SIZE_A2,
    SP_PAPER_SIZE_A3,
    SP_PAPER_SIZE_A4,
    SP_PAPER_SIZE_A5,
    SP_PAPER_SIZE_A6,
    SP_PAPER_SIZE_2A0,

    SP_PAPER_SIZE_JIS_B0,
    SP_PAPER_SIZE_JIS_B1,
    SP_PAPER_SIZE_JIS_B2,
    SP_PAPER_SIZE_JIS_B3,
    SP_PAPER_SIZE_JIS_B4,
    SP_PAPER_SIZE_JIS_B5,
    SP_PAPER_SIZE_JIS_B6,

    SP_PAPER_SIZE_LETTER,
    SP_PAPER_SIZE_GOVERNMENT_LETTER,
    SP_PAPER_SIZE_LEGAL,
    SP_PAPER_SIZE_TABLOID,
    SP_PAPER_SIZE_STATEMENT,
    SP_PAPER_SIZE_QUARTO,
    SP_PAPER_SIZE_FOOLSCAP,
    SP_PAPER_SIZE_EXECUTIVE,
    SP_PAPER_SIZE_HAGAKI,
};

#if defined(IPHONE) || defined(ANDROID)
#define SP_PAGE_SETUP_PAPER_FIELD_OFFSET 70
#define SP_PAGE_SETUP_PAPER_FIELD_SIZE 105
#define SP_PAGE_SETUP_PAPER_DIMENSION_FIELD_SIZE 83
#define SP_PAGE_SETUP_DEFAULT_CANVAS_WIDTH 100
#define SP_PAGE_SETUP_DEFAULT_CANVAS_HEIGHT 100
#define SP_PAGE_SETUP_MARGIN_BOX_WIDTH 130
#define SP_PAGE_SETUP_MARGIN_LABEL_SIZE 50
#define SP_PAGE_SETUP_MARGIN_FIELD_SIZE 60
#define SP_PAGE_SETUP_DEFAULT_MARGIN 5
#define SP_PAGE_SETUP_DEFAULT_SPACING 5
#else
#define SP_PAGE_SETUP_PAPER_FIELD_OFFSET 50
#define SP_PAGE_SETUP_PAPER_FIELD_SIZE 170
#define SP_PAGE_SETUP_PAPER_DIMENSION_FIELD_SIZE 70
#define SP_PAGE_SETUP_DEFAULT_CANVAS_WIDTH 150
#define SP_PAGE_SETUP_DEFAULT_CANVAS_HEIGHT 150
#define SP_PAGE_SETUP_MARGIN_LABEL_SIZE 50
#define SP_PAGE_SETUP_MARGIN_FIELD_SIZE 50
#endif
#define SP_PAGE_SETUP_ORIENTATION_BOX_WIDTH 100
#define SP_PAGE_SETUP_PREVIEW_PAPER_PERCENTAGE 85
#define SP_PAGE_SETUP_PREVIEW_SHADOW_OFFSET_PERCENTAGE 4
#define SP_PAGE_SETUP_PREVIEW_SHADOW_COLOR "#aca899"
#define SP_PAGE_SETUP_PREVIEW_PAPER_BRIGHT_LINE "#808080"
#define SP_PAGE_SETUP_PREVIEW_PAPER_DARK_LINE "black"

static double getCurrentDoubleValue(spComponent component)
{
    const char *str;
    
    if ((str = spGetTextString(component)) != NULL) {
	return MAX(atof(str), 0.0);
    } else {
	return 0.0;
    }
}

static void getCurrentMarginInPoint(spPageSetupDialog dialog,
				    double factor, double paper_width_mm, double paper_height_mm,
				    int *left_margin_pt, int *right_margin_pt,
				    int *top_margin_pt, int *bottom_margin_pt)
{
    double left_margin_mm, top_margin_mm;
    double right_margin_mm, bottom_margin_mm;
    
    left_margin_mm = getCurrentDoubleValue(dialog->margin_left_field);
    if (dialog->inch_flag == SP_TRUE) left_margin_mm *= 25.4;
    *left_margin_pt = (int)spRound(left_margin_mm * factor);

    right_margin_mm = getCurrentDoubleValue(dialog->margin_right_field);
    if (dialog->inch_flag == SP_TRUE) right_margin_mm *= 25.4;
    *right_margin_pt = (int)spRound(right_margin_mm * factor);
    
    top_margin_mm = getCurrentDoubleValue(dialog->margin_top_field);
    if (dialog->inch_flag == SP_TRUE) top_margin_mm *= 25.4;
    *top_margin_pt = (int)spRound(top_margin_mm * factor);

    bottom_margin_mm = getCurrentDoubleValue(dialog->margin_bottom_field);
    if (dialog->inch_flag == SP_TRUE) bottom_margin_mm *= 25.4;
    *bottom_margin_pt = (int)spRound(bottom_margin_mm * factor);
    
    return;
}

static spBool getPaperDimensions(spPageSetupDialog dialog, double *paper_width_mm, double *paper_height_mm)
{
    spPaperOrientation orient;
    
    if (dialog->landscape_flag_temp == SP_TRUE) {
	orient = SP_PAPER_ORIENTATION_LANDSCAPE;
    } else {
	orient = SP_PAPER_ORIENTATION_PORTRAIT;
    }
    
    return spGetPaperDimensions(dialog->paper_temp, orient, paper_width_mm, paper_height_mm);
}

static void drawPagePreview(spComponent component, spPageSetupDialog dialog)
{
    int width, height;
    double paper_width_mm, paper_height_mm;
    int paper_width_pt, paper_height_pt;
    int left_offset_pt, top_offset_pt;
    int shadow_left_offset_pt, shadow_top_offset_pt;
    double width_factor, height_factor, factor;
    int left_margin_pt, right_margin_pt;
    int top_margin_pt, bottom_margin_pt;
    int client_width_pt, client_height_pt;
    
    if (spGetClientSize(component, &width, &height) == SP_TRUE) {
#if defined(IPHONE)
        spSetGraphicsParams(dialog->preview_graphics, SppForeground, "white", NULL);
        spFillRectangle(component, dialog->preview_graphics, 0, 0, width, height);
#endif
	if (getPaperDimensions(dialog, &paper_width_mm, &paper_height_mm) == SP_TRUE) {
	    paper_width_pt = (SP_PAGE_SETUP_PREVIEW_PAPER_PERCENTAGE * width) / 100;
	    paper_height_pt = (SP_PAGE_SETUP_PREVIEW_PAPER_PERCENTAGE * height) / 100;
	    width_factor = (double)paper_width_pt / MAX(paper_width_mm, paper_height_mm);
	    height_factor = (double)paper_height_pt / MAX(paper_height_mm, paper_width_mm);
	    factor = MIN(width_factor, height_factor);

	    paper_width_pt = (int)spRound(paper_width_mm * factor);
	    paper_height_pt = (int)spRound(paper_height_mm * factor);
	    
	    left_offset_pt = (width - paper_width_pt) / 2;
	    top_offset_pt = (height - paper_height_pt) / 2;
	    shadow_left_offset_pt = (SP_PAGE_SETUP_PREVIEW_SHADOW_OFFSET_PERCENTAGE * width) / 100;
	    shadow_top_offset_pt = shadow_left_offset_pt;
	    
	    spSetGraphicsParams(dialog->preview_graphics,
				SppForeground, SP_PAGE_SETUP_PREVIEW_SHADOW_COLOR,
				SppLineType, SP_LINE_SOLID,
				NULL);
	    spFillRectangle(component, dialog->preview_graphics,
			    left_offset_pt + shadow_left_offset_pt, top_offset_pt + shadow_top_offset_pt,
			    paper_width_pt, paper_height_pt);

	    spSetGraphicsParams(dialog->preview_graphics, SppForeground, "white", NULL);
	    spFillRectangle(component, dialog->preview_graphics, left_offset_pt, top_offset_pt,
			    paper_width_pt, paper_height_pt);

	    spSetGraphicsParams(dialog->preview_graphics, SppForeground, SP_PAGE_SETUP_PREVIEW_PAPER_BRIGHT_LINE, NULL);
	    spDrawLine(component, dialog->preview_graphics, left_offset_pt + paper_width_pt, top_offset_pt,
		       left_offset_pt, top_offset_pt);
	    spDrawLine(component, dialog->preview_graphics, left_offset_pt, top_offset_pt,
		       left_offset_pt, top_offset_pt + paper_height_pt);
	    
	    spSetGraphicsParams(dialog->preview_graphics, SppForeground, SP_PAGE_SETUP_PREVIEW_PAPER_DARK_LINE, NULL);
	    spDrawLine(component, dialog->preview_graphics, left_offset_pt, top_offset_pt + paper_height_pt,
		       left_offset_pt + paper_width_pt, top_offset_pt + paper_height_pt);
	    spDrawLine(component, dialog->preview_graphics, left_offset_pt + paper_width_pt, top_offset_pt + paper_height_pt,
		       left_offset_pt + paper_width_pt, top_offset_pt);

	    getCurrentMarginInPoint(dialog, factor, paper_width_mm, paper_height_mm,
				    &left_margin_pt, &right_margin_pt, &top_margin_pt, &bottom_margin_pt);
	    client_width_pt = paper_width_pt - left_margin_pt - right_margin_pt;
	    client_height_pt = paper_height_pt - top_margin_pt - bottom_margin_pt;
	    spDebug(80, "drawPagePreview", "client_width_pt = %d, client_height_pt = %d\n",
		    client_width_pt, client_height_pt);

	    if (client_width_pt > 0 && client_height_pt > 0) {
		spSetGraphicsParams(dialog->preview_graphics,
				    SppForeground, SP_PAGE_SETUP_PREVIEW_PAPER_BRIGHT_LINE,
				    SppLineType, SP_LINE_DASH,
				    NULL);
		spDrawRectangle(component, dialog->preview_graphics,
				left_offset_pt + left_margin_pt, top_offset_pt + top_margin_pt,
				client_width_pt, client_height_pt);
	    }
	}
    }

    return;
}

static void setDimensionString(double value, char *out_str)
{
    double value_r, value_r10;

    value_r = spRound(value);
    value_r10 = spRound(10.0 * value);
    
    if (FABS(value - value_r) < 0.1) {
	sprintf(out_str, "%.0f", value_r);
    } else if (FABS(10.0 * value - value_r10) < 0.1) {
	sprintf(out_str, "%.1f", value_r10 / 10.0);
    } else {
	sprintf(out_str, "%.2f", spRound(100.0 * value) / 100.0);
    }
    
    return;
}

static void convertDimension(spPageSetupDialog dialog, double value, char *out_str, spBool inch_flag)
{
    if (inch_flag == SP_TRUE) {
	value /= 25.4;
    } else {
	value *= 25.4;
    }
    
    setDimensionString(value, out_str);
    
    return;
}

static void convertFieldDimension(spPageSetupDialog dialog, spComponent component, spBool inch_flag)
{
    double value;
    char buf[SP_MAX_LINE];

    value = getCurrentDoubleValue(component);
    convertDimension(dialog, value, buf, inch_flag);
    spSetTextString(component, buf);

    return;
}

static void createPaperListStrings(spPageSetupDialog dialog)
{
    int i;

    dialog->num_paper_list = spArraySize(sp_paper_size_list);
    dialog->paper_list_strings = xalloc(dialog->num_paper_list + 1, char *);
    dialog->paper_orig_index = 0;
    
    for (i = 0; i < dialog->num_paper_list; i++) {
	dialog->paper_list_strings[i] = xalloc(SP_MAX_LINE, char);
	spGetPaperSizeString(sp_paper_size_list[i], 0L, dialog->paper_list_strings[i], SP_MAX_LINE);
	if (sp_paper_size_list[i] == dialog->paper) {
	    dialog->paper_orig_index = i;
	    spGetPaperDimensionString(dialog->paper, SP_PAPER_SIZE_STRING_NO_UNIT,
				      dialog->paper_size_string, sizeof(dialog->paper_size_string));
	}
    }
    dialog->paper_list_strings[i] = NULL;
    spDebug(80, "createPaperListStrings", "paper_orig_index = %d\n", dialog->paper_orig_index);
    
    return;
}

static int getPaperIndex(spPageSetupDialog dialog)
{
    int i;
    int index;

    index = dialog->paper_orig_index;
    
    for (i = 0; i < dialog->num_paper_list; i++) {
	if (sp_paper_size_list[i] == dialog->paper_temp) {
	    index = i;
	}
    }

    return index;
}

static void updatePaperSize(spPageSetupDialog dialog)
{
    unsigned long options = SP_PAPER_SIZE_STRING_NO_UNIT;

    spDebug(80, "updatePaperSize", "inch_flag = %d\n", dialog->inch_flag);
    
    if (dialog->inch_flag == SP_TRUE) {
	options |= SP_PAPER_SIZE_STRING_INCH;
    }
    spGetPaperDimensionString(dialog->paper_temp, options,
			      dialog->paper_size_string, sizeof(dialog->paper_size_string));
    spDebug(80, "updatePaperSize", "paper_size_string = %s\n", dialog->paper_size_string);
	
    spSetTextString(dialog->paper_size_field, dialog->paper_size_string);

    spDebug(80, "updatePaperSize", "done\n");
    
    return;
}

static void changePaperTypeCB(spComponent component, spPageSetupDialog dialog)
{
    int index;

    if ((index = spGetSelectedListIndex(component)) >= 0) {
	spDebug(80, "changePaperTypeCB", "index = %d\n", index);
	dialog->paper_temp = sp_paper_size_list[index];
	updatePaperSize(dialog);
	spRedrawCanvas(dialog->preview_canvas);
    }
    
    spDebug(80, "changePaperTypeCB", "done\n");
    
    return;
}

static void changeDimensionCB(spComponent component, spPageSetupDialog dialog)
{
    int index;
    spBool updated = SP_FALSE;

    if ((index = spGetSelectedListIndex(component)) >= 0) {
	spDebug(80, "changeDimensionCB", "index = %d\n", index);
	if (index == 1) {
	    if (dialog->inch_flag != SP_TRUE) {
		dialog->inch_flag = SP_TRUE;
		updated = SP_TRUE;
	    }
	} else {
	    if (dialog->inch_flag != SP_FALSE) {
		dialog->inch_flag = SP_FALSE;
		updated = SP_TRUE;
	    }
	}

	if (updated == SP_TRUE) {
	    convertFieldDimension(dialog, dialog->margin_left_field, dialog->inch_flag);
	    convertFieldDimension(dialog, dialog->margin_right_field, dialog->inch_flag);
	    convertFieldDimension(dialog, dialog->margin_top_field, dialog->inch_flag);
	    convertFieldDimension(dialog, dialog->margin_bottom_field, dialog->inch_flag);
	    
	    updatePaperSize(dialog);
	}
    }
    
    spDebug(80, "changeDimensionCB", "done\n");
    
    return;
}

static void checkOrientationButtonCB(spComponent component, spPageSetupDialog dialog)
{
    spBool flag;
    
    if (spGetToggleState(component, &flag) == SP_TRUE) {
	spDebug(80, "checkOrientationButtonCB", "%s: flag = %d\n", SpGetName(component), flag);
	if (streq(SpGetName(component), "_PageOrientationLandscapeButton")) {
	    dialog->landscape_flag_temp = flag;
	} else {
	    dialog->landscape_flag_temp = spIsFalse(flag);
	}
	
	spRedrawCanvas(dialog->preview_canvas);
    }
    
    spDebug(80, "checkOrientationButtonCB", "landscape_flag_temp = %d\n", dialog->landscape_flag_temp);
    
    return;
}

static void changeMarginValueCB(spComponent component, spPageSetupDialog dialog)
{
    spDebug(80, "changeMarginValueCB", "in\n");
    spRedrawCanvas(dialog->preview_canvas);
    spDebug(80, "changeMarginValueCB", "done\n");
    
    return;
}

static void popdownPageSetupDialogCB(spComponent component, spPageSetupDialog dialog)
{
    double paper_width_mm, paper_height_mm;
    double left_margin_mm, right_margin_mm, top_margin_mm, bottom_margin_mm;
    
    dialog->reason = spGetCallbackReason(component);

    if (dialog->reason == SP_CR_OK) {
	if (getPaperDimensions(dialog, &paper_width_mm, &paper_height_mm) == SP_TRUE) {
	    left_margin_mm = getCurrentDoubleValue(dialog->margin_left_field);
	    if (dialog->inch_flag == SP_TRUE) left_margin_mm *= 25.4;
	    right_margin_mm = getCurrentDoubleValue(dialog->margin_right_field);
	    if (dialog->inch_flag == SP_TRUE) right_margin_mm *= 25.4;
	    top_margin_mm = getCurrentDoubleValue(dialog->margin_top_field);
	    if (dialog->inch_flag == SP_TRUE) top_margin_mm *= 25.4;
	    bottom_margin_mm = getCurrentDoubleValue(dialog->margin_bottom_field);
	    if (dialog->inch_flag == SP_TRUE) bottom_margin_mm *= 25.4;

	    if (paper_width_mm - left_margin_mm - right_margin_mm <= 0.0
		|| paper_height_mm - top_margin_mm - bottom_margin_mm <= 0) {
		spDisplayWarning(component, SP_PAGE_SETUP_MARGIN_WARNING_TITLE,
				 SP_PAGE_SETUP_MARGIN_WARNING_MESSAGE);
		return;
	    }

	    dialog->left_margin_mm = left_margin_mm;
	    dialog->right_margin_mm = right_margin_mm;
	    dialog->top_margin_mm = top_margin_mm;
	    dialog->bottom_margin_mm = bottom_margin_mm;
	    dialog->paper = dialog->paper_temp;
	    dialog->landscape_flag = dialog->landscape_flag_temp;
	}
    }

    spPopdownWindow(component);
    
    return;
}

static void destroyPageSetupDialog(spPageSetupDialog dialog)
{
    int i;
    
    if (dialog != NULL) {
	if (dialog->paper_list_strings != NULL) {
	    for (i = 0; dialog->paper_list_strings[i] != NULL; i++) {
		xfree(dialog->paper_list_strings[i]);
	    }
	}
	xfree(dialog);
    }
    
    return;
}

static void destroyPageSetupDialogCB(void *data)
{
    spPageSetupDialog *pdialog = (spPageSetupDialog *)data;

    destroyPageSetupDialog(*pdialog);
    *pdialog = NULL;

    return;
}

spBool spGetPageSetupDefault(spComponent component, spPaperSize *paper_size, spPaperOrientation *paper_orientation,
			     double *left_margin_mm, double *right_margin_mm, double *top_margin_mm, double *bottom_margin_mm)
{
    static spPageSetupDialog dialog = NULL;
    int default_field_height;
    int box_title_margin;
    spBool create_dialog = SP_FALSE;

    spDebug(100, "spGetPageSetupDefault", "in\n");
    
    if (dialog == NULL) {
	dialog = xalloc(1, struct _spPageSetupDialog);
	dialog->reason = SP_CR_NONE;
	dialog->inch_flag = SP_FALSE;
	dialog->paper = SP_PAPER_SIZE_A4;
	dialog->landscape_flag = SP_FALSE;
	dialog->left_margin_mm = SP_PAGE_SETUP_DEFAULT_HORIZONTAL_MARGIN_MM;
	dialog->right_margin_mm = SP_PAGE_SETUP_DEFAULT_HORIZONTAL_MARGIN_MM;
	dialog->top_margin_mm = SP_PAGE_SETUP_DEFAULT_VERTICAL_MARGIN_MM;
	dialog->bottom_margin_mm = SP_PAGE_SETUP_DEFAULT_VERTICAL_MARGIN_MM;
	dialog->paper_list_strings = NULL;
	
	spAddExitCallback(destroyPageSetupDialogCB, &dialog);
	
	create_dialog = SP_TRUE;
    }
    
    if (paper_size != NULL) {
	dialog->paper = *paper_size;
    }
    dialog->paper_temp = dialog->paper;
    
    if (paper_orientation != NULL) {
	dialog->landscape_flag = spIsPaperOrientationHorizontal(*paper_orientation);
    }
    dialog->landscape_flag_temp = dialog->landscape_flag;

    if (left_margin_mm != NULL) dialog->left_margin_mm = *left_margin_mm;
    if (right_margin_mm != NULL) dialog->right_margin_mm = *right_margin_mm;
    if (top_margin_mm != NULL) dialog->top_margin_mm = *top_margin_mm;
    if (bottom_margin_mm != NULL) dialog->bottom_margin_mm = *bottom_margin_mm;
    
    setDimensionString(dialog->left_margin_mm, dialog->left_margin_string);
    setDimensionString(dialog->right_margin_mm, dialog->right_margin_string);
    setDimensionString(dialog->top_margin_mm, dialog->top_margin_string);
    setDimensionString(dialog->bottom_margin_mm, dialog->bottom_margin_string);

    if (create_dialog == SP_TRUE) {
	createPaperListStrings(dialog);

#if defined(IPHONE) || defined(ANDROID)
	box_title_margin = 0;
	default_field_height = spGetTextFieldDefaultHeight(SP_FALSE);
        default_field_height += SP_PAGE_SETUP_DEFAULT_MARGIN * 2;
#else
	default_field_height = spGetTextFieldDefaultHeight(SP_TRUE);
	box_title_margin = 1 * default_field_height;
#endif

	dialog->dialog = spCreateDialogBox("_PageSetupDialog",
					   SppTitle, SP_PAGE_SETUP_TITLE,
					   SppCloseStyle, SP_UNMAP_CLOSE,
					   SppCallbackFunc, popdownPageSetupDialogCB,
					   SppCallbackData, dialog,
#if defined(IPHONE) || defined(ANDROID)
                                           SppSpacingOn, SP_FALSE,
					   SppOrientation, SP_VERTICAL,
#else
					   SppOrientation, SP_HORIZONTAL,
#endif
					   NULL);

#if defined(IPHONE) || defined(ANDROID)
        dialog->main_box = dialog->dialog;
#else
	/* main container for paper, orientation, and margin areas */
	dialog->main_box = spCreateBox(dialog->dialog, "_PageSetupMainBox",
                                       -SP_PAGE_SETUP_DEFAULT_CANVAS_WIDTH - default_field_height,
				       SppSpacingOn, SP_FALSE,
				       NULL);
#endif

	/* paper area */
	dialog->paper_box = spCreateBox(dialog->main_box, "_PageSetupPaperBox", default_field_height * 2 + box_title_margin,
#if defined(IPHONE) || defined(ANDROID)
                                        SppMarginWidth, SP_PAGE_SETUP_DEFAULT_MARGIN,
                                        SppMarginHeight, SP_PAGE_SETUP_DEFAULT_MARGIN,
                                        SppSpacing, SP_PAGE_SETUP_DEFAULT_SPACING,
#else
					SppTitle, SP_PAGE_SETUP_PAPER_TITLE,
					SppSpacingOn, SP_TRUE,
					SppBorderOn, SP_TRUE,
					SppTitleOn, SP_TRUE,
#endif
					SppWidth, -1,
					NULL);
	dialog->paper_type_field = spCreateParamField(dialog->paper_box, "_PaperTypeField",
                                                      default_field_height,
						      SppTitle, SP_PAGE_SETUP_PAPER_TYPE,
						      SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
						      SppFieldOffset, SP_PAGE_SETUP_PAPER_FIELD_OFFSET,
						      SppFieldSize, SP_PAGE_SETUP_PAPER_FIELD_SIZE,
						      SppEditable, SP_FALSE,
						      SppCallbackFunc, changePaperTypeCB,
						      SppCallbackData, dialog,
						      SppListStrings, dialog->paper_list_strings,
						      SppSelectedListIndex, dialog->paper_orig_index,
#if defined(IPHONE) || defined(ANDROID)
                                                      SppSpacing, SP_FALSE,
                                                      SppMarginWidth, SP_PAGE_SETUP_DEFAULT_MARGIN,
                                                      SppMarginHeight, SP_PAGE_SETUP_DEFAULT_MARGIN,
#endif
						      NULL);
	dialog->paper_dimension_box = spCreateBox(dialog->paper_box, "_PaperDimensionBox", default_field_height,
						  SppOrientation, SP_HORIZONTAL,
#if defined(IPHONE) || defined(ANDROID)
                                                  SppMarginWidth, SP_PAGE_SETUP_DEFAULT_MARGIN,
                                                  SppMarginHeight, SP_PAGE_SETUP_DEFAULT_MARGIN,
#endif
                                                  SppSpacingOn, SP_FALSE,
						  NULL);
	dialog->paper_size_field = spCreateParamField(dialog->paper_dimension_box, "_PaperSizeField", 5,
						      SppTitle, SP_PAGE_SETUP_PAPER_SIZE,
						      SppFieldOffset, SP_PAGE_SETUP_PAPER_FIELD_OFFSET,
						      SppFieldSize, SP_PAGE_SETUP_PAPER_FIELD_SIZE,
						      SppEditable, SP_FALSE,
						      SppTextString, dialog->paper_size_string,
                                                      SppHeight, default_field_height,
#if defined(IPHONE) || defined(ANDROID)
                                                      SppSpacing, SP_FALSE,
                                                      SppMarginWidth, SP_PAGE_SETUP_DEFAULT_MARGIN,
                                                      SppMarginHeight, SP_PAGE_SETUP_DEFAULT_MARGIN,
#endif
						      NULL);
	dialog->paper_dimension_combo = spCreateComboBox(dialog->paper_dimension_box, "_PaperDimensionCombo",
							 SppWidth, SP_PAGE_SETUP_PAPER_DIMENSION_FIELD_SIZE,
                                                         SppHeight, spGetTextFieldDefaultHeight(SP_FALSE),
							 SppListStrings, sp_paper_dimension_strings,
							 SppSelectedListIndex, 0,
							 SppEditable, SP_FALSE,
							 SppCallbackFunc, changeDimensionCB,
							 SppCallbackData, dialog,
							 NULL);

	/* container for orientation and margin areas */
	dialog->bottom_box = spCreateBox(dialog->main_box, "_PageSetupBottomBox", 0,
					 SppOrientation, SP_HORIZONTAL,
					 SppSpacingOn, SP_FALSE,
#if defined(IPHONE) || defined(ANDROID)
                                         SppMarginWidth, SP_PAGE_SETUP_DEFAULT_MARGIN,
                                         SppMarginHeight, SP_PAGE_SETUP_DEFAULT_MARGIN,
#endif
					 NULL);

#if !(defined(IPHONE) || defined(ANDROID))
	/* orientation area */
	dialog->orientation_box = spCreateBox(dialog->bottom_box, "_PageSetupOrientationBox", SP_PAGE_SETUP_ORIENTATION_BOX_WIDTH,
					      SppTitle, SP_PAGE_SETUP_ORIENTATION_TITLE,
					      SppSpacingOn, SP_TRUE,
					      SppBorderOn, SP_TRUE,
					      SppTitleOn, SP_TRUE,
					      NULL);
	dialog->orientation_portrait_button = spCreateRadioButton(dialog->orientation_box, "_PageOrientationPortraitButton",
								  SppTitle, SP_PAGE_SETUP_ORIENTATION_PORTRAIT,
								  SppCallbackFunc, checkOrientationButtonCB,
								  SppCallbackData, dialog,
								  SppSet, spIsFalse(dialog->landscape_flag),
								  NULL);
	dialog->orientation_landscape_button = spCreateRadioButton(dialog->orientation_box, "_PageOrientationLandscapeButton",
								   SppTitle, SP_PAGE_SETUP_ORIENTATION_LANDSCAPE,
								   SppCallbackFunc, checkOrientationButtonCB,
								   SppCallbackData, dialog,
								   SppSet, dialog->landscape_flag,
								   NULL);
#endif

	/* margin area */
	dialog->margin_box = spCreateBox(dialog->bottom_box, "_PageSetupMarginBox",
#if defined(IPHONE) || defined(ANDROID)
                                         SP_PAGE_SETUP_MARGIN_BOX_WIDTH,
					 SppSpacingOn, SP_FALSE,
                                         SppMarginWidth, SP_PAGE_SETUP_DEFAULT_MARGIN,
                                         SppMarginHeight, SP_PAGE_SETUP_DEFAULT_MARGIN,
#else
                                         0,
					 SppTitle, SP_PAGE_SETUP_MARGIN_TITLE,
					 SppSpacingOn, SP_TRUE,
					 SppBorderOn, SP_TRUE,
					 SppTitleOn, SP_TRUE,
#endif
					 NULL);

#if defined(IPHONE) || defined(ANDROID)
        dialog->orientation_box = NULL;
        dialog->orientation_portrait_button = NULL;
	dialog->orientation_landscape_button = spCreateCheckBox(dialog->margin_box, "_PageOrientationLandscapeButton",
								SppTitle, SP_PAGE_SETUP_ORIENTATION_LANDSCAPE,
								SppCallbackFunc, checkOrientationButtonCB,
								SppCallbackData, dialog,
								SppSet, dialog->landscape_flag,
								NULL);
#endif
        
#if defined(IPHONE) || defined(ANDROID)
	dialog->margin_left_box = spCreateBox(dialog->margin_box, "_PageMarginLeftBox", default_field_height,
                                              SppOrientation, SP_HORIZONTAL,
                                              SppSpacingOn, SP_FALSE,
                                              SppMarginWidth, SP_PAGE_SETUP_DEFAULT_MARGIN,
                                              SppMarginHeight, SP_PAGE_SETUP_DEFAULT_MARGIN,
                                              NULL);
        dialog->margin_horizontal_box = dialog->margin_left_box;
#else
	dialog->margin_horizontal_box = spCreateBox(dialog->margin_box, "_PageMarginHorizontalBox", default_field_height,
						    SppOrientation, SP_HORIZONTAL,
						    NULL);
#endif
        
	spCreateLabel(dialog->margin_horizontal_box, "_PageMarginLeftLabel",
		      SppWidth, SP_PAGE_SETUP_MARGIN_LABEL_SIZE,
		      SppHeight, spGetTextFieldDefaultHeight(SP_FALSE),
		      SppAlignment, SP_ALIGNMENT_BEGINNING,
		      SppTitle, SP_PAGE_SETUP_MARGIN_LEFT_LABEL,
		      NULL);
	dialog->margin_left_field = spCreateTextField(dialog->margin_horizontal_box, "_PageMarginLeftField",
						      SppWidth, SP_PAGE_SETUP_MARGIN_FIELD_SIZE,
						      SppHeight, spGetTextFieldDefaultHeight(SP_FALSE),
                                                      SppKeyboardType, SP_KEYBOARD_TYPE_NUMBER_PAD,
						      SppTextString, dialog->left_margin_string,
						      SppCallbackFunc, changeMarginValueCB,
						      SppCallbackData, dialog,
						      NULL);
        
#if defined(IPHONE) || defined(ANDROID)
	dialog->margin_right_box = spCreateBox(dialog->margin_box, "_PageMarginRightBox", default_field_height,
                                               SppOrientation, SP_HORIZONTAL,
                                               SppSpacingOn, SP_FALSE,
                                               SppMarginWidth, SP_PAGE_SETUP_DEFAULT_MARGIN,
                                               SppMarginHeight, SP_PAGE_SETUP_DEFAULT_MARGIN,
                                               NULL);
        dialog->margin_horizontal_box = dialog->margin_right_box;
#endif
        
	spCreateLabel(dialog->margin_horizontal_box, "_PageMarginRightLabel",
		      SppWidth, SP_PAGE_SETUP_MARGIN_LABEL_SIZE,
                      SppHeight, spGetTextFieldDefaultHeight(SP_FALSE),
		      SppAlignment, SP_ALIGNMENT_BEGINNING,
		      SppTitle, SP_PAGE_SETUP_MARGIN_RIGHT_LABEL,
		      NULL);
	dialog->margin_right_field = spCreateTextField(dialog->margin_horizontal_box, "_PageMarginRightField",
						       SppWidth, SP_PAGE_SETUP_MARGIN_FIELD_SIZE,
						       SppHeight, spGetTextFieldDefaultHeight(SP_FALSE),
                                                       SppKeyboardType, SP_KEYBOARD_TYPE_NUMBER_PAD,
						       SppTextString, dialog->right_margin_string,
						       SppCallbackFunc, changeMarginValueCB,
						       SppCallbackData, dialog,
						       NULL);
	
#if defined(IPHONE) || defined(ANDROID)
	dialog->margin_top_box = spCreateBox(dialog->margin_box, "_PageMarginTopBox", default_field_height,
                                             SppOrientation, SP_HORIZONTAL,
                                             SppSpacingOn, SP_FALSE,
                                             SppMarginWidth, SP_PAGE_SETUP_DEFAULT_MARGIN,
                                             SppMarginHeight, SP_PAGE_SETUP_DEFAULT_MARGIN,
                                             NULL);
        dialog->margin_vertical_box = dialog->margin_top_box;
#else
	dialog->margin_vertical_box = spCreateBox(dialog->margin_box, "_PageMarginHorizontalBox", default_field_height,
						  SppOrientation, SP_HORIZONTAL,
						  NULL);
#endif
        
	spCreateLabel(dialog->margin_vertical_box, "_PageMarginTopLabel",
		      SppWidth, SP_PAGE_SETUP_MARGIN_LABEL_SIZE,
                      SppHeight, spGetTextFieldDefaultHeight(SP_FALSE),
		      SppAlignment, SP_ALIGNMENT_BEGINNING,
		      SppTitle, SP_PAGE_SETUP_MARGIN_TOP_LABEL,
		      NULL);
	dialog->margin_top_field = spCreateTextField(dialog->margin_vertical_box, "_PageMarginTopField",
						     SppWidth, SP_PAGE_SETUP_MARGIN_FIELD_SIZE,
						     SppHeight, spGetTextFieldDefaultHeight(SP_FALSE),
                                                     SppKeyboardType, SP_KEYBOARD_TYPE_NUMBER_PAD,
						     SppTextString, dialog->top_margin_string,
						     SppCallbackFunc, changeMarginValueCB,
						     SppCallbackData, dialog,
						     NULL);

#if defined(IPHONE) || defined(ANDROID)
	dialog->margin_bottom_box = spCreateBox(dialog->margin_box, "_PageMarginBottomBox", default_field_height,
                                                SppOrientation, SP_HORIZONTAL,
                                                SppSpacingOn, SP_FALSE,
                                                SppMarginWidth, SP_PAGE_SETUP_DEFAULT_MARGIN,
                                                SppMarginHeight, SP_PAGE_SETUP_DEFAULT_MARGIN,
                                                NULL);
        dialog->margin_vertical_box = dialog->margin_bottom_box;
#endif
        
	spCreateLabel(dialog->margin_vertical_box, "_PageMarginBottomLabel",
		      SppWidth, SP_PAGE_SETUP_MARGIN_LABEL_SIZE,
		      SppHeight, spGetTextFieldDefaultHeight(SP_FALSE),
		      SppAlignment, SP_ALIGNMENT_BEGINNING,
		      SppTitle, SP_PAGE_SETUP_MARGIN_BOTTOM_LABEL,
		      NULL);
	dialog->margin_bottom_field = spCreateTextField(dialog->margin_vertical_box, "_PageMarginBottomField",
							SppWidth, SP_PAGE_SETUP_MARGIN_FIELD_SIZE,
							SppHeight, spGetTextFieldDefaultHeight(SP_FALSE),
                                                        SppKeyboardType, SP_KEYBOARD_TYPE_NUMBER_PAD,
							SppTextString, dialog->bottom_margin_string,
							SppCallbackFunc, changeMarginValueCB,
							SppCallbackData, dialog,
							NULL);
	
#if defined(IPHONE) || defined(ANDROID)
        dialog->preview_box = dialog->bottom_box;
#else
	/* preview area */
	dialog->preview_box = spCreateBox(dialog->dialog, "_PageSetupPreviewBox", 0,
					  SppTitle, SP_PAGE_SETUP_PREVIEW_TITLE,
					  SppSpacingOn, SP_TRUE,
					  SppBorderOn, SP_TRUE,
					  SppTitleOn, SP_TRUE,
					  NULL);
#endif
                                         
	dialog->preview_graphics = spCreateGraphics("_PageSetupPreviewGraphics",
						    SppForeground, "white",
						    NULL);
	dialog->preview_canvas = spCreateCanvas(dialog->preview_box, "_PageSetupPreviewCanvas",
						SP_PAGE_SETUP_DEFAULT_CANVAS_WIDTH, SP_PAGE_SETUP_DEFAULT_CANVAS_HEIGHT,
						SppBorderOn, SP_FALSE,
#if defined(IPHONE) || defined(ANDROID)
						SppSpacingOn, SP_TRUE,
#else
						SppSpacingOn, SP_TRUE,
#endif
						SppDrawBackground, SP_TRUE,
                                                SppFocusable, SP_FALSE,
						SppCallbackFunc, drawPagePreview,
						SppCallbackData, dialog,
						NULL);
    } else {
	spSetTextString(dialog->margin_left_field, dialog->left_margin_string);
	spSetTextString(dialog->margin_right_field, dialog->right_margin_string);
	spSetTextString(dialog->margin_top_field, dialog->top_margin_string);
	spSetTextString(dialog->margin_bottom_field, dialog->bottom_margin_string);
        if (dialog->orientation_portrait_button != NULL) {
            spSetToggleState(dialog->orientation_portrait_button, spIsFalse(dialog->landscape_flag));
        }
	spSetToggleState(dialog->orientation_landscape_button, dialog->landscape_flag);
	
	spSelectListIndex(dialog->paper_type_field, getPaperIndex(dialog));
	updatePaperSize(dialog);
	spRedrawCanvas(dialog->preview_canvas);
    }
    
    dialog->reason = SP_CR_NONE;
    spPopupWindow(dialog->dialog);
    spDebug(100, "spGetPageSetupDefault", "reason = %d\n", dialog->reason);

    if (dialog->reason == SP_CR_OK) {
	if (paper_orientation != NULL) {
            spDebug(80, "spGetPageSetupDefault", "dialog->landscape_flag = %d\n", dialog->landscape_flag);
	    if (dialog->landscape_flag == SP_TRUE) {
		*paper_orientation = SP_PAPER_ORIENTATION_LANDSCAPE;
	    } else {
		*paper_orientation = SP_PAPER_ORIENTATION_PORTRAIT;
	    }
	}
	if (paper_size != NULL) *paper_size = dialog->paper;
	if (left_margin_mm != NULL) *left_margin_mm = dialog->left_margin_mm;
	if (right_margin_mm != NULL) *right_margin_mm = dialog->right_margin_mm;
	if (top_margin_mm != NULL) *top_margin_mm = dialog->top_margin_mm;
	if (bottom_margin_mm != NULL) *bottom_margin_mm = dialog->bottom_margin_mm;
    
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

#if 1
typedef struct _spPrinterDialog {
    spComponent dialog;
    
    spComponent printer_box;
    spComponent printer_name_field;
    
    spComponent print_setting_box;
    
    spComponent print_range_box;
    spComponent print_range_all_button;
    spComponent print_range_selection_button;
    spComponent print_range_current_button;
    
    spComponent print_range_pages_box;
    spComponent print_range_pages_button;
    spComponent print_range_from_label;
    spComponent print_range_from_field;
    spComponent print_range_to_label;
    spComponent print_range_to_field;
    
    spComponent print_copies_box;
    spComponent print_copies_field;
    spComponent print_collate_checkbox;
    spComponent print_double_sided_checkbox;
    spComponent print_double_sided_tumble_checkbox;

    int print_range_type;
    int max_page;
    int current_page;
    
    spCallbackReason reason;
    
} *spPrinterDialog;

#define SP_PRINTER_NAME_FIELD_OFFSET /*60*/80
#define SP_PRINTER_NAME_FIELD_SIZE 180
#define SP_PRINT_RANGE_BOX_WIDTH 160
#define SP_PRINT_RANGE_BOX_PERCENTAGE_WIDTH 63
#define SP_PRINT_RANGE_PAGES_BUTTON_SIZE 90
#define SP_PRINT_RANGE_LABEL_SIZE 30
#define SP_PRINT_RANGE_FIELD_SIZE 40
#define SP_PRINT_COPIES_FIELD_OFFSET 100
#define SP_PRINT_COPIES_FIELD_SIZE 40

static void editNumCopiesCB(spComponent component, spPrinterDialog dialog)
{
    int num_copies;
    char *string;

    if ((string = xspGetTextString(component)) != NULL) {
	num_copies = atoi(string);

	if (num_copies > 1) {
	    spSetSensitive(dialog->print_collate_checkbox, SP_TRUE);
	} else {
	    spSetSensitive(dialog->print_collate_checkbox, SP_FALSE);
	}
	
	xfree(string);
    }

    return;
}

static void checkDoubleSidedCB(spComponent component, spPrinterDialog dialog)
{
    spBool flag;

    if (spGetToggleState(component, &flag) == SP_TRUE) {
	if (flag == SP_TRUE) {
	    spSetSensitive(dialog->print_double_sided_tumble_checkbox, SP_TRUE);
	} else {
	    spSetSensitive(dialog->print_double_sided_tumble_checkbox, SP_FALSE);
	}
    }
    
    return;
}

static void setPrintRangeSensitive(spPrinterDialog dialog, spBool sensitive)
{
    spSetSensitive(dialog->print_range_from_label, sensitive);
    spSetSensitive(dialog->print_range_from_field, sensitive);
    spSetSensitive(dialog->print_range_to_label, sensitive);
    spSetSensitive(dialog->print_range_to_field, sensitive);

    return;
}

static void checkPageRangeRadioCB(spComponent component, spPrinterDialog dialog)
{
    if (streq(SpGetName(component), "_PrintRangePagesButton")) {
	setPrintRangeSensitive(dialog, SP_TRUE);
	dialog->print_range_type = 0;
    } else {
	setPrintRangeSensitive(dialog, SP_FALSE);
	if (streq(SpGetName(component), "_PrintRangeSelectionButton")) {
	    dialog->print_range_type = SP_PAGE_RANGE_SELECTION;
	} else if (streq(SpGetName(component), "_PrintRangeCurrentButton")) {
	    dialog->print_range_type = SP_PAGE_RANGE_CURRENT;
	} else {
	    dialog->print_range_type = SP_PAGE_RANGE_ALL;
	}
    }
    
    return;
}

static void getRangeValue(spPrinterDialog dialog, int *start, int *end)
{
    char *string;
    
    if ((string = xspGetTextString(dialog->print_range_from_field)) != NULL) {
        *start = atoi(string);
        xfree(string);
    }
    if ((string = xspGetTextString(dialog->print_range_to_field)) != NULL) {
        *end = atoi(string);
        xfree(string);
    }

    return;
}

static void popdownPrinterDialogCB(spComponent component, spPrinterDialog dialog)
{
    int start, end;
    
    dialog->reason = spGetCallbackReason(component);
    spDebug(80, "popdownPrinterDialogCB", "reason = %d\n", dialog->reason);

    if (dialog->reason == SP_CR_OK) {
	start = end = 0;
	getRangeValue(dialog, &start, &end);
	if (start > end || (dialog->max_page >= 1 && end > dialog->max_page)) {
	    spDisplayWarning(component, SP_PRINT_WARNING_TITLE, SP_PRINT_RANGE_WARNING_MESSAGE);
	    return;
	}
    }

    spPopdownWindow(component);

    return;
}

static void destroyPrinterDialog(spPrinterDialog dialog)
{
    if (dialog != NULL) {
	xfree(dialog);
    }
    
    return;
}

static void destroyPrinterDialogCB(void *data)
{
    spPrinterDialog *pdialog = (spPrinterDialog *)data;

    destroyPrinterDialog(*pdialog);
    *pdialog = NULL;

    return;
}

char *xspOpenPrinterDialogDefault(int *start_page, int *end_page, int *num_copies, spBool *collate,
				  spBool *double_sided, spBool *tumble, spArg *args, int num_arg)
{
    static spPrinterDialog dialog = NULL;
    int i;
    int default_field_height;

    if (dialog == NULL) {
	dialog = xalloc(1, struct _spPrinterDialog);
	dialog->reason = SP_CR_NONE;
	dialog->max_page = 0;
	dialog->current_page = 0;
	dialog->print_range_type = SP_PAGE_RANGE_ALL;
	
	default_field_height = spGetTextFieldDefaultHeight(SP_TRUE);
	
	dialog->dialog = spCreateDialogBox("_PrinterDialog",
					   SppTitle, SP_PRINTER_DIALOG_TITLE,
					   SppCloseStyle, SP_UNMAP_CLOSE,
					   SppCallbackFunc, popdownPrinterDialogCB,
					   SppCallbackData, dialog,
					   SppOrientation, SP_VERTICAL,
					   NULL);
	dialog->printer_box = spCreateBox(dialog->dialog, "_PrinterBox", default_field_height,
					  SppTitle, SP_PRINTER_BOX_TITLE,
					  SppSpacingOn, SP_TRUE,
					  SppBorderOn, SP_TRUE,
					  SppTitleOn, SP_TRUE,
					  SppWidth, -1,
					  NULL);
	dialog->printer_name_field = spCreateParamField(dialog->printer_box, "_PrinterNameField", default_field_height,
							SppTitle, SP_PRINTER_NAME_LABEL,
							SppFieldOffset, SP_PRINTER_NAME_FIELD_OFFSET,
							SppFieldSize, SP_PRINTER_NAME_FIELD_SIZE,
							SppEditable, SP_TRUE,
							NULL);
	
	dialog->print_setting_box = spCreateBox(dialog->dialog, "_PrintSettingBox", 0,
						SppOrientation, SP_HORIZONTAL,
						SppSpacingOn, SP_FALSE,
						NULL);
	dialog->print_range_box = spCreateBox(dialog->print_setting_box, "_PrintRangeBox", /*SP_PRINT_RANGE_BOX_WIDTH*/0,
					      SppTitle, SP_PRINT_RANGE_BOX_TITLE,
					      SppSpacingOn, SP_TRUE,
					      SppBorderOn, SP_TRUE,
					      SppTitleOn, SP_TRUE,
					      SppInitialWidth, SP_PRINT_RANGE_BOX_WIDTH,
					      SppPercentageWidth, SP_PRINT_RANGE_BOX_PERCENTAGE_WIDTH,
					      NULL);
	dialog->print_range_all_button = spCreateRadioButton(dialog->print_range_box, "_PrintRangeAllButton",
							     SppTitle, SP_PRINT_RANGE_ALL_LABEL,
							     SppSet, SP_TRUE,
							     SppSpacingOn, SP_TRUE,
							     SppCallbackFunc, checkPageRangeRadioCB,
							     SppCallbackData, dialog,
							     NULL);
	dialog->print_range_selection_button = spCreateRadioButton(dialog->print_range_box, "_PrintRangeSelectionButton",
								   SppTitle, SP_PRINT_RANGE_SELECTION_LABEL,
								   SppSet, SP_FALSE,
								   SppSpacingOn, SP_TRUE,
								   SppCallbackFunc, checkPageRangeRadioCB,
								   SppCallbackData, dialog,
								   NULL);
	spSetSensitive(dialog->print_range_selection_button, SP_FALSE);
	dialog->print_range_current_button = spCreateRadioButton(dialog->print_range_box, "_PrintRangeCurrentButton",
								 SppTitle, SP_PRINT_RANGE_CURRENT_LABEL,
								 SppSet, SP_FALSE,
								 SppSpacingOn, SP_TRUE,
								 SppCallbackFunc, checkPageRangeRadioCB,
								 SppCallbackData, dialog,
								 NULL);
	spSetSensitive(dialog->print_range_current_button, SP_FALSE);
	dialog->print_range_pages_box = spCreateBox(dialog->print_range_box, "_PrintRangePagesBox", 0,
						    SppOrientation, SP_HORIZONTAL,
						    /*SppSpacingOn, SP_FALSE,*/
						    SppSpacingOn, SP_TRUE,
						    SppMarginHeight, 0,
						    NULL);
	dialog->print_range_pages_button = spCreateRadioButton(dialog->print_range_pages_box, "_PrintRangePagesButton",
							       SppTitle, SP_PRINT_RANGE_PAGES_LABEL,
							       SppSet, SP_FALSE,
							       SppWidth, SP_PRINT_RANGE_PAGES_BUTTON_SIZE,
							       SppHeight, spGetTextFieldDefaultHeight(SP_FALSE),
							       SppRadioGroupParent, dialog->print_range_box,
							       SppCallbackFunc, checkPageRangeRadioCB,
							       SppCallbackData, dialog,
							       NULL);
	spSetSensitive(dialog->print_range_pages_button, SP_FALSE);
	dialog->print_range_from_label = spCreateLabel(dialog->print_range_pages_box, "_PrintRangeFromLabel",
						       SppWidth, SP_PRINT_RANGE_LABEL_SIZE,
						       SppAlignment, SP_ALIGNMENT_BEGINNING,
						       SppTitle, SP_PRINT_RANGE_FROM_LABEL,
						       SppHeight, spGetTextFieldDefaultHeight(SP_FALSE),
						       NULL);
	dialog->print_range_from_field = spCreateTextField(dialog->print_range_pages_box, "_PrintRangeFromField",
							   SppWidth, SP_PRINT_RANGE_FIELD_SIZE,
							   SppHeight, spGetTextFieldDefaultHeight(SP_FALSE),
							   NULL);
	dialog->print_range_to_label = spCreateLabel(dialog->print_range_pages_box, "_PrintRangeToLabel",
						     SppWidth, SP_PRINT_RANGE_LABEL_SIZE,
						     SppAlignment, SP_ALIGNMENT_END,
						     SppTitle, SP_PRINT_RANGE_TO_LABEL,
						     SppHeight, spGetTextFieldDefaultHeight(SP_FALSE),
						     NULL);
	dialog->print_range_to_field = spCreateTextField(dialog->print_range_pages_box, "_PrintRangeToField",
							 SppWidth, SP_PRINT_RANGE_FIELD_SIZE,
							 SppHeight, spGetTextFieldDefaultHeight(SP_FALSE),
							 NULL);

	dialog->print_copies_box = spCreateBox(dialog->print_setting_box, "_PrintCopiesBox", 0,
					       SppTitle, SP_PRINT_COPIES_BOX_TITLE,
					       SppSpacingOn, SP_TRUE,
					       SppBorderOn, SP_TRUE,
					       SppTitleOn, SP_TRUE,
					       NULL);
	dialog->print_copies_field = spCreateParamField(dialog->print_copies_box, "_PrintCopiesField", default_field_height,
							SppTitle, SP_PRINT_COPIES_FIELD_LABEL,
							SppFieldOffset, SP_PRINT_COPIES_FIELD_OFFSET,
							SppFieldSize, SP_PRINT_COPIES_FIELD_SIZE,
							SppEditable, SP_TRUE,
							SppTextString, "1",
							SppCallbackFunc, editNumCopiesCB,
							SppCallbackData, dialog,
							NULL);
	dialog->print_collate_checkbox = spCreateCheckBox(dialog->print_copies_box, "_PrintCollateCheckBox",
							  SppTitle, SP_PRINT_COLLATE_LABEL,
							  SppSet, SP_TRUE,
							  NULL);
	spSetSensitive(dialog->print_collate_checkbox, SP_FALSE);

	dialog->print_double_sided_checkbox = spCreateCheckBox(dialog->print_copies_box, "_PrintDoubleSidedCheckBox",
							       SppTitle, SP_PRINT_DOUBLE_SIDED_LABEL,
							       SppSet, SP_FALSE,
							       
							       SppCallbackFunc, checkDoubleSidedCB,
							       SppCallbackData, dialog,
							       NULL);
	dialog->print_double_sided_tumble_checkbox = spCreateCheckBox(dialog->print_copies_box, "_PrintDoubleSidedTumbleCheckBox",
								      SppTitle, SP_PRINT_DOUBLE_SIDED_TUMBLE_LABEL,
								      SppSet, SP_FALSE,
								      NULL);
	spSetSensitive(dialog->print_double_sided_tumble_checkbox, SP_FALSE);
	
	setPrintRangeSensitive(dialog, SP_FALSE);

	spAddExitCallback(destroyPrinterDialogCB, &dialog);
    }
    
    if (start_page == NULL) {
	spSetSensitive(dialog->print_range_box, SP_FALSE);
    } else {
	spSetSensitive(dialog->print_range_box, SP_TRUE);
    }
    if (num_copies == NULL) {
	spSetSensitive(dialog->print_copies_field, SP_FALSE);
    } else {
	spSetSensitive(dialog->print_copies_field, SP_TRUE);
    }
    if (collate == NULL) {
	spSetSensitive(dialog->print_collate_checkbox, SP_FALSE);
    }
    if (double_sided == NULL) {
	spSetSensitive(dialog->print_double_sided_checkbox, SP_FALSE);
	spSetSensitive(dialog->print_double_sided_tumble_checkbox, SP_FALSE);
    } else if (tumble == NULL) {
	spSetSensitive(dialog->print_double_sided_tumble_checkbox, SP_FALSE);
    }

    dialog->current_page = 0;

    for (i = 0; i < num_arg; i++) {
	if (spEqParam(args[i].param, SppEnablePageRange)) {
	    if ((spBool)args[i].value == SP_TRUE) {
		spSetSensitive(dialog->print_range_pages_button, SP_TRUE);
	    }
	} else if (spEqParam(args[i].param, SppEnablePageRangeSelection)) {
	    if ((spBool)args[i].value == SP_TRUE) {
		spSetSensitive(dialog->print_range_selection_button, SP_TRUE);
	    }
	} else if (spEqParam(args[i].param, SppEnablePageRangeCurrent)) {
	    if ((spBool)args[i].value == SP_TRUE) {
		spSetSensitive(dialog->print_range_current_button, SP_TRUE);
	    }
	} else if (spEqParam(args[i].param, SppMaxPageRange)) {
	    dialog->max_page = (int)args[i].value;
	} else if (spEqParam(args[i].param, SppCurrentPage)) {
	    dialog->current_page = (int)args[i].value;
	}
    }
    
    dialog->reason = SP_CR_NONE;
    spPopupWindow(dialog->dialog);

    if (dialog->reason == SP_CR_OK) {
	int start = 0, end = 0;
	char *string;
	
	spDebug(80, "xspOpenPrinterDialogDefault", "print_range_type = %d\n", dialog->print_range_type);
	if (dialog->print_range_type >= 0) {
	    getRangeValue(dialog, &start, &end);
	} else if (dialog->print_range_type == SP_PAGE_RANGE_CURRENT && dialog->current_page >= 1) {
	    start = end = dialog->current_page;
	} else {
	    start = end = dialog->print_range_type;
	}
	spDebug(80, "xspOpenPrinterDialogDefault", "start = %d, end = %d\n", start, end);

	if (start_page != NULL) *start_page = start;
	if (end_page != NULL) *end_page = end;

	if (num_copies != NULL) {
	    if ((string = xspGetTextString(dialog->print_copies_field)) != NULL) {
		*num_copies = MAX(atoi(string), 0);
		xfree(string);
	    }
	}
	if (collate != NULL) {
	    spGetToggleState(dialog->print_collate_checkbox, collate);
	}

	if (double_sided != NULL) {
	    spGetToggleState(dialog->print_double_sided_checkbox, double_sided);
	    if (tumble != NULL) {
		spGetToggleState(dialog->print_double_sided_tumble_checkbox, tumble);
	    }
	}
	
	return xspGetTextString(dialog->printer_name_field);
    }
    
    return NULL;
}
#endif
