/*
 *	spDraw_Win.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spPrimitiveP.h>
#include <sp/spCanvasP.h>
#include <sp/spDrawP.h>
#include <sp/spGraphicsP.h>

extern spTopLevel sp_toplevel;

spBool spIsAlphaSupported(spComponent component)
{
    return SP_FALSE;
}

spBool spIsDrawable(spComponent component)
{
    if (spIsPrimitive(component) == SP_TRUE
	&& SpPrimitiveArch(component).memdc != NULL) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

void *spGetImageHandleArch(spComponent component)
{
    return SpPrimitiveArch(component).memdc;
}

spBool spCopyImageArch(spComponent src, spComponent dest,
		       int src_x, int src_y, int width, int height,
		       int dest_x, int dest_y)
{
    if (BitBlt(SpPrimitiveArch(dest).memdc, dest_x, dest_y,
	       MIN(SpComponentPart(src).client_width - src_x, width),
	       MIN(SpComponentPart(src).client_height - src_y, height),
	       SpPrimitiveArch(src).memdc, src_x, src_y, SRCCOPY)) {
	return SP_TRUE;
    } else {
	spDebug(50, "spCopyImageArch", "BitBlt failed\n");
	return SP_FALSE;
    }
}

spBool spStretchImageArch(spComponent src, spComponent dest,
			  int src_x, int src_y, int src_width, int src_height,
			  int dest_x, int dest_y, int dest_width, int dest_height)
{
    BOOL ret;
    int oldMode;
    POINT oldPt;
    BOOL setBrushRet;

    oldMode = SetStretchBltMode(SpPrimitiveArch(dest).memdc, COLORONCOLOR/*HALFTONE*/);
    //setBrushRet = SetBrushOrgEx(SpPrimitiveArch(dest).memdc, 0, 0, &oldPt);
    
    ret = StretchBlt(SpPrimitiveArch(dest).memdc,
		     dest_x, dest_y, dest_width, dest_height,
		     SpPrimitiveArch(src).memdc,
		     src_x, src_y, 
		     MIN(SpComponentPart(src).client_width - src_x, src_width),
		     MIN(SpComponentPart(src).client_height - src_y, src_height),
		     SRCCOPY);

    //if (setBrushRet) SetBrushOrgEx(SpPrimitiveArch(dest).memdc, oldPt.x, oldPt.y, NULL);
    if (oldMode) SetStretchBltMode(SpPrimitiveArch(dest).memdc, oldMode);
    
    if (ret) {
	return SP_TRUE;
    } else {
	spDebug(50, "spStretchImageArch", "StretchBlt failed\n");
	return SP_FALSE;
    }
}

void spDrawRectangleWin(HDC hdc, spGraphics graphics,
			spGraphicsType type, HPEN null_pen, HBRUSH null_brush,
			int x, int y, int width, int height)
{
    int prev_mode;
    int prev_bkmode;
    int ux, uy;
    HPEN hprevpen;
    HBRUSH hprevbrush;
    
    if (x <= 0) {
	ux = -1;
    } else {
	ux = x;
    }
    if (y <= 0) {
	uy = -1;
    } else {
	uy = y;
    }
    
    if (type == SP_GRAPHICS_BRUSH) {
	hprevpen = SelectObject(hdc, null_pen);
	hprevbrush = SelectObject(hdc, SpGraphicsArch(graphics).brush);
    } else {
	hprevbrush = SelectObject(hdc, null_brush);
	hprevpen = SelectObject(hdc, SpGraphicsArch(graphics).pen);
    }
    prev_mode = SetROP2(hdc, SpGraphicsArch(graphics).mode);
    prev_bkmode = SetBkMode(hdc, TRANSPARENT);
    Rectangle(hdc, ux, uy, x + width + 1, y + height + 1);
    SelectObject(hdc, /*null_pen*/hprevpen);
    SelectObject(hdc, /*null_brush*/hprevbrush);
    SetROP2(hdc, prev_mode);
    SetBkMode(hdc, prev_bkmode);

    return;
}

void spFillRectangleArch(spComponent component, spGraphics graphics,
			 int x, int y, int width, int height)
{
    if (graphics == NULL) graphics = SpTopLevelPart(sp_toplevel).graphics;

    spDrawRectangleWin(SpPrimitiveArch(component).memdc, graphics, SP_GRAPHICS_BRUSH,
		       SpTopLevelArch(sp_toplevel).null_pen, SpTopLevelArch(sp_toplevel).null_brush,
		       x, y, width, height);
    
    return;
}

void spDrawRectangleArch(spComponent component, spGraphics graphics,
			 int x, int y, int width, int height)
{
    if (graphics == NULL) graphics = SpTopLevelPart(sp_toplevel).graphics;

    spDrawRectangleWin(SpPrimitiveArch(component).memdc, graphics, SP_GRAPHICS_PEN,
		       SpTopLevelArch(sp_toplevel).null_pen, SpTopLevelArch(sp_toplevel).null_brush,
		       x, y, width, height);
    
    return;
}

static void convArcRadianToCoord(int x, int y, int width, int height, double radian, int *ox, int *oy)
{
    double w2, h2;
    double value;
    double t0, t1;

    w2 = (double)width / 2.0;
    h2 = (double)height / 2.0;
    t0 = h2 / w2;
	
    value = radian;
    t1 = tan(value);
    
    if (t1 >= t0) {		/* top-right or bottom-left */
	if (cos(value) >= 0.0) { /* top-right */
	    *ox = x + (int)spRound(w2 + w2 / t1);
	    *oy = y;
	} else {		/* bottom-left */
	    *ox = x + (int)spRound(w2 - w2 / t1);
	    *oy = y + height;
	}
    } else {		/* top-left or bottom-right */
	if (cos(value) >= 0.0) { /* bottom-right */
	    *ox = x + width;
	    *oy = y + (int)spRound(h2 - h2 * t1);
	} else {		/* top-left */
	    *ox = x;
	    *oy = y + (int)spRound(h2 + h2 * t1);
	}
    }
    
    return;
}

void spDrawArcWin(HDC hdc, spGraphics graphics,
		  spGraphicsType type, HPEN null_pen, HBRUSH null_brush,
		  int x, int y, int width, int height, double angle1, double angle2)
{
    double c;
    int sx, sy;
    int ex, ey;
    int prev_mode;
    int prev_bkmode;
    HPEN hprevpen;
    HBRUSH hprevbrush;
    
    width += 1; height += 1;
    //angle2 += 1;
    
    if (angle2 >= 360.0 || angle2 <= -360.0) {
	sx = x; sy = y + (int)spRound((double)height / 2.0);
	ex = sx; ey = sy;
    } else {
	c = PI / 180.0;
	convArcRadianToCoord(x, y, width, height, c * angle1, &sx, &sy);
	convArcRadianToCoord(x, y, width, height, c * (angle1 + angle2), &ex, &ey);
    }

    spDebug(60, "spDrawArcWin", "sx = %d, sy = %d, ex = %d, ey = %d\n",
	    sx, sy, ex, ey);
    
    if (type == SP_GRAPHICS_BRUSH) {
	hprevpen = SelectObject(hdc, null_pen);
	hprevbrush = SelectObject(hdc, SpGraphicsArch(graphics).brush);
    } else {
	hprevbrush = SelectObject(hdc, null_brush);
	hprevpen = SelectObject(hdc, SpGraphicsArch(graphics).pen);
    }
    prev_bkmode = SetBkMode(hdc, TRANSPARENT);
    prev_mode = SetROP2(hdc, SpGraphicsArch(graphics).mode);
    if (type == SP_GRAPHICS_BRUSH) {
	if (angle2 < 0.0) {
	    Pie(hdc, x, y, x + width, y + height, ex, ey, sx, sy);
	} else {
	    Pie(hdc, x, y, x + width, y + height, sx, sy, ex, ey);
	}
    } else {
	if (angle2 < 0.0) {
	    Arc(hdc, x, y, x + width, y + height, ex, ey, sx, sy);
	} else {
	    Arc(hdc, x, y, x + width, y + height, sx, sy, ex, ey);
	}
    }
    SelectObject(hdc, /*null_pen*/hprevpen);
    SelectObject(hdc, /*null_brush*/hprevbrush);
    SetROP2(hdc, prev_mode);
    SetBkMode(hdc, prev_bkmode);

    return;
}

void spFillArcArch(spComponent component, spGraphics graphics,
		   int x, int y, int width, int height, double angle1, double angle2)
{
    if (graphics == NULL) graphics = SpTopLevelPart(sp_toplevel).graphics;

    spDrawArcWin(SpPrimitiveArch(component).memdc, graphics, SP_GRAPHICS_BRUSH,
		 SpTopLevelArch(sp_toplevel).null_pen, SpTopLevelArch(sp_toplevel).null_brush,
		 x, y, width, height, angle1, angle2);
    
    return;
}

void spDrawArcArch(spComponent component, spGraphics graphics,
		   int x, int y, int width, int height, double angle1, double angle2)
{
    if (graphics == NULL) graphics = SpTopLevelPart(sp_toplevel).graphics;

    spDrawArcWin(SpPrimitiveArch(component).memdc, graphics, SP_GRAPHICS_PEN,
		 SpTopLevelArch(sp_toplevel).null_pen, SpTopLevelArch(sp_toplevel).null_brush,
		 x, y, width, height, angle1, angle2);
    
    return;
}

void spDrawLineWin(HDC hdc, spGraphics graphics, int x1, int y1, int x2, int y2)
{
    int prev_mode;
    int prev_bkmode;
    HPEN hprevpen;
    
    hprevpen = SelectObject(hdc, SpGraphicsArch(graphics).pen);
    prev_bkmode = SetBkMode(hdc, TRANSPARENT);
    prev_mode = SetROP2(hdc, SpGraphicsArch(graphics).mode);
    MoveToEx(hdc, x1, y1, NULL);
    LineTo(hdc, x2, y2);
    SetPixel(hdc, x2, y2, SpGraphicsPart(graphics).fg_pixel);
    SelectObject(hdc, hprevpen);
    SetROP2(hdc, prev_mode);
    SetBkMode(hdc, prev_bkmode);
    
    return;
}

void spDrawLineArch(spComponent component, spGraphics graphics,
		    int x1, int y1, int x2, int y2)
{
    if (graphics == NULL) graphics = SpTopLevelPart(sp_toplevel).graphics;

    spDrawLineWin(SpPrimitiveArch(component).memdc, graphics, x1, y1, x2, y2);
    
    return;
}

void spDrawPointWin(HDC hdc, spGraphics graphics, int x, int y)
{
    int prev_mode;
    HPEN hprevpen;
    
    hprevpen = SelectObject(hdc, SpGraphicsArch(graphics).pen);
    prev_mode = SetROP2(hdc, SpGraphicsArch(graphics).mode);
    SetPixel(hdc, x, y, SpGraphicsPart(graphics).fg_pixel);
    SelectObject(hdc, hprevpen);
    SetROP2(hdc, prev_mode);
    
    return;
}

void spDrawPointArch(spComponent component, spGraphics graphics, int x, int y)
{
    if (spIsDrawable(component) == SP_FALSE) return;

    if (graphics == NULL) graphics = SpTopLevelPart(sp_toplevel).graphics;
    
    spDrawPointWin(SpPrimitiveArch(component).memdc, graphics, x, y);
    
    return;
}

void spDrawStringWin(HDC hdc, spGraphics graphics, int x, int y, const char *string)
{
    SIZE size;
    TEXTMETRIC text_metric;
    int x_shift, y_shift;
    int x_textout, y_textout;
    int prev_mode;
    int prev_bkmode;
    COLORREF prev_color;
    HFONT prev_font = NULL;
    
    if (SpGraphicsArch(graphics).font != NULL) {
	prev_font = SelectObject(hdc,
				 SpGraphicsArch(graphics).font);
    }

    GetTextMetrics(hdc, &text_metric);
    y_shift = -text_metric.tmAscent;

    x_shift = 0;
    if (SpGraphicsPart(graphics).string_alignment == SP_ALIGNMENT_CENTER
	|| SpGraphicsPart(graphics).string_alignment == SP_ALIGNMENT_END) {
	GetTextExtentPoint32(hdc, string, (int)strlen(string), &size);
	if (SpGraphicsPart(graphics).string_alignment == SP_ALIGNMENT_CENTER) {
	    x_shift = -size.cx / 2;
	} else {
	    x_shift = -size.cx;
	}
    }
    spDebug(80, "spDrawStringWin", "no rotation case: x_shift = %d, y_shift = %d\n",
	    x_shift, y_shift);

    if (SpGraphicsPart(graphics).string_rotation != 0) {
	double costheta, sintheta;
	long x_shift_temp;
	
	costheta = cos(PI * (double)SpGraphicsPart(graphics).string_rotation / 180.0);
	sintheta = sin(PI * (double)SpGraphicsPart(graphics).string_rotation / 180.0);
	spDebug(100, "spDrawStringWin", "rotation = %d, costheta = %f, sintheta = %f\n",
		SpGraphicsPart(graphics).string_rotation, costheta, sintheta);
	
	x_shift_temp = (int)spRound((double)x_shift * costheta + (double)y_shift * sintheta);
	y_shift = (int)spRound((double)y_shift * costheta - (double)x_shift * sintheta);
	x_shift = x_shift_temp;
	spDebug(80, "spDrawStringWin", "rotation case: x_shift = %d, y_shift = %d\n",
		x_shift, y_shift);
    }
    
    x_textout = x + x_shift;
    y_textout = y + y_shift;
    
    prev_color = SetTextColor(hdc, SpGraphicsPart(graphics).fg_pixel);
    prev_bkmode = SetBkMode(hdc, TRANSPARENT);
    prev_mode = SetROP2(hdc, SpGraphicsArch(graphics).mode);
    TextOut(hdc, x_textout, y_textout, string, (int)strlen(string));
    SetROP2(hdc, prev_mode);
    SetBkMode(hdc, prev_bkmode);
    SetTextColor(hdc, prev_color);
    
    if (SpGraphicsArch(graphics).font != NULL) {
	SelectObject(hdc, prev_font);
    }
    
    return;
}

void spDrawStringArch(spComponent component, spGraphics graphics,
		      int x, int y, const char *string)
{
    if (graphics == NULL) graphics = SpTopLevelPart(sp_toplevel).graphics;

    spDrawStringWin(SpPrimitiveArch(component).memdc, graphics, x, y, string);

    return;
}

spBool spGetStringExtentWin(HDC hdc, spGraphics graphics, const char *string,
			    int *x, int *y, int *width, int *height, int *next_x)
{
    HFONT prev_font = NULL;
    SIZE size;
    int overhang;
    TEXTMETRIC text_metric;

    if (SpGraphicsArch(graphics).font != NULL) {
	prev_font = SelectObject(hdc, SpGraphicsArch(graphics).font);
    }

    GetTextExtentPoint32(hdc, string, (int)strlen(string), &size);
    GetTextMetrics(hdc, &text_metric);
    spDebug(100, "spGetStringExtentWin", "tmAscent = %ld, tmOverhang = %ld, tmItalic = %d, tmWeight = %ld\n",
	    text_metric.tmAscent, text_metric.tmOverhang, text_metric.tmItalic, text_metric.tmWeight);

    if (x != NULL) *x = -1;
    if (y != NULL) {
	*y = -text_metric.tmAscent - 1;
    }
    if (width != NULL) {
	*width = size.cx;
	overhang = 0;
	if (text_metric.tmOverhang > 0) {
	    overhang = text_metric.tmOverhang;
	} else {
	    if (text_metric.tmItalic) {
		overhang = (int)spRound((double)text_metric.tmAscent * 0.26);
	    }
	    if (text_metric.tmWeight > 400) {
		double overhang2;
		
		overhang2 = (double)text_metric.tmAscent * ((double)(text_metric.tmWeight - 400) / 8000.0);
		spDebug(50, "spGetStringExtentWin", "overhang2 = %f\n", overhang2);
		
		overhang += (int)spRound(overhang2);
	    }
	    
	    spDebug(50, "spGetStringExtentWin", "overhang = %d\n", overhang);
	}
	*width += overhang;
    }
    if (height != NULL) *height = size.cy;
    if (next_x != NULL) {
	*next_x = size.cx;
    }

    if (SpGraphicsArch(graphics).font != NULL) {
	SelectObject(hdc, prev_font);
    }

    return SP_TRUE;
}

spBool spGetStringExtentArch(spComponent component, spGraphics graphics, const char *string,
			     int *x, int *y, int *width, int *height, int *next_x)
{
    HDC hdc;
    HWND hwnd;
    
    if (graphics == NULL) graphics = SpTopLevelPart(sp_toplevel).graphics;

    if (component != NULL) {
	hwnd = SpPrimitiveArch(component).hwnd;
    } else {
	hwnd = GetDesktopWindow();
    }
    
    hdc = GetDC(hwnd);
    spGetStringExtentWin(hdc, graphics, string, x, y, width, height, next_x);
    ReleaseDC(hwnd, hdc);

    return SP_TRUE;
}

spBool spCreatePolygonArch(spPolygon polygon)
{
    SpPolygonArch(polygon).points = xalloc(polygon->max_point, POINT);
    return SP_TRUE;
}

spBool spDestroyPolygonArch(spPolygon polygon)
{
    xfree(SpPolygonArch(polygon).points);
    return SP_TRUE;
}

spBool spStartPolygonPointArch(spPolygon polygon, int x, int y)
{
    SpPolygonArch(polygon).points[0].x = x;
    SpPolygonArch(polygon).points[0].y = y;
    
    return SP_TRUE;
}

spBool spAddPolygonPointArch(spPolygon polygon, int x, int y)
{
    SpPolygonArch(polygon).points[polygon->num_point].x = x;
    SpPolygonArch(polygon).points[polygon->num_point].y = y;
    
    return SP_TRUE;
}

spBool spAddPolygonRelativePointArch(spPolygon polygon, int dx, int dy)
{
    SpPolygonArch(polygon).points[polygon->num_point].x
	= polygon->px + dx;
    SpPolygonArch(polygon).points[polygon->num_point].y
	= polygon->py + dy;
    
    return SP_TRUE;
}

spBool spEndPolygonPointArch(spPolygon polygon)
{
    return SP_TRUE;
}

spBool spGetPolygonPointArch(spPolygon polygon, int index, int *x, int *y)
{
    if (x != NULL) *x = SpPolygonArch(polygon).points[index].x;
    if (y != NULL) *y = SpPolygonArch(polygon).points[index].y;
    
    return SP_TRUE;
}

spBool spMovePolygonArch(spPolygon polygon, int dx, int dy)
{
    int i;

    for (i = 0; i < polygon->num_point; i++) {
	SpPolygonArch(polygon).points[i].x += dx;
	SpPolygonArch(polygon).points[i].y += dy;
    }
    
    return SP_TRUE;
}

spBool spDrawPolygonWin(HDC hdc, spGraphics graphics,
			spGraphicsType type, HPEN null_pen, HBRUSH null_brush, spPolygon polygon)
{
    int prev_mode;
    int prev_bkmode;
    int prev_fillmode;
    HPEN hprevpen;
    HBRUSH hprevbrush;
    BOOL flag;

    if (type == SP_GRAPHICS_BRUSH) {
	hprevpen = SelectObject(hdc, null_pen);
	hprevbrush = SelectObject(hdc, SpGraphicsArch(graphics).brush);
    } else {
	hprevbrush = SelectObject(hdc, null_brush);
	hprevpen = SelectObject(hdc, SpGraphicsArch(graphics).pen);
    }
    prev_mode = SetROP2(hdc, SpGraphicsArch(graphics).mode);
    prev_bkmode = SetBkMode(hdc, TRANSPARENT);
    prev_fillmode = SetPolyFillMode(hdc, WINDING/*ALTERNATE*/);

    flag = Polygon(hdc, SpPolygonArch(polygon).points, polygon->num_point);
    
    SetPolyFillMode(hdc, prev_fillmode);
    SelectObject(hdc, hprevpen);
    SelectObject(hdc, hprevbrush);
    SetROP2(hdc, prev_mode);
    SetBkMode(hdc, prev_bkmode);

    if (flag) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spBool spFillPolygonArch(spComponent component, spGraphics graphics, spPolygon polygon)
{
    if (graphics == NULL) graphics = SpTopLevelPart(sp_toplevel).graphics;

    return spDrawPolygonWin(SpPrimitiveArch(component).memdc, graphics,
			    SP_GRAPHICS_BRUSH,
			    SpTopLevelArch(sp_toplevel).null_pen, SpTopLevelArch(sp_toplevel).null_brush,
			    polygon);
}
     
spBool spDrawPolygonArch(spComponent component, spGraphics graphics, spPolygon polygon)
{
    if (graphics == NULL) graphics = SpTopLevelPart(sp_toplevel).graphics;
    

    return spDrawPolygonWin(SpPrimitiveArch(component).memdc, graphics,
			    SP_GRAPHICS_PEN,
			    SpTopLevelArch(sp_toplevel).null_pen, SpTopLevelArch(sp_toplevel).null_brush,
			    polygon);
}

unsigned long spGetDrawingCapsWin(void)
{
    unsigned long caps;
    
    caps = SP_DRAWING_CAPS_COPY_IMAGE
	| SP_DRAWING_CAPS_COPY_IMAGE_STRETCHABLE
	| SP_DRAWING_CAPS_FILL_RECTANGLE | SP_DRAWING_CAPS_DRAW_RECTANGLE
	| SP_DRAWING_CAPS_FILL_ARC | SP_DRAWING_CAPS_DRAW_ARC
	| SP_DRAWING_CAPS_FILL_POLYGON | SP_DRAWING_CAPS_DRAW_POLYGON
	| SP_DRAWING_CAPS_DRAW_LINE
	| SP_DRAWING_CAPS_STRING_EXTENT | SP_DRAWING_CAPS_STRING_EXTENT_ACCURATE
	| SP_DRAWING_CAPS_STRING_ROTATION
	| SP_DRAWING_CAPS_DRAW_WIDE_LINE | SP_DRAWING_CAPS_DRAW_STYLED_LINE
	| SP_DRAWING_CAPS_DRAW_WIDE_RECTANGLE | SP_DRAWING_CAPS_DRAW_STYLED_RECTANGLE
	| SP_DRAWING_CAPS_DRAW_WIDE_POLYGON | SP_DRAWING_CAPS_DRAW_STYLED_POLYGON
	| SP_DRAWING_CAPS_DRAW_WIDE_ARC | SP_DRAWING_CAPS_DRAW_STYLED_ARC
	;

    if (spIsOSWinNT() == SP_TRUE) {
	caps |= SP_DRAWING_CAPS_DRAW_WIDE_STYLED_LINE
	    | SP_DRAWING_CAPS_DRAW_WIDE_STYLED_RECTANGLE
	    | SP_DRAWING_CAPS_DRAW_WIDE_STYLED_POLYGON
	    | SP_DRAWING_CAPS_DRAW_WIDE_STYLED_ARC
	    ;
    }

    return caps;
}

unsigned long spGetDrawingCapsArch(spComponent component)
{
    return spGetDrawingCapsWin();
}

void spUpdateCompatibleBitmapWin(spComponent component, int width, int height)
{
    HDC hdc;
    
    /* delete DC */
    if (SpPrimitiveArch(component).memdc != NULL) {
	DeleteDC(SpPrimitiveArch(component).memdc);
	SpPrimitiveArch(component).memdc = NULL;
    }
    
    hdc = GetDC(SpPrimitiveArch(component).hwnd);
    SpPrimitiveArch(component).memdc = CreateCompatibleDC(hdc);
    
    if (width > 0 && height > 0) {
	if (SpPrimitiveArch(component).hbitmap != NULL) {
	    DeleteObject(SpPrimitiveArch(component).hbitmap);
	    SpPrimitiveArch(component).hbitmap = NULL;
	}
	SpPrimitiveArch(component).hbitmap = CreateCompatibleBitmap(hdc, width, height);
	spDebug(80, "spUpdateCompatibleBitmapWin", "CreateCompatibleBitmap done: %s\n", spGetName(component));
    }
    SelectObject(SpPrimitiveArch(component).memdc, SpPrimitiveArch(component).hbitmap);
    spDebug(80, "spUpdateCompatibleBitmapWin", "SelectObject done: %s\n", spGetName(component));
    
    /* release DC */
    ReleaseDC(SpPrimitiveArch(component).hwnd, hdc);
    
    return;
}
