#include <sp/spDefs.h>
#include <sp/spBase.h>

#include <gtk/gtk.h>

#include <sp/spCanvas.h>

#include <sp/spTopLevelP.h>
#include <sp/spPrimitiveP.h>
#include <sp/spDrawP.h>
#include <sp/spDrawingPluginP.h>

#if defined(GTK3)
						
typedef struct _spPluginInstancePrinter
{
    spComponent component;
    double xfactor, yfactor;
    gboolean support_selection;
    int n_pages;
    int current_page;

    GtkWindow *parent_window;
    GtkWidget *parent_widget;
    
    GtkPrintSettings *settings;
    GtkPageSetup *page_setup;

    GtkPrintOperationResult print_result;
    GtkPrintOperation *print;
    GtkPrintContext *context;
} *spPluginInstancePrinter;

static spBool spInitPrinter(const char *lang)
{
    return SP_TRUE;
}

static spBool spFreePrinter(void)
{
    return SP_TRUE;
}

static void *spInitInstancePrinter(const char *lang)
{
    spPluginInstancePrinter pinstance;

    pinstance = xalloc(1, struct _spPluginInstancePrinter);
    memset(pinstance, 0, sizeof(struct _spPluginInstancePrinter));

    pinstance->settings = gtk_print_settings_new();
    pinstance->page_setup = gtk_page_setup_new();
    pinstance->n_pages = 1;
    pinstance->current_page = -1;
    pinstance->xfactor = 1.0;
    pinstance->yfactor = 1.0;
    
    return (void *)pinstance;
}

static spBool spFreeInstancePrinter(void *instance)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    if (pinstance->page_setup != NULL) {
	g_object_unref(pinstance->page_setup);
    }
    if (pinstance->settings != NULL) {
	g_object_unref(pinstance->settings);
    }

    xfree(instance);
    
    return SP_TRUE;
}

static spBool spGetClientSizePrinter(void *instance, int *width, int *height);

static spBool spCreateComponentPrinter(void *instance, void *parent_arch, spArg *args, int num_arg)
{
    int width, height;
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    spGetClientSizePrinter(instance, &width, &height);
    spSetArg(args[num_arg], SppWidth, width); num_arg++;
    spSetArg(args[num_arg], SppHeight, height); num_arg++;
    pinstance->component = spCreateComponentArg(SpPrimitiveClass, SpPrintCanvas, "print", NULL, args, num_arg);
    
    if (pinstance->component == NULL) return SP_FALSE;

    if (parent_arch != NULL) {
	GtkWidget *toplevel;
	pinstance->parent_widget = (GtkWidget *)parent_arch;
	toplevel = gtk_widget_get_toplevel(pinstance->parent_widget);
	if (gtk_widget_is_toplevel(toplevel)) {
	    pinstance->parent_window = GTK_WINDOW(toplevel);
	} else {
	    pinstance->parent_window = NULL;
	}
    }
    
    return SP_TRUE;
}

static spBool spDestroyComponentPrinter(void *instance)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    spDestroyComponent(pinstance->component);
    pinstance->component = NULL;
    
    pinstance->parent_widget = NULL;
    pinstance->parent_window = NULL;
    
    return SP_TRUE;
}

static spBool spOpenSetupDialogPrinter(void *instance, spArg *args, int num_arg)
{
    GtkPageSetup *page_setup;
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    page_setup = gtk_print_run_page_setup_dialog(pinstance->parent_window, pinstance->page_setup,
						 pinstance->settings);
    
    if (pinstance->page_setup != page_setup) {
	g_object_unref(pinstance->page_setup);
	pinstance->page_setup = page_setup;
    }

    return SP_TRUE;
}

static spBool spOpenTargetDialogPrinter(void *instance, spArg *args, int num_arg)
{
    int i;
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    
    spDebug(100, "spOpenTargetDialogPrinter", "num_arg = %d\n", num_arg);
    
    for (i = 0; i < num_arg; i++) {
	if (spEqParam(args[i].param, SppEnablePageRangeSelection)) {
	    pinstance->support_selection = (gboolean)args[i].value;
	} else if (spEqParam(args[i].param, SppMaxPageRange)) {
	    pinstance->n_pages = MAX((int)args[i].value, 1);
	} else if (spEqParam(args[i].param, SppCurrentPage)) {
	    pinstance->current_page = (int)args[i].value - 1;
	}
    }
    
    return SP_TRUE;
}

static spBool spCloseTargetPrinter(void *instance)
{
    spDebug(100, "spCloseTargetPrinter", "done\n");
    
    return SP_TRUE;
}

spBool spSetPageInfoPrinter(void *instance, spPaperSize paper_size, spPaperOrientation paper_orientation,
	double left_margin_mm, double right_margin_mm, double top_margin_mm, double bottom_margin_mm)
{
    char *paper_name;
    GtkPaperSize *gtk_paper_size;
    GtkPageOrientation orientation;
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    if (paper_size == SP_PAPER_SIZE_A3) {
	paper_name = GTK_PAPER_NAME_A3;
    } else if (paper_size == SP_PAPER_SIZE_A4) {
	paper_name = GTK_PAPER_NAME_A4;
    } else if (paper_size == SP_PAPER_SIZE_A5) {
	paper_name = GTK_PAPER_NAME_A5;
    } else if (paper_size == SP_PAPER_SIZE_B5) {
	paper_name = GTK_PAPER_NAME_B5;
    } else if (paper_size == SP_PAPER_SIZE_LETTER) {
	paper_name = GTK_PAPER_NAME_LETTER;
    } else if (paper_size == SP_PAPER_SIZE_EXECUTIVE) {
	paper_name = GTK_PAPER_NAME_EXECUTIVE;
    } else if (paper_size == SP_PAPER_SIZE_LEGAL) {
	paper_name = GTK_PAPER_NAME_LEGAL;
    } else {
	paper_name = NULL;
    }

    gtk_paper_size = gtk_paper_size_new(paper_name);
    if (paper_name == NULL) {
	double width_mm, height_mm;
	if (spGetPaperDimensions(paper_size, SP_PAPER_ORIENTATION_PORTRAIT,
				 &width_mm, &height_mm) == SP_TRUE) {
	    gtk_paper_size_set_size(gtk_paper_size, width_mm, height_mm, GTK_UNIT_MM);
	}
    }
    gtk_page_setup_set_paper_size(pinstance->page_setup, gtk_paper_size);
    gtk_paper_size_free(gtk_paper_size);

    if (paper_orientation == SP_PAPER_ORIENTATION_LANDSCAPE) {
	orientation = GTK_PAGE_ORIENTATION_LANDSCAPE;
    } else if (paper_orientation == SP_PAPER_ORIENTATION_SEASCAPE) {
	orientation = GTK_PAGE_ORIENTATION_REVERSE_LANDSCAPE;
    } else if (paper_orientation == SP_PAPER_ORIENTATION_UPSIDE_DOWN) {
	orientation = GTK_PAGE_ORIENTATION_REVERSE_PORTRAIT;
    } else {
	orientation = GTK_PAGE_ORIENTATION_PORTRAIT;
    }
    gtk_page_setup_set_orientation(pinstance->page_setup, orientation);

    gtk_page_setup_set_left_margin(pinstance->page_setup, left_margin_mm, GTK_UNIT_MM);
    gtk_page_setup_set_right_margin(pinstance->page_setup, right_margin_mm, GTK_UNIT_MM);
    gtk_page_setup_set_top_margin(pinstance->page_setup, top_margin_mm, GTK_UNIT_MM);
    gtk_page_setup_set_bottom_margin(pinstance->page_setup, bottom_margin_mm, GTK_UNIT_MM);
    
    return SP_TRUE;
}


spBool spGetPageInfoPrinter(void *instance, spPaperSize *paper_size, spPaperOrientation *paper_orientation,
        double *left_margin_mm, double *right_margin_mm, double *top_margin_mm, double *bottom_margin_mm)
{
    GtkPaperSize *gtk_paper_size;
    double width_mm, height_mm;
    GtkPageOrientation orientation;
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    gtk_paper_size = gtk_page_setup_get_paper_size(pinstance->page_setup);
    width_mm = gtk_paper_size_get_width(gtk_paper_size, GTK_UNIT_MM);
    height_mm = gtk_paper_size_get_height(gtk_paper_size, GTK_UNIT_MM);
    spGetPaperFromDimensions(width_mm, height_mm, paper_size, NULL);
    
    orientation = gtk_page_setup_get_orientation(pinstance->page_setup);
    
    if (orientation == GTK_PAGE_ORIENTATION_LANDSCAPE) {
	*paper_orientation = SP_PAPER_ORIENTATION_LANDSCAPE;
    } else if (orientation == GTK_PAGE_ORIENTATION_REVERSE_LANDSCAPE) {
	*paper_orientation = SP_PAPER_ORIENTATION_SEASCAPE;
    } else if (orientation == GTK_PAGE_ORIENTATION_REVERSE_PORTRAIT) {
	*paper_orientation = SP_PAPER_ORIENTATION_UPSIDE_DOWN;
    } else {
	*paper_orientation = SP_PAPER_ORIENTATION_PORTRAIT;
    }

    *left_margin_mm = gtk_page_setup_get_left_margin(pinstance->page_setup, GTK_UNIT_MM);
    *right_margin_mm = gtk_page_setup_get_right_margin(pinstance->page_setup, GTK_UNIT_MM);
    *top_margin_mm = gtk_page_setup_get_top_margin(pinstance->page_setup, GTK_UNIT_MM);
    *bottom_margin_mm = gtk_page_setup_get_bottom_margin(pinstance->page_setup, GTK_UNIT_MM);
    
    return SP_TRUE;
}

spBool spSetPageRangePrinter(void *instance, int start_page, int end_page, int max_page)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    
    pinstance->n_pages = MAX(max_page, 1);
    spDebug(50, "spSetPageRangePrinter", "n_pages = %d\n", pinstance->n_pages);

    return SP_TRUE;
}
	
unsigned long spGetDrawingCapsPrinter(void *instance)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    return spGetDrawingCapsArch(pinstance->component) | SP_DRAWING_CAPS_MULTIPLE_PAGES;
}

unsigned long spGetGraphicsModeCapsPrinter(void *instance)
{
    return 0L;
}

static spBool spIsDrawablePrinter(void *instance)
{
    spBool flag = SP_TRUE;
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    /*flag = spIsDrawable(pinstance->component);*/

    spDebug(100, "spIsDrawablePrinter", "flag = %d\n", flag);

    return flag;
}

static spBool spGetClientSizePrinter(void *instance, int *width, int *height)
{
    double widthd, heightd;
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    widthd = gtk_page_setup_get_page_width(pinstance->page_setup, GTK_UNIT_POINTS/*GTK_UNIT_PIXEL*/);
    widthd /= pinstance->xfactor;
    *width = (int)spRound(widthd);
    
    heightd = gtk_page_setup_get_page_height(pinstance->page_setup, GTK_UNIT_POINTS);
    heightd /= pinstance->yfactor;
    *height = (int)spRound(heightd);

    spDebug(20, "spGetClientSizePrinter", "widthd = %f, heightd = %f\n", widthd, heightd);
    
    return SP_TRUE;
}

static spBool spGetPhysicalSizePrinter(void *instance, double *width_mm, double *height_mm)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    *width_mm = gtk_page_setup_get_page_width(pinstance->page_setup, GTK_UNIT_MM);
    *height_mm = gtk_page_setup_get_page_height(pinstance->page_setup, GTK_UNIT_MM);
 
    return SP_TRUE;
}

static spBool spSetScalingFactorPrinter(void *instance, double xfactor, double yfactor)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    
    pinstance->xfactor = xfactor;
    pinstance->yfactor = yfactor;
    
    return SP_TRUE;
}
     
static void spBeginPrintGtkCB(GtkPrintOperation *print, GtkPrintContext *context, gpointer user_data)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)user_data;
  
    SpPrimitiveArch(pinstance->component).pixmap_cr = gtk_print_context_get_cairo_context(context);
    if (pinstance->xfactor != 1.0 || pinstance->yfactor != 1.0) {
	cairo_scale(SpPrimitiveArch(pinstance->component).pixmap_cr, pinstance->xfactor, pinstance->yfactor);
    }
    SpPrimitiveArch(pinstance->component).layout = gtk_print_context_create_pango_layout(context);
    SpComponentPart(pinstance->component).width = (int)spRound(gtk_print_context_get_width(context));
    SpComponentPart(pinstance->component).height = (int)spRound(gtk_print_context_get_height(context));

    spDebug(80, "spBeginPrintGtkCB", "width = %d, height = %d\n",
	    SpComponentPart(pinstance->component).width, SpComponentPart(pinstance->component).height);

    return;
}

static void spEndPrintGtkCB(GtkPrintOperation *print, GtkPrintContext *context, gpointer user_data)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)user_data;

    SpComponentPart(pinstance->component).width = -1;
    SpComponentPart(pinstance->component).height = -1;
    if (SpPrimitiveArch(pinstance->component).layout != NULL) {
	g_object_unref(SpPrimitiveArch(pinstance->component).layout);
    }
    SpPrimitiveArch(pinstance->component).pixmap_cr = NULL;

    spDebug(100, "spEndPrintGtkCB", "done\n");
    
    return;
}

static spPluginRec *spGetPluginRecPrinter(void);

static void spDrawPageGtkCB(GtkPrintOperation *print, GtkPrintContext *context,
			    gint page_nr, gpointer user_data)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)user_data;
  
    spEmitPluginHostCallback(spGetPluginRecPrinter(), SP_DRAWING_PLUGIN_HOST_CR_DRAW_PAGE, (void *)(page_nr + 1));
    spDebug(80, "spDrawPageGtkCB", "spEmitPluginHostCallback done\n");

    return;
}

static spBool spStartDrawingPrinter(void *instance, const char *job_name)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    spBool flag;
    GError *error;

    spDebug(100, "spStartDrawingPrinter", "in\n");
    
    pinstance->print = gtk_print_operation_new();
    gtk_print_operation_set_show_progress(pinstance->print, TRUE);
    gtk_print_operation_set_support_selection(pinstance->print, pinstance->support_selection);
    gtk_print_operation_set_has_selection(pinstance->print, pinstance->support_selection);
    gtk_print_operation_set_n_pages(pinstance->print, pinstance->n_pages);

    spDebug(80, "spStartDrawingPrinter", "pinstance->current_page = %d\n", pinstance->current_page);
    if (pinstance->current_page >= 0) {
        gtk_print_operation_set_current_page(pinstance->print, pinstance->current_page);
    }

    if (pinstance->settings != NULL) {
	gtk_print_operation_set_print_settings(pinstance->print, pinstance->settings);
    }
    if (pinstance->page_setup != NULL) {
	gtk_print_operation_set_default_page_setup(pinstance->print, pinstance->page_setup);
    }

    g_signal_connect(pinstance->print, "begin_print", G_CALLBACK(spBeginPrintGtkCB), pinstance);
    g_signal_connect(pinstance->print, "end_print", G_CALLBACK(spEndPrintGtkCB), pinstance);
    g_signal_connect(pinstance->print, "draw_page", G_CALLBACK(spDrawPageGtkCB), pinstance);

    pinstance->print_result = gtk_print_operation_run(pinstance->print, GTK_PRINT_OPERATION_ACTION_PRINT_DIALOG,
						      pinstance->parent_window, &error);

    if (pinstance->print_result == GTK_PRINT_OPERATION_RESULT_ERROR) {
	GtkWidget *error_dialog;
	error_dialog = gtk_message_dialog_new(pinstance->parent_window,
					      GTK_DIALOG_DESTROY_WITH_PARENT,
					      GTK_MESSAGE_ERROR,
					      GTK_BUTTONS_CLOSE,
					      "%s",
					      error->message);
	g_signal_connect(error_dialog, "response", 
			 G_CALLBACK (gtk_widget_destroy), NULL);
	gtk_widget_show(error_dialog);
	g_error_free(error);	
	flag = SP_FALSE;
    } else if (pinstance->print_result == GTK_PRINT_OPERATION_RESULT_APPLY) {
	if (pinstance->settings != NULL) {
	    g_object_unref(pinstance->settings);
	}
	pinstance->settings = g_object_ref(gtk_print_operation_get_print_settings(pinstance->print));
	flag = SP_TRUE;
    } else {
	flag = SP_TRUE;
    }

    g_object_unref(pinstance->print);
    pinstance->print = NULL;
    
    spDebug(100, "spStartDrawingPrinter", "done: flag = %d\n", flag);
    
    return flag;
}

static spBool spEndDrawingPrinter(void *instance)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    if (pinstance->print != NULL) {
	g_object_unref(pinstance->print);
    }
    pinstance->print = NULL;
    
    spDebug(100, "spEndDrawingPrinter", "done\n");
    
    return SP_TRUE;
}

static spBool spCopyImagePrinter(void *instance, void *src_image_arch,
				 int src_x, int src_y, int src_width, int src_height,
				 int dest_x, int dest_y, int dest_width, int dest_height)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    spDebug(80, "spCopyImagePrinter", "in\n");
    
    return spStretchImageArch((spComponent)src_image_arch, pinstance->component, 
			      src_x, src_y, src_width, src_height,
			      dest_x, dest_y, dest_width, dest_height);
}

void spFillRectanglePrinter(void *instance, spGraphics graphics,
			    int x, int y, int width, int height)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    spFillRectangleArch(pinstance->component, graphics, x, y, width, height);

    return;
}

void spDrawRectanglePrinter(void *instance, spGraphics graphics,
			    int x, int y, int width, int height)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    spDrawRectangleArch(pinstance->component, graphics, x, y, width, height);

    return;
}

void spFillArcPrinter(void *instance, spGraphics graphics,
		      int x, int y, int width, int height, double angle1, double angle2)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    
    spFillArcArch(pinstance->component, graphics, x, y, width, height, angle1, angle2);
    
    return;
}
     
void spDrawArcPrinter(void *instance, spGraphics graphics,
		      int x, int y, int width, int height, double angle1, double angle2)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    
    spDrawArcArch(pinstance->component, graphics, x, y, width, height, angle1, angle2);
    
    return;
}

void spDrawLinePrinter(void *instance, spGraphics graphics,
		       int x1, int y1, int x2, int y2)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    
    spDrawLineArch(pinstance->component, graphics, x1, y1, x2, y2);
    
    return;
}

void spDrawPointPrinter(void *instance, spGraphics graphics,
			int x, int y)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    spDrawPointArch(pinstance->component, graphics, x, y);
    
    return;
}

void spDrawStringPrinter(void *instance, spGraphics graphics,
			 int x, int y, const char *string)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    
    spDrawStringArch(pinstance->component, graphics, x, y, string);
    
    return;
}

spBool spGetStringExtentPrinter(void *instance, spGraphics graphics, const char *string,
				int *x, int *y, int *width, int *height, int *next_x)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    
    return spGetStringExtentArch(pinstance->component, graphics, string, x, y, width, height, next_x);
}

spBool spFillPolygonPrinter(void *instance, spGraphics graphics, void *polygon_instance)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    spPolygon polygon = (spPolygon)polygon_instance;
    
    return spFillPolygonArch(pinstance->component, graphics, polygon);
}

spBool spDrawPolygonPrinter(void *instance, spGraphics graphics, void *polygon_instance)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    spPolygon polygon = (spPolygon)polygon_instance;
    
    return spDrawPolygonArch(pinstance->component, graphics, polygon);
}

static spDrawingPluginRec sp_drawing_plugin_printer = {
    NULL,
    NULL,

    SP_PLUGIN_DRAWING,
    SP_DRAWING_PLUGIN_PRINTER,
    1,
    SP_PLUGIN_PRIORITY_MIDDLE,
    SP_DRAWING_PLUGIN_CAPS_MULTIPLE_PAGES | SP_DRAWING_PLUGIN_CAPS_CALLBACK_DRAW,		/* caps */
    spInitPrinter,
    spFreePrinter,
    "Printer",
    "Printer Plugin for GTK+  Version 0.1\nCopyright(C) Hideki Banno",
    
    spInitInstancePrinter,
    spFreeInstancePrinter,
    NULL,
    NULL,

    SP_DRAWING_DEVICE_PRINTER,

    NULL,
    NULL,
    NULL,
    NULL,

    NULL,
    spCreateComponentPrinter,
    spDestroyComponentPrinter,
    NULL/*spSetParamsPrinter*/,
    NULL/*spGetParamsPrinter*/,

    spOpenSetupDialogPrinter,
    
    spSetPageInfoPrinter,
    spGetPageInfoPrinter,
    spSetPageRangePrinter,
    NULL,
    NULL,

    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    
    NULL,
    spOpenTargetDialogPrinter,

    spCloseTargetPrinter,
    
    spIsDrawablePrinter,

    spGetDrawingCapsPrinter,
    spGetGraphicsModeCapsPrinter,
    NULL,
    spGetClientSizePrinter,
    spGetPhysicalSizePrinter,
    spSetScalingFactorPrinter,

    spStartDrawingPrinter,
    spEndDrawingPrinter,
    NULL,

    NULL,
    spCopyImagePrinter,

    spFillRectanglePrinter,
    spDrawRectanglePrinter,
    spFillArcPrinter,
    spDrawArcPrinter,
    spDrawLinePrinter,
    spDrawPointPrinter,
    spDrawStringPrinter,
    spGetStringExtentPrinter,

    NULL,
    NULL,
    NULL,
    
    NULL,
    spFillPolygonPrinter,
    spDrawPolygonPrinter,
};

static spPluginRec *spGetPluginRecPrinter(void)
{
    return (spPluginRec *)&sp_drawing_plugin_printer;
}

spBool spAddPrinterPlugin(void)
{
    return spAddPluginRecList((spPluginRec *)&sp_drawing_plugin_printer, SP_DRAWING_PLUGIN_PRINTER);
}

spBool spRemovePrinterPlugin(void)
{
    return spRemovePluginRecList((spPluginRec *)&sp_drawing_plugin_printer);
}
#endif
