#include <sp/spDefs.h>
#include <sp/spBase.h>

#include <sp/spTopLevelP.h>
#include <sp/spPrimitiveP.h>
#include <sp/spDrawP.h>
#include <sp/spDrawingPluginP.h>

#if defined(MACOSX)
#include <Carbon/Carbon.h>
#define SP_MAC_USE_PRINTING_MANAGER
#else
#if TARGET_API_MAC_CARBON
#define SP_MAC_USE_PRINTING_MANAGER
#else
#undef SP_MAC_USE_PRINTING_MANAGER
#endif
#ifdef SP_MAC_USE_PRINTING_MANAGER
#include <PMApplication.h>
#else
#include <Printing.h>
#endif
#endif

typedef struct _spPluginInstancePrinter
{
    WindowRef window;
    GrafPtr save_port;
    
    spBool target_ok;
    
#ifdef SP_MAC_USE_PRINTING_MANAGER
    PMPrintSession session;
    PMPrintSettings settings;
    PMPageFormat page_format;
    GrafPtr pport;
    double default_hdpi;
    double default_vdpi;
#else
    THPrint hprint;
    TPPrPort pport;

    int x_min_dpi;
    int x_max_dpi;
    int y_min_dpi;
    int y_max_dpi;
    
    spBool landscape;
    spBool support_enhanced;
#endif
    int current_page;
    int num_page;
    int start_page;
    int end_page;
    
    double xfactor, yfactor;

    double paper_width_mm;
    double paper_height_mm;
    double left_margin_mm;
    double right_margin_mm;
    double top_margin_mm;
    double bottom_margin_mm;
} *spPluginInstancePrinter;

#ifdef SP_MAC_USE_PRINTING_MANAGER
static spBool getDPI(spPluginInstancePrinter pinstance, double *hdpi, double *vdpi)
{
    PMResolution res;

    if (PMGetResolution(pinstance->page_format, &res) == noErr) {
	spDebug(80, "getDPI", "hRes = %f, vRes = %f\n", res.hRes, res.vRes);
	if (hdpi != NULL) *hdpi = res.hRes;
	if (vdpi != NULL) *vdpi = res.vRes;
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static spBool getPointToMMFactor(spPluginInstancePrinter pinstance, 
				 double *hfactor, double *vfactor)
{
    PMResolution res;

    if (PMGetResolution(pinstance->page_format, &res) == noErr) {
	spDebug(80, "getPointToMMFactor", "hRes = %f, vRes = %f\n", res.hRes, res.vRes);
	if (hfactor != NULL) *hfactor = 25.4 / res.hRes;
	if (vfactor != NULL) *vfactor = 25.4 / res.vRes;
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}
#else
static int getVDPI(THPrint hprint)
{
    spDebug(80, "getVDPI", "iVRes = %d\n", (*hprint)->prInfo.iVRes);
    return (*hprint)->prInfo.iVRes;
}

static int getHDPI(THPrint hprint)
{
    spDebug(80, "getHDPI", "iHRes = %d\n", (*hprint)->prInfo.iHRes);
    return (*hprint)->prInfo.iHRes;
}

static spBool checkIfLandscape(THPrint hprint)
{
    TGetRotnBlk rotnblk;
    
    rotnblk.iOpCode = getRotnOp;
    rotnblk.hPrint = hprint;
    PrGeneral((Ptr)&rotnblk);
    if (PrError() == noErr && rotnblk.iError == noErr) {
        if (rotnblk.fLandscape) {
            return SP_TRUE;
        } else {
            return SP_FALSE;
        }
    }

    return SP_FALSE;
}
#endif

static spBool spInitPrinter(const char *lang)
{
    return SP_TRUE;
}

static spBool spFreePrinter(void)
{
    return SP_TRUE;
}

static spBool updatePaperSizePrinter(spPluginInstancePrinter pinstance)
{
#ifdef SP_MAC_USE_PRINTING_MANAGER
    {
	PMRect paperrect, pagerect;
	double hfactor, vfactor;

	if (PMGetUnadjustedPaperRect(pinstance->page_format, &paperrect) != noErr
	    || PMGetUnadjustedPageRect(pinstance->page_format, &pagerect) != noErr
	    || getPointToMMFactor(pinstance, &hfactor, &vfactor) == SP_FALSE) {
	    return SP_FALSE;
	}

	pinstance->paper_width_mm = (paperrect.right - paperrect.left) * hfactor;
	pinstance->paper_height_mm = (paperrect.bottom - paperrect.top) * vfactor;
	
	pinstance->left_margin_mm = (pagerect.left - paperrect.left) * hfactor;
	pinstance->right_margin_mm = (paperrect.right - pagerect.right) * hfactor;
	pinstance->top_margin_mm = (pagerect.top - paperrect.top) * vfactor;
	pinstance->bottom_margin_mm = (paperrect.bottom - pagerect.bottom) * vfactor;
    }
#else
    {
        Rect rPaper, rPage;

        rPaper = (*pinstance->hprint)->rPaper;
        rPage = (*pinstance->hprint)->prInfo.rPage;

        spDebug(50, "updatePaperSizePrinter", "landscape = %d, rPaper.left = %d, rPaper.right = %d, rPaper.top = %d, rPaper.bottom = %d\n",
                pinstance->landscape, rPaper.left, rPaper.right, rPaper.top, rPaper.bottom);
        spDebug(50, "updatePaperSizePrinter", "rPage.left = %d, rPage.right = %d, rPage.top = %d, rPage.bottom = %d\n",
                rPage.left, rPage.right, rPage.top, rPage.bottom);
	
        pinstance->paper_width_mm = (double)(rPaper.right - rPaper.left)
            / (double)getHDPI(pinstance->hprint) * 25.4;
        pinstance->paper_height_mm = (double)(rPaper.bottom - rPaper.top)
            / (double)getVDPI(pinstance->hprint) * 25.4;
	
        pinstance->left_margin_mm = (double)(rPage.left - rPaper.left)
            / (double)getHDPI(pinstance->hprint) * 25.4;
        pinstance->right_margin_mm = (double)(rPaper.right - rPage.right)
            / (double)getHDPI(pinstance->hprint) * 25.4;
        pinstance->top_margin_mm = (double)(rPage.top - rPaper.top)
            / (double)getVDPI(pinstance->hprint) * 25.4;
        pinstance->bottom_margin_mm = (double)(rPaper.bottom - rPage.bottom)
            / (double)getVDPI(pinstance->hprint) * 25.4;
        spDebug(50, "updatePaperSizePrinter", "left_margin_mm = %f, right_margin_mm = %f, top_margin_mm = %f, bottom_margin_mm = %f\n",
                pinstance->left_margin_mm, pinstance->right_margin_mm, pinstance->top_margin_mm, pinstance->bottom_margin_mm);
    }
#endif
    
    return SP_TRUE;
}

static void *spInitInstancePrinter(const char *lang)
{
    spPluginInstancePrinter pinstance;
#ifdef SP_MAC_USE_PRINTING_MANAGER
    PMPrintSession session = NULL;
    PMPrintSettings settings = NULL;
    PMPageFormat page_format = NULL;
    spBool ok = SP_FALSE;
    static double default_hdpi = 0.0, default_vdpi = 0.0;
    
    if (PMCreateSession(&session) == noErr) {
	if (PMCreatePageFormat(&page_format) == noErr) {
	    PMSessionDefaultPageFormat(session, page_format);
	    
	    if (PMCreatePrintSettings(&settings) == noErr) {
		PMSessionDefaultPrintSettings(session, settings);
		ok = SP_TRUE;
	    }
	    if (ok == SP_FALSE) {
		PMRelease(page_format);
	    }
	}
	if (ok == SP_FALSE) {
	    PMRelease(session);
	}
    }
    if (ok == SP_FALSE) {
	return SP_FALSE;
    }
#else
    PrOpen();
    if (PrError() != noErr) {
	return NULL;
    }
#endif

    pinstance = xalloc(1, struct _spPluginInstancePrinter);
    memset(pinstance, 0, sizeof(struct _spPluginInstancePrinter));
    
#ifdef SP_MAC_USE_PRINTING_MANAGER
    pinstance->session = session;
    pinstance->settings = settings;
    pinstance->page_format = page_format;
    if (default_hdpi <= 0.0 || default_vdpi <= 0.0) {
	getDPI(pinstance, &default_hdpi, &default_vdpi);
    }
    pinstance->default_hdpi = default_hdpi;
    pinstance->default_vdpi = default_vdpi;
    spDebug(80, "getDPI", "default_hdpi = %f, default_vdpi = %f\n", default_hdpi, default_vdpi);
#else
    pinstance->hprint = (THPrint)NewHandleClear(sizeof(TPrint));
    PrintDefault(pinstance->hprint);
    
    pinstance->landscape = SP_FALSE;
    pinstance->support_enhanced = SP_FALSE;
    {
	TDftBitsBlk bitsblk;
	bitsblk.iOpCode = draftBitsOp;
	bitsblk.hPrint = pinstance->hprint;
	PrGeneral((Ptr)&bitsblk);
	if (PrError() == noErr && bitsblk.iError == noErr) {
	    pinstance->support_enhanced = SP_TRUE;
	}
    }
    {
	int i;
	TGetRslBlk getrslblk;

	getrslblk.iOpCode = getRslDataOp;
	PrGeneral((Ptr)&getrslblk);
	if (PrError() == noErr && getrslblk.iError == noErr) {
	    pinstance->x_min_dpi = getrslblk.xRslRg.iMin;
	    pinstance->x_max_dpi = getrslblk.xRslRg.iMax;
	    pinstance->y_min_dpi = getrslblk.yRslRg.iMin;
	    pinstance->y_max_dpi = getrslblk.yRslRg.iMax;
	    
	    spDebug(50, "spInitInstancePrinter", "xRslRg = (%d-%d), yRslRg = (%d-%d)\n",
		    getrslblk.xRslRg.iMin, getrslblk.xRslRg.iMax,
		    getrslblk.yRslRg.iMin, getrslblk.yRslRg.iMax);
	    for (i = 0; i < getrslblk.iRslRecCnt; i++) {
		spDebug(50, "spInitInstancePrinter", "resolution %d: iXRsl = %d, iYRsl = %d\n",
			i, getrslblk.rgRslRec[i].iXRsl, getrslblk.rgRslRec[i].iYRsl);
		
	    }
	}
    }
#endif
    
    pinstance->current_page = 0;
    pinstance->num_page = 1;
    pinstance->start_page = 0;
    pinstance->end_page = 0;
    
    pinstance->xfactor = 1.0;
    pinstance->yfactor = 1.0;
    pinstance->paper_width_mm = 0.0;
    pinstance->paper_height_mm = 0.0;
    spGetPaperDimensions(SP_PAPER_SIZE_A4, SP_PAPER_ORIENTATION_PORTRAIT,
			 &pinstance->paper_width_mm, &pinstance->paper_height_mm);
    pinstance->left_margin_mm = 20.0;
    pinstance->right_margin_mm = 20.0;
    pinstance->top_margin_mm = 20.0;
    pinstance->bottom_margin_mm = 20.0;
    
    updatePaperSizePrinter(pinstance);
    
    return (void *)pinstance;
}


static spBool spFreeInstancePrinter(void *instance)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

#ifdef SP_MAC_USE_PRINTING_MANAGER
    if (pinstance->settings != NULL) {
        PMRelease(pinstance->settings);
    }
    if (pinstance->page_format != NULL) {
        PMRelease(pinstance->page_format);
    }
    if (pinstance->session != NULL) {
	PMRelease(pinstance->session);
    }
#else
    if (pinstance->hprint != NULL) {
	DisposeHandle((Handle)pinstance->hprint);
    }
    PrClose();
#endif
    
    xfree(instance);
    return SP_TRUE;
}

static spBool spCreateComponentPrinter(void *instance, void *parent_arch, spArg *args, int num_arg)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    if (parent_arch != NULL) {
	pinstance->window = (WindowRef)parent_arch;
    }
    
    return SP_TRUE;
}

static spBool spDestroyComponentPrinter(void *instance)
{
    return SP_TRUE;
}

static spBool spOpenSetupDialogPrinter(void *instance, spArg *args, int num_arg)
{
    Boolean ok = true;
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
#ifdef SP_MAC_USE_PRINTING_MANAGER
    Boolean accepted = true;

    if (PMSessionPageSetupDialog(pinstance->session, pinstance->page_format, &accepted) != noErr
	|| !accepted) {
	spDebug(50, "spOpenTargetDialogPrinter", "accepted = %d\n", accepted);
	ok = false;
    }
#else
    ok = PrStlDialog(pinstance->hprint);

    if (ok) {
        pinstance->landscape = checkIfLandscape(pinstance->hprint);
    }
#endif
    
    spDebug(50, "spOpenSetupDialogPrinter", "ok = %d\n", ok);
	
    if (ok) {
	updatePaperSizePrinter(pinstance);
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static spBool spOpenTargetDialogPrinter(void *instance, spArg *args, int num_arg)
{
    int i;
    Boolean ok = true;
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    for (i = 0; i < num_arg; i++) {
	if (spEqParam(args[i].param, SppMaxPageRange)) {
	    pinstance->num_page = MAX((int)args[i].value, 1);
	}
    }
    
#ifdef SP_MAC_USE_PRINTING_MANAGER
    {
        Boolean accepted = true;

        if (pinstance->start_page >= 1 && pinstance->end_page >= 1 && pinstance->num_page >= 2) {
            if (PMSetFirstPage(pinstance->settings, pinstance->start_page, false) != noErr
                || PMSetLastPage(pinstance->settings, pinstance->end_page, false) != noErr) {
                spDebug(50, "spOpenTargetDialogPrinter", "PMSetFirstPage failed\n");
            }
        }

        if (PMSessionPrintDialog(pinstance->session, pinstance->settings, 
                                 pinstance->page_format, &accepted) != noErr
            || !accepted) {
            spDebug(50, "spOpenTargetDialogPrinter", "accepted = %d\n", accepted);
            ok = false;
        } else if (pinstance->num_page >= 2) {
            UInt32 first, last;
        
            if (PMGetFirstPage(pinstance->settings, &first) != noErr
                || PMGetLastPage(pinstance->settings, &last) != noErr) {
                spDebug(50, "spOpenTargetDialogPrinter", "PMGetFirstPage failed\n");
                pinstance->start_page = 1;
                pinstance->end_page = pinstance->num_page;
            } else {
                pinstance->start_page = first;
                pinstance->end_page = last;
            }
        }
    }
#else
    if (pinstance->start_page >= 1 && pinstance->end_page >= 1 && pinstance->num_page >= 2) {
        (*pinstance->hprint)->prJob.iFstPage = pinstance->start_page;
        (*pinstance->hprint)->prJob.iLstPage = pinstance->end_page;
    }
    
    ok = PrJobDialog(pinstance->hprint);

    if (ok) {
        if (pinstance->num_page >= 2) {
            pinstance->end_page = (*pinstance->hprint)->prJob.iLstPage;
            pinstance->start_page = (*pinstance->hprint)->prJob.iFstPage;
        }
    }
#endif
    
    spDebug(50, "spOpenTargetDialogPrinter", "ok = %d\n", ok);
    
    if (ok) {
        pinstance->end_page = MIN(MAX(pinstance->end_page, 1), pinstance->num_page);
        pinstance->start_page = MIN(MAX(pinstance->start_page, 1), pinstance->end_page);
	pinstance->target_ok = SP_TRUE;
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static spBool spCloseTargetPrinter(void *instance)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    pinstance->target_ok = SP_FALSE;
    
    return SP_TRUE;
}

spBool spSetPageInfoPrinter(void *instance, spPaperSize paper_size, spPaperOrientation paper_orientation,
	double left_margin_mm, double right_margin_mm, double top_margin_mm, double bottom_margin_mm)
{
#if 1
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    pinstance->left_margin_mm = left_margin_mm;
    pinstance->right_margin_mm = right_margin_mm;
    pinstance->top_margin_mm = top_margin_mm;
    pinstance->bottom_margin_mm = bottom_margin_mm;
    
    return spGetPaperDimensions(paper_size, paper_orientation,
				&pinstance->paper_width_mm, &pinstance->paper_height_mm);
#else
    return SP_FALSE;
#endif
}


spBool spGetPageInfoPrinter(void *instance, spPaperSize *paper_size, spPaperOrientation *paper_orientation,
        double *left_margin_mm, double *right_margin_mm, double *top_margin_mm, double *bottom_margin_mm)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    *left_margin_mm = pinstance->left_margin_mm;
    *right_margin_mm = pinstance->right_margin_mm;
    *top_margin_mm = pinstance->top_margin_mm;
    *bottom_margin_mm = pinstance->bottom_margin_mm;
    
    return spGetPaperFromDimensions(pinstance->paper_width_mm, pinstance->paper_height_mm,
				    paper_size, paper_orientation);
}

spBool spSetPageRangePrinter(void *instance, int start_page, int end_page, int max_page)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    pinstance->start_page = start_page;
    pinstance->end_page = end_page;
    pinstance->num_page = MAX(max_page, 1);
    
    return SP_TRUE;
}

spBool spGetPageRangePrinter(void *instance, int *start_page, int *end_page)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    
    *start_page = pinstance->start_page;
    *end_page = pinstance->end_page;

    return SP_TRUE;
}
	
unsigned long spGetDrawingCapsPrinter(void *instance)
{
    unsigned long caps = 0L;

    caps |= SP_DRAWING_CAPS_COPY_IMAGE | SP_DRAWING_CAPS_MULTIPLE_PAGES
	| SP_DRAWING_CAPS_FILL_RECTANGLE | SP_DRAWING_CAPS_DRAW_RECTANGLE
	| SP_DRAWING_CAPS_FILL_ARC | SP_DRAWING_CAPS_DRAW_ARC
	| SP_DRAWING_CAPS_FILL_POLYGON | SP_DRAWING_CAPS_DRAW_POLYGON
	| SP_DRAWING_CAPS_DRAW_LINE
	| SP_DRAWING_CAPS_STRING_EXTENT 
	| SP_DRAWING_CAPS_DRAW_WIDE_LINE | SP_DRAWING_CAPS_DRAW_STYLED_LINE | SP_DRAWING_CAPS_DRAW_WIDE_STYLED_LINE
	| SP_DRAWING_CAPS_DRAW_WIDE_RECTANGLE | SP_DRAWING_CAPS_DRAW_STYLED_RECTANGLE | SP_DRAWING_CAPS_DRAW_WIDE_STYLED_RECTANGLE
	| SP_DRAWING_CAPS_DRAW_WIDE_POLYGON 
	| SP_DRAWING_CAPS_DRAW_WIDE_ARC | SP_DRAWING_CAPS_DRAW_STYLED_ARC | SP_DRAWING_CAPS_DRAW_WIDE_STYLED_ARC
	;
    
    return caps;
}

unsigned long spGetGraphicsModeCapsPrinter(void *instance)
{
    return 0L;
}

static spBool spIsDrawablePrinter(void *instance)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    if (pinstance->target_ok == SP_TRUE || pinstance->pport != NULL) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

#ifndef SP_MAC_USE_PRINTING_MANAGER
static void getClientSize(spPluginInstancePrinter pinstance, int *width, int *height)
{
    Rect rPage;
    int page_width, page_height;

    rPage = (*pinstance->hprint)->prInfo.rPage;
    
    page_width = (rPage.right - rPage.left);
    page_height = (rPage.bottom - rPage.top);
    
    *width = (int)spRound((double)page_width / pinstance->xfactor);
    *height = (int)spRound((double)page_height / pinstance->yfactor);
    
    return;
}

static void getClientRect(spPluginInstancePrinter pinstance, Rect *rect)
{
    int width, height;
    
    getClientSize(pinstance, &width, &height);

    rect->left = rect->top = 0;
    rect->right = width;
    rect->bottom = height;
    spDebug(50, "getClientRect", "left = %d, right = %d, top = %d, bottom = %d\n",
	    rect->left, rect->right, rect->top, rect->bottom);

    return;
}
#endif

static spBool spGetClientSizePrinter(void *instance, int *width, int *height)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

#ifdef SP_MAC_USE_PRINTING_MANAGER
    {
	PMRect pmrect;

	if (PMGetAdjustedPageRect(pinstance->page_format, &pmrect) != noErr) {
	    return SP_FALSE;
	}
	
	*width = (int)spRound((pmrect.right - pmrect.left) / pinstance->xfactor);
	*height = (int)spRound((pmrect.bottom - pmrect.top) / pinstance->yfactor);
    }
#else
    getClientSize(pinstance, width, height);
#endif
    spDebug(50, "spGetClientSizePrinter", "width = %d, height = %d\n", *width, *height);

    return SP_TRUE;
}

static spBool spGetPhysicalSizePrinter(void *instance, double *width_mm, double *height_mm)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    *width_mm = pinstance->paper_width_mm - pinstance->left_margin_mm - pinstance->right_margin_mm;
    *height_mm = pinstance->paper_height_mm - pinstance->top_margin_mm - pinstance->bottom_margin_mm;
    
    return SP_TRUE;
}

static spBool spSetScalingFactorPrinter(void *instance, double xfactor, double yfactor)
{
    spBool ok = SP_FALSE;
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

#ifdef SP_MAC_USE_PRINTING_MANAGER
    {
	double hdpi, vdpi;
	PMResolution res;

	hdpi = pinstance->default_hdpi;
	vdpi = pinstance->default_vdpi;

	spDebug(50, "spSetScalingFactorPrinter", "current dpi: %f, %f\n", hdpi, vdpi);
	res.hRes = hdpi / xfactor;
	res.vRes = vdpi / yfactor;
	spDebug(50, "spSetScalingFactorPrinter", "new dpi: %f, %f\n", res.hRes, res.vRes);

	if (PMSetResolution(pinstance->page_format, &res) == noErr) {
	    ok = SP_TRUE;
	    if (PMGetResolution(pinstance->page_format, &res) == noErr) {
		spDebug(50, "spSetScalingFactorPrinter", "updated dpi: %f, %f\n", res.hRes, res.vRes);
		xfactor = hdpi / res.hRes;
		yfactor = vdpi / res.vRes;
	    }
	}
    }
#else
    {
	int hdpi, vdpi;
	TSetRslBlk setrslblk;

	hdpi = (int)spRound((double)getHDPI(pinstance->hprint) / xfactor);
	vdpi = (int)spRound((double)getVDPI(pinstance->hprint) / yfactor);
	spDebug(50, "spSetScalingFactorPrinter", "new dpi: %d, %d\n", hdpi, vdpi);

	if (hdpi >= pinstance->x_min_dpi && hdpi <= pinstance->x_max_dpi
	    && vdpi >= pinstance->y_min_dpi && vdpi <= pinstance->y_max_dpi) {
	    setrslblk.iOpCode = setRslOp;
	    setrslblk.hPrint = pinstance->hprint;
	    setrslblk.iXRsl = hdpi;
	    setrslblk.iYRsl = vdpi;
	    PrGeneral((Ptr)&setrslblk);
	    if (PrError() == noErr && setrslblk.iError == noErr) {
		ok = SP_TRUE;
	    }
	}
    }
#endif

    if (ok == SP_TRUE) {
	pinstance->xfactor = xfactor;
	pinstance->yfactor = yfactor;
	spDebug(50, "spSetScalingFactorPrinter", "ok: xfactor = %f, yfactor = %f\n", xfactor, yfactor);
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static void spSetPortPrinter(spPluginInstancePrinter pinstance)
{
#ifdef SP_MAC_USE_PRINTING_MANAGER
    GetPort(&pinstance->save_port); SetPort(pinstance->pport);
#endif
    return;
}

static void spUnsetPortPrinter(spPluginInstancePrinter pinstance)
{
#ifdef SP_MAC_USE_PRINTING_MANAGER
    SetPort(pinstance->save_port);
#endif
    return;
}

static spBool spStartDrawingPrinter(void *instance, const char *job_name)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    pinstance->current_page = 0;
    pinstance->save_port = NULL;
    pinstance->pport = NULL;
    
#ifdef SP_MAC_USE_PRINTING_MANAGER
    {
	spBool ok = SP_FALSE;

#if defined(MACOSX)
	{
	    CFStringRef cfstr;
	    
	    cfstr = CFStringCreateWithCString(kCFAllocatorDefault, job_name, kCFStringEncodingUTF8);
	    
	    if (PMSetJobNameCFString(pinstance->settings, cfstr) != noErr) {
	        CFRelease(cfstr);
		return SP_FALSE;
	    }
	    CFRelease(cfstr);
	}
#else
	{
	    Str255 pstr;

	    spStrCToP(job_name, pstr);
	    
	    if (PMSetJobName(pinstance->settings, pstr) != noErr) {
		return SP_FALSE;
	    }
	}
#endif
        
	if (PMSessionBeginDocument(pinstance->session, pinstance->settings,
				   pinstance->page_format) == noErr) {
	    if (PMSessionBeginPage(pinstance->session, pinstance->page_format, NULL) == noErr) {
		if (PMSessionGetGraphicsContext(pinstance->session, kPMGraphicsContextQuickdraw,
						(void**)&pinstance->pport) == noErr) {
#if 0
		    GetPort(&pinstance->save_port);
		    SetPort(pinstance->pport);
#else
		    pinstance->save_port = NULL;
#endif
		    ok = SP_TRUE;
		}
		
		if (ok == SP_FALSE) {
		    PMSessionEndPage(pinstance->session);
		}
	    }

	    if (ok == SP_FALSE) {
		PMSessionEndDocument(pinstance->session);
	    }
	}

	if (ok == SP_FALSE) {
	    return SP_FALSE;
	}
    }
#else
    pinstance->pport = PrOpenDoc(pinstance->hprint, NULL, NULL);
    if (pinstance->pport == NULL || PrError() != noErr) {
	return SP_FALSE;
    }

    /* Since scaling doesn't work, the followin code was invalidated */
    if (0 && (pinstance->xfactor != 1.0 || pinstance->yfactor != 1.0)) { 
	Rect rect;

	getClientRect(pinstance, &rect);
	
	PrOpenPage(pinstance->pport, &rect);
    } else {
	PrOpenPage(pinstance->pport, NULL);
    }
    
    if (PrError() != noErr) {
	PrCloseDoc(pinstance->pport);
	pinstance->pport = NULL;
	return SP_FALSE;
    }
#endif

    pinstance->current_page = 1;
    
    return SP_TRUE;
}

static spBool spEndDrawingPrinter(void *instance)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    if (pinstance->pport == NULL) {
	return SP_FALSE;
    }
    
#ifdef SP_MAC_USE_PRINTING_MANAGER
    PMSessionEndPage(pinstance->session);
    PMSessionEndDocument(pinstance->session);

#if 0
    if (pinstance->save_port != NULL) {
	SetPort(pinstance->save_port);
	pinstance->save_port = NULL;
    }
#else
    pinstance->save_port = NULL;
#endif
#else
    if (pinstance->current_page >= 1) {
	PrClosePage(pinstance->pport);
    }
    PrCloseDoc(pinstance->pport);
    
    spDebug(50, "spEndDrawingPrinter", "bJDocLoop = %d\n", (*pinstance->hprint)->prJob.bJDocLoop);
    
    if (PrError() != noErr
	&& (*pinstance->hprint)->prJob.bJDocLoop == bSpoolLoop) {
	TPrStatus prstatus;
	PrPicFile(pinstance->hprint, NULL, NULL, NULL, &prstatus);
	spDebug(50, "spEndDrawingPrinter", "deferred printing\n");
    }
#endif
    
    return SP_TRUE;
}

static spBool spForwardPagePrinter(void *instance)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    if (pinstance->pport == NULL || pinstance->current_page < 1) return SP_FALSE;
    
    spDebug(50, "spForwardPagePrinter", "pinstance->current_page = %d\n", pinstance->current_page);
    
#ifdef SP_MAC_USE_PRINTING_MANAGER
    if (PMSessionEndPage(pinstance->session) != noErr
	|| PMSessionBeginPage(pinstance->session, pinstance->page_format, NULL) != noErr) {
	return SP_FALSE;
    }
    if (PMSessionGetGraphicsContext(pinstance->session, kPMGraphicsContextQuickdraw,
				    (void**)&pinstance->pport) != noErr) {
	PMSessionEndPage(pinstance->session);
	return SP_FALSE;
    }

#if 0
    SetPort(pinstance->pport);
#endif
#else
    PrClosePage(pinstance->pport);
    
    if (0 && (pinstance->xfactor != 1.0 || pinstance->yfactor != 1.0)) {
	Rect rect;

	getClientRect(pinstance, &rect);
	
	PrOpenPage(pinstance->pport, &rect);
    } else {
	PrOpenPage(pinstance->pport, NULL);
    }
#endif

    pinstance->current_page++;
    
    return SP_TRUE;
}

static spBool spCopyImagePrinter(void *instance, void *src_image_arch,
				 int src_x, int src_y, int src_width, int src_height,
				 int dest_x, int dest_y, int dest_width, int dest_height)
{
    PixMapHandle src_pixmap;
    BitMap *dest_bitmap_ptr;
    int width, height;
    Rect src_rect, dest_rect;
    GWorldPtr src_gworld = (GWorldPtr)src_image_arch;
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    if (pinstance->pport == NULL) return SP_FALSE;
    
    if ((src_pixmap = GetGWorldPixMap(src_gworld)) == nil || GetPixBaseAddr(src_pixmap) == NULL) {
	return SP_FALSE;
    }

    spSetPortPrinter(pinstance);
    
    width = MIN(src_width, dest_width);
    height = MIN(src_height, dest_height);
    
    src_rect.left = src_x;
    src_rect.top = src_y;
    src_rect.right = src_x + width;
    src_rect.bottom = src_y + height;

    dest_rect.left = dest_x;
    dest_rect.top = dest_y;
    dest_rect.right = dest_x + width;
    dest_rect.bottom = dest_y + height;
    
    /* set normal RGB */
    spGetOriginalRGBMac();
    spSetNormalRGBMac();

#ifdef SP_MAC_USE_PRINTING_MANAGER
#if TARGET_API_MAC_CARBON
    dest_bitmap_ptr = (BitMap *)GetPortBitMapForCopyBits((CGrafPtr)pinstance->pport);
#else
    dest_bitmap_ptr = &pinstance->pport->portBits;
#endif
#else
    dest_bitmap_ptr = &pinstance->pport->gPort.portBits;
#endif
    
    /* copy pixmap */
    CopyBits((BitMap *)*src_pixmap, dest_bitmap_ptr,
	     &src_rect, &dest_rect, srcCopy, NULL);
    
    /* restore RGB */
    spSetOriginalRGBMac();
    
    spUnsetPortPrinter(pinstance);
    
    return SP_TRUE;
}

void spFillRectanglePrinter(void *instance, spGraphics graphics,
			    int x, int y, int width, int height)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    if (pinstance->pport == NULL) return;
    
    spSetPortPrinter(pinstance);
    
    spSetGraphicsMac(graphics);
    spDrawRectangleMac(graphics, x, y, width, height, SP_TRUE);
    spUnsetGraphicsMac(graphics);

    spUnsetPortPrinter(pinstance);
    
    return;
}

void spDrawRectanglePrinter(void *instance, spGraphics graphics,
			    int x, int y, int width, int height)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    if (pinstance->pport == NULL) return;
    
    spSetPortPrinter(pinstance);
    
    spSetGraphicsMac(graphics);
    spDrawRectangleMac(graphics, x, y, width, height, SP_FALSE);
    spUnsetGraphicsMac(graphics);

    spUnsetPortPrinter(pinstance);
    
    return;
}

void spFillArcPrinter(void *instance, spGraphics graphics,
		      int x, int y, int width, int height, double angle1, double angle2)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    
    if (pinstance->pport == NULL) return;
    
    spSetPortPrinter(pinstance);
    
    spSetGraphicsMac(graphics);
    spDrawArcMac(graphics, x, y, width, height, angle1, angle2, SP_TRUE);
    spUnsetGraphicsMac(graphics);

    spUnsetPortPrinter(pinstance);
    
    return;
}
     
void spDrawArcPrinter(void *instance, spGraphics graphics,
		      int x, int y, int width, int height, double angle1, double angle2)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    
    if (pinstance->pport == NULL) return;
    
    spSetPortPrinter(pinstance);
    
    spSetGraphicsMac(graphics);
    spDrawArcMac(graphics, x, y, width, height, angle1, angle2, SP_FALSE);
    spUnsetGraphicsMac(graphics);

    spUnsetPortPrinter(pinstance);
    
    return;
}

void spDrawLinePrinter(void *instance, spGraphics graphics,
		       int x1, int y1, int x2, int y2)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    
    if (pinstance->pport == NULL) return;
    
    spSetPortPrinter(pinstance);
    
    spSetGraphicsMac(graphics);
    spDrawLineMac(graphics, x1, y1, x2, y2);
    spUnsetGraphicsMac(graphics);

    spUnsetPortPrinter(pinstance);
    
    return;
}

void spDrawPointPrinter(void *instance, spGraphics graphics,
			int x, int y)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    if (pinstance->pport == NULL) return;
    
    spSetPortPrinter(pinstance);
    
    spSetGraphicsMac(graphics);
    spDrawPointMac(graphics, x, y);
    spUnsetGraphicsMac(graphics);

    spUnsetPortPrinter(pinstance);
    
    return;
}

void spDrawStringPrinter(void *instance, spGraphics graphics,
			 int x, int y, const char *string)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    if (pinstance->pport == NULL) return;
    
    spSetPortPrinter(pinstance);
    
    spSetGraphicsMac(graphics);
    spDrawStringMac(graphics, x, y, string);
    spUnsetGraphicsMac(graphics);

    spUnsetPortPrinter(pinstance);
    
    return;
}

spBool spGetStringExtentPrinter(void *instance, spGraphics graphics, const char *string,
				int *x, int *y, int *width, int *height, int *next_x)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    
    if (pinstance->pport == NULL) return SP_FALSE;
    
    spSetPortPrinter(pinstance);
    
    spGetStringExtentMac(graphics, string, x, y, width, height, next_x);

    spUnsetPortPrinter(pinstance);
    
    return SP_TRUE;
}

spBool spFillPolygonPrinter(void *instance, spGraphics graphics, void *polygon_instance)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    spPolygon polygon = (spPolygon)polygon_instance;
    
    if (pinstance->pport == NULL) return SP_FALSE;
    
    spSetPortPrinter(pinstance);
    
    spSetGraphicsMac(graphics);
    PaintPoly(SpPolygonArch(polygon).hpoly);
    spUnsetGraphicsMac(graphics);
    
    spUnsetPortPrinter(pinstance);
    
    return SP_TRUE;
}

spBool spDrawPolygonPrinter(void *instance, spGraphics graphics, void *polygon_instance)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    spPolygon polygon = (spPolygon)polygon_instance;
    
    if (pinstance->pport == NULL) return SP_FALSE;
    
    spSetPortPrinter(pinstance);
    
    spSetGraphicsMac(graphics);
    FramePoly(SpPolygonArch(polygon).hpoly);
    spUnsetGraphicsMac(graphics);
    
    spUnsetPortPrinter(pinstance);
    
    return SP_TRUE;
}

static spDrawingPluginRec sp_drawing_plugin_printer = {
    NULL,
    NULL,

    SP_PLUGIN_DRAWING,
    SP_DRAWING_PLUGIN_PRINTER,
    1,
    SP_PLUGIN_PRIORITY_MIDDLE,
    SP_DRAWING_PLUGIN_CAPS_MULTIPLE_PAGES,		/* caps */
    spInitPrinter,
    spFreePrinter,
    "Printer",
    "Printer Plugin for Mac  Version 0.1\nCopyright(C) Hideki Banno",
    
    spInitInstancePrinter,
    spFreeInstancePrinter,
    NULL,
    NULL,

    SP_DRAWING_DEVICE_PRINTER,

    NULL,
    NULL,
    NULL,
    NULL,

    NULL,
    spCreateComponentPrinter,
    spDestroyComponentPrinter,
    NULL/*spSetParamsPrinter*/,
    NULL/*spGetParamsPrinter*/,

    spOpenSetupDialogPrinter,
    
    spSetPageInfoPrinter,
    spGetPageInfoPrinter,
    spSetPageRangePrinter,
    spGetPageRangePrinter,
    NULL,

    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    
    NULL,
    spOpenTargetDialogPrinter,

    spCloseTargetPrinter,
    
    spIsDrawablePrinter,

    spGetDrawingCapsPrinter,
    spGetGraphicsModeCapsPrinter,
    NULL,
    spGetClientSizePrinter,
    spGetPhysicalSizePrinter,
    spSetScalingFactorPrinter,

    spStartDrawingPrinter,
    spEndDrawingPrinter,
    spForwardPagePrinter,

    NULL,
    spCopyImagePrinter,

    spFillRectanglePrinter,
    spDrawRectanglePrinter,
    spFillArcPrinter,
    spDrawArcPrinter,
    spDrawLinePrinter,
    spDrawPointPrinter,
    spDrawStringPrinter,
    spGetStringExtentPrinter,

    NULL,
    NULL,
    NULL,
    
    NULL,
    spFillPolygonPrinter,
    spDrawPolygonPrinter,
};

spBool spAddPrinterPlugin(void)
{
    return spAddPluginRecList((spPluginRec *)&sp_drawing_plugin_printer, SP_DRAWING_PLUGIN_PRINTER);
}

spBool spRemovePrinterPlugin(void)
{
    return spRemovePluginRecList((spPluginRec *)&sp_drawing_plugin_printer);
}
