/*
 *	spDraw_Win.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spPrimitiveP.h>
#include <sp/spCanvasP.h>
#include <sp/spDrawP.h>
#include <sp/spImageP.h>
#include <sp/spGraphicsP.h>

extern spTopLevel sp_toplevel;

void spImageCreateArch(spComponent component)
{
    SpPrimitiveArch(component).hwnd = SpParentPrimitiveArch(component).hwnd;

    return;
}

spBool spImageLoadArch(spComponent component, const char *filename)
{
    BITMAP bitmap;
    char filenamebuf[SP_MAX_PATHNAME];
    
    strcpy(filenamebuf, filename);
    if (spGetSuffix(filenamebuf) == NULL || spIsFile(filenamebuf) == SP_FALSE) {
	spStrCat(filenamebuf, sizeof(filenamebuf), ".bmp");
    }
    spDebug(50, "spImageLoadArch", "filename = %s, filenamebuf = %s\n", filename, filenamebuf);
    
    SpPrimitiveArch(component).hbitmap =
	(HBITMAP)LoadImage(NULL, filenamebuf, IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE);
    if (SpPrimitiveArch(component).hbitmap == NULL) {
	spDebug(50, "spImageLoadArch", "cannot load bitmap file: %s\n", filenamebuf);
	return SP_FALSE;
    }

    GetObject(SpPrimitiveArch(component).hbitmap, sizeof(BITMAP), &bitmap);
    spDebug(50, "spImageLoadArch", "bitmap size: %ld x %ld\n", bitmap.bmWidth, bitmap.bmHeight);
    SpComponentPart(component).width = bitmap.bmWidth;
    SpComponentPart(component).height = bitmap.bmHeight;

    /* Don't set client size. */
#if 0				
    SpComponentPart(component).client_width = bitmap.bmWidth;
    SpComponentPart(component).client_height = bitmap.bmHeight;
#endif
    
    return SP_TRUE;
}

void spDrawImageArch(spComponent component)
{
    spDebug(100, "spDrawImageArch", "fill background: client_width = %d, client_height = %d\n",
	    SpComponentPart(component).client_width, SpComponentPart(component).client_height);
#if 0
    PatBlt(SpPrimitiveArch(component).memdc, 0, 0,
	   SpComponentPart(component).client_width, SpComponentPart(component).client_height,
	   WHITENESS);
#else
    /* fill background */
    SelectObject(SpPrimitiveArch(component).memdc, SpTopLevelArch(sp_toplevel).bg_brush);
    Rectangle(SpPrimitiveArch(component).memdc, -1, -1,
	      SpComponentPart(component).client_width + 1,
	      SpComponentPart(component).client_height + 1);
    SelectObject(SpPrimitiveArch(component).memdc, SpTopLevelArch(sp_toplevel).null_brush);
#endif
    
    return;
}

void spRedrawImageArch(spComponent component)
{
    int width, height;

    if (spIsSubClass(component, SpImage) == SP_FALSE
	|| strnone(SpComponentPart(component).description)) {
	width = SpComponentPart(component).client_width;
	height = SpComponentPart(component).client_height;
    } else {
	width = height = 0;
    }
    
    spUpdateCompatibleBitmapWin(component, width, height);
    
    return;
}

