/*
 *	spMenu_Gtk.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sp/spBase.h>
#include <sp/spOption.h>

#include <sp/spCanvas.h>

#include <sp/spTopLevelP.h>
#include <sp/spFrameP.h>
#include <sp/spButtonP.h>
#include <sp/spMenuItemP.h>
#include <sp/spMenuP.h>

void spPulldownMenuCreateArch(spComponent component)
{
    spComponent window;
    static char label[SP_MAX_MENU_LABEL] = "";
    
    window = SpGetWindow(component);
    
    if (SpFrameArch(window).accel_group == NULL) {
#if GTK_CHECK_VERSION(1,2,0)
	SpFrameArch(window).accel_group = gtk_accel_group_new();
	gtk_window_add_accel_group(GTK_WINDOW(SpFrameArch(window).toplevel),
				   SpFrameArch(window).accel_group);
#else
	SpFrameArch(window).accel_group = gtk_accelerator_table_new();
	gtk_window_add_accelerator_table(GTK_WINDOW(SpFrameArch(window).toplevel),
					 SpFrameArch(window).accel_group);
#endif
	spDebug(40, "spPulldownMenuCreateArch", "accel group added\n");
    }

    strcpy(label, "");
    if (!strnone(SpComponentPart(component).title)) {
	spGetMnemonicGtk(SpComponentPart(component).title, label);
    }

    if (strnone(label)) {
	strcpy(label, (!strnone(SpGetName(component)) ? SpGetName(component) : ""));
    }
    
    SpPrimitiveArch(component).widget = gtk_menu_new();

    if (spIsSubClass(component, SpPopupMenu) == SP_TRUE) {
	GtkWidget *parent;

	if (spIsLabel(SpGetParent(component)) == SP_TRUE) {
	    parent = SpParentPrimitiveArch(component).top_widget;
	} else {
	    parent = SpParentPrimitiveArch(component).widget;
	}
	gtk_menu_attach_to_widget(GTK_MENU(SpPrimitiveArch(component).widget),
				  parent, NULL);
    } else {
#if GTK_CHECK_VERSION(2,0,0)
	SpPrimitiveArch(component).sub_widget = gtk_menu_item_new_with_mnemonic(label);
#else
	SpPrimitiveArch(component).sub_widget = gtk_menu_item_new_with_label(label);
#endif
	gtk_widget_show(SpPrimitiveArch(component).sub_widget);

	gtk_menu_item_set_submenu(GTK_MENU_ITEM(SpPrimitiveArch(component).sub_widget),
				  SpPrimitiveArch(component).widget);

#if GTK_CHECK_VERSION(3,0,0)
	gtk_menu_shell_append(GTK_MENU_SHELL(SpParentPrimitiveArch(component).widget),
			      SpPrimitiveArch(component).sub_widget);
#else
	if (spIsSubClass(SpGetParent(component), SpMenuBar) == SP_TRUE) {
	    gtk_menu_bar_append(GTK_MENU_BAR(SpParentPrimitiveArch(component).widget),
				SpPrimitiveArch(component).sub_widget);
	} else if (spIsSubClass(SpGetParent(component), SpMenu) == SP_TRUE) {
	    gtk_menu_append(GTK_MENU(SpParentPrimitiveArch(component).widget),
			    SpPrimitiveArch(component).sub_widget);
	}
#endif
	
	spSignalConnectEventGtk(SpPrimitiveArch(component).sub_widget,
				"enter_notify_event", spShowMenuHelpCB,
				(gpointer)component);
	spSignalConnectEventGtk(SpPrimitiveArch(component).sub_widget,
				"leave_notify_event", spShowMenuHelpCB,
				(gpointer)component);
    }

    gtk_menu_set_accel_group(GTK_MENU(SpPrimitiveArch(component).widget),
			     SpFrameArch(window).accel_group);
    
    spDebug(30, "spPulldownMenuCreateArch", "done\n");

    return;
}

void spPulldownMenuSetParamsArch(spComponent component)
{
    static char label[SP_MAX_BUTTON_LABEL] = "";

    if (SpComponentPart(SpOldObject(component)).title != SpComponentPart(component).title) {    
	strcpy(label, "");
	if (!strnone(SpComponentPart(component).title)) {
	    spGetMnemonicGtk(SpComponentPart(component).title, label);
	}
	
	if (!strnone(label)) {
	    spUpdateWidgetTitleGtk(SpPrimitiveArch(component).sub_widget, label);
	}
    }
    
    spDebug(40, "spPulldownMenuSetParamsArch", "done\n");

    return;
}

#if !GTK_CHECK_VERSION(2,0,0)
static void canvasButtonPress(spComponent component, GdkEvent *event)
{
    long k;
    long component_id;
    spComponent canvas;
    spCallback *callback;
    
    spDebug(40, "canvasButtonPress", "in\n");
    canvas = SpGetParent(component);
    spDebug(40, "canvasButtonPress", "canvas = %ld\n", (long)canvas);
    
    if (spIsCanvas(canvas) == SP_TRUE) {
	spDebug(40, "canvasButtonPress", "num_callback = %d\n", SpPrimitivePart(canvas).num_callback);
	
	for (k = 0; k < SpPrimitivePart(canvas).num_callback; k++) {
	    if (streq(SpPrimitivePart(canvas).callbacks[k].call_name,
		      "button_press_event")) {
		callback = &SpPrimitivePart(canvas).callbacks[k];
			
		component_id = SpGetComponentId(callback->component);
		SpPrimitiveArch(callback->component).call_name = callback->call_name;
		SpPrimitiveArch(callback->component).event = event;
			
		spComponentCallbackFunc(callback);
			
		if (component_id != spGetDestroyedComponentId()
		    && callback != NULL && callback->component != NULL) {
		    SpPrimitiveArch(callback->component).call_name = NULL;
		    SpPrimitiveArch(callback->component).event = NULL;
		}
	    }
	}
    }

    spDebug(40, "canvasButtonPress", "done\n");
    
    return;
}
#endif

static gint popupMenuCB(GtkWidget *widget, GdkEvent *event)
{
    spComponent component;
    guint popup_button = 3;

    spDebug(80, "popupMenuCB", "event->type = %d\n", event->type);
    
     if (event->type == GDK_BUTTON_PRESS) {
        GdkEventButton *bevent = (GdkEventButton *)event;

#if GTK_CHECK_VERSION(3,0,0)
	component = (spComponent)g_object_get_data(G_OBJECT(widget), "value");
#else
	component = (spComponent)gtk_object_get_data(GTK_OBJECT(widget), "value");
#endif
	
	if (component != NULL) {
	    if (SpMenuPart(component).popup_button == SP_LBUTTON) {
		popup_button = 1;
	    } else if (SpMenuPart(component).popup_button == SP_RBUTTON) {
		popup_button = 3;
	    } else if (SpMenuPart(component).popup_button == SP_MBUTTON) {
		popup_button = 2;
	    }
	}

	spDebug(80, "popupMenuCB", "component = %ld, popup_button = %d, bevent->button = %d\n",
		(long)component, popup_button, bevent->button);
	
	if (component == NULL) return FALSE;

#if !GTK_CHECK_VERSION(2,0,0)
	/* call canvas callback because signal is stopped by menu popup */
	canvasButtonPress(component, event);
#endif
	
	if (bevent->button != popup_button) return FALSE;

        gtk_menu_popup(GTK_MENU(widget), NULL, NULL, NULL, NULL,
		       bevent->button, bevent->time);
	
#if !GTK_CHECK_VERSION(2,0,0)
        return TRUE;
#else
	return FALSE;
#endif
    }

     spDebug(80, "popupMenuCB", "return FALSE\n");
     
    return FALSE;
}

#if GTK_CHECK_VERSION(2,10,0)
static void popupStatusIconMenu(GtkStatusIcon *status_icon,
				guint button,
				guint activate_time,
				gpointer user_data)
{
    spComponent component = (spComponent)user_data;

    spDebug(80, "popupStatusIconMenu", "button = %d, activate_time = %d\n",
	    button, activate_time);
    
    gtk_menu_popup(GTK_MENU(SpPrimitiveArch(component).widget),
		   NULL, NULL, gtk_status_icon_position_menu, status_icon, button, activate_time);
    
    return;
}

static void activateStatusIcon(GtkStatusIcon* status_icon,
			       gpointer user_data)
{
    spComponent component = (spComponent)user_data;
    
    spDebug(80, "activateStatusIcon", "in\n");
    
    gtk_menu_popup(GTK_MENU(SpPrimitiveArch(component).widget),
		   NULL, NULL, gtk_status_icon_position_menu, status_icon,
		   0, gtk_get_current_event_time());

    return;
}
#endif

void spPopupMenuCreateArch(spComponent component)
{
    spComponent parent;
    GtkWidget *widget = NULL;
    GtkWidget *parent_widget = NULL;
    
    spPulldownMenuCreateArch(component);

    if (SpMenuPart(component).task_tray_flag == SP_TRUE) {
#if GTK_CHECK_VERSION(2,10,0)
	spComponent window;
	
	window = SpGetWindow(component);
	
	if (spIsTaskTrayIconCreatedGtk(window) == SP_FALSE) {
	    spCreateTaskTrayIconGtk(window);
	}

#if defined(HAVE_APP_INDICATOR)
	if (SpFrameArch(window).indicator != NULL) {
	    spDebug(80, "spPopupMenuCreateArch", "set menu to indicator\n");
	    app_indicator_set_menu(SpFrameArch(window).indicator, GTK_MENU(SpPrimitiveArch(component).widget));
	}
	else 
#endif
	if (SpFrameArch(window).status_icon != NULL) {
	    spDebug(80, "spPopupMenuCreateArch", "g_signal_connect of popup-menu\n");
	    g_signal_connect(G_OBJECT(SpFrameArch(window).status_icon), "popup-menu",
			     G_CALLBACK(popupStatusIconMenu), component);
	    g_signal_connect(G_OBJECT(SpFrameArch(window).status_icon), "activate",
			     G_CALLBACK(activateStatusIcon), component);
	}
#endif
    } else {
	parent = SpGetParent(component);
    
	if (spIsSubClass(parent, SpLabel) == SP_TRUE) {
	    parent_widget = SpPrimitiveArch(parent).top_widget;
	} else if (SpPrimitiveArch(parent).widget != NULL) {
	    parent_widget = SpPrimitiveArch(parent).widget;
	}
	widget = SpPrimitiveArch(component).widget;

	if (parent_widget != NULL) {
	    spDebug(80, "spPopupMenuCreateArch", "add callback function of popupMenuCB\n");
#if GTK_CHECK_VERSION(3,0,0)
	    g_object_set_data(G_OBJECT(widget),
			      "value", (gpointer)component);
	    g_signal_connect_object(G_OBJECT(parent_widget),
				    "button_press_event", G_CALLBACK(popupMenuCB),
				    G_OBJECT(widget), G_CONNECT_SWAPPED);
#else
	    gtk_object_set_data(GTK_OBJECT(widget),
				"value", (gpointer)component);
	    gtk_signal_connect_object(GTK_OBJECT(parent_widget),
				      "button_press_event", GTK_SIGNAL_FUNC(popupMenuCB),
				      GTK_OBJECT(widget));
#endif
	}
    }
 
    spDebug(80, "spPopupMenuCreateArch", "done\n");
    
    return;
}

void spPopupMenuSetParamsArch(spComponent component)
{
    spPulldownMenuSetParamsArch(component);
    return;
}

void spMenuBarCreateArch(spComponent component)
{
    if (spIsCreated(component) == SP_FALSE
	&& spIsWindow(SpGetParent(component)) == SP_TRUE) {
	/* create menu bar */
	SpPrimitiveArch(component).widget = gtk_menu_bar_new();
	gtk_box_pack_start(GTK_BOX(SpParentPrimitiveArch(component).top_widget),
			   SpPrimitiveArch(component).widget, FALSE, FALSE, 0);
	gtk_widget_show(SpPrimitiveArch(component).widget);
	
	spDebug(30, "spMenuBarCreateArch", "created\n");
    }

    return;
}
