/*
 *	spTabBox.c
 */

#include <stdio.h>
#include <stdlib.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spContainer.h>
#include <sp/spList.h>

#include <sp/spContainerP.h>
#include <sp/spTabBoxP.h>

static spParamTable sp_tab_box_param_tables[] = {
    {SppSelectedTabIndex, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spTabBox, tab_box.selected_index), "0"},
    {SppSize, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spTabBox, tab_box.size), "0"},
};

spTabBoxClassRec SpTabBoxClassRec = {
    /* spObjectClassPart */
    {
	SpTabBox,
	(spObjectClass)&SpPrimitiveClassRec,
	sizeof(spTabBoxRec),
	spArraySize(sp_tab_box_param_tables),
	sp_tab_box_param_tables,
	spTabBoxPartInit,
	spTabBoxPartFree,
	SP_FALSE,
	NULL,
	NULL,
	spTabBoxCreate,
	spTabBoxDestroy,
	spTabBoxSetParams,
	NULL,
    },
    /* spComponentClassPart */
    {
	SP_TRUE,
	SP_FALSE,
	SP_TRUE,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,

	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
    },
    /* spPrimitiveClassPart */
    {
	0,
    },
    /* spTabBoxClassPart */
    {
	0,
    },
};

spComponentClass SpTabBoxClass = (spComponentClass)&SpTabBoxClassRec;

void spTabBoxPartInit(spObject object)
{
    spComponent component = (spComponent)object;
    
    spTabBoxPartInitArch(component);
    
    SpComponentPart(component).orientation = SP_VERTICAL;
    SpComponentPart(component).margin_width = SP_DEFAULT_MARGIN;
    SpComponentPart(component).margin_height = SP_DEFAULT_MARGIN;
    SpComponentPart(component).spacing = SP_DEFAULT_SPACING;
    
    SpTabBoxPart(component).num_item = 0;

    return;
}

void spTabBoxPartFree(spObject object)
{
    spComponent component = (spComponent)object;
    
    spTabBoxPartFreeArch(component);
    return;
}

spBool spTabBoxCreate(spObject object)
{
    spComponent component = (spComponent)object;
    
    if (component == NULL) return SP_FALSE;

    spTabBoxSetDefaultSize(component);
    if (SpComponentPart(component).disable_initial_scaling == SP_FALSE)
	spPrimitiveScaleDefaultSizeArch(component);
    spTabBoxCreateArch(component);
	
    return SP_TRUE;
}

spBool spTabBoxSetParams(spObject object)
{
    spComponent component = (spComponent)object;
    
    if (component == NULL) return SP_FALSE;

    spTabBoxSetParamsArch(component);
	
    return SP_TRUE;
}

spBool spTabBoxDestroy(spObject object)
{
    spComponent component = (spComponent)object;
    
    spTabBoxDestroyArch(component);
    spPrimitiveDestroyArch(component);
    
    return SP_TRUE;
}

void spTabBoxSetDefaultSize(spComponent component)
{
    if (spIsTabBox(component) == SP_FALSE) return;
    
    SpComponentPart(component).x = 0;
    SpComponentPart(component).y = 0;
    if (SpParentComponentPart(component).orientation == SP_HORIZONTAL) {
	SpComponentPart(component).width = SpTabBoxPart(component).size;
	SpComponentPart(component).height = 0;
    } else {
	SpComponentPart(component).width = 0;
	SpComponentPart(component).height = SpTabBoxPart(component).size;
    }
    SpComponentPart(component).current_width = SpComponentPart(component).width;
    SpComponentPart(component).current_height = SpComponentPart(component).height;

    return;
}

spBool spIsTabBox(spComponent component)
{
    return spIsSubClass(component, SpTabBox);
}

spComponent spCreateTabBox(spComponent parent, const char *name, int size, ...)
{
    int num_arg = 0;
    spArg args[SP_MAX_NUM_ARG];
    va_list argp;
    
    if (spIsCreated(parent) == SP_FALSE || spIsContainer(parent) == SP_FALSE)
	return NULL;
    
    va_start(argp, size);
    spGetArgs(argp, args, num_arg);
    va_end(argp);

    spSetArg(args[num_arg], SppSize, size); num_arg++;

    return spCreateComponentArg(SpTabBoxClass, NULL, name, parent, args, num_arg);
}

spBool spIsTabItem(spComponent component)
{
    if (spIsSubClass(component, SpBox) == SP_TRUE
	&& spIsTabBox(SpGetParent(component)) == SP_TRUE) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

int spMapTabItemDefault(spComponent component, int index)
{
    spComponent child;
    spComponent selected_tab = NULL;
    
    child = SpGetChild(component);

    while (child != NULL) {
	if (SpPrimitivePart(child).index == index) {
	    selected_tab = child;

	    child = SpGetChild(component);
	    while (child != NULL) {
		spDebug(30, "spMapTabItemDefault", "child index = %d\n", SpPrimitivePart(child).index);
		SpComponentPart(child).geometry_flag = SP_TRUE;
		if (child != selected_tab) {
		    spDebug(30, "spMapTabItemDefault", "call spUnmapComponent\n");
		    spUnmapComponent(child);
		} else {
		    SpComponentPart(child).visible_flag = SP_TRUE;
		}
		SpComponentPart(child).geometry_flag = SP_FALSE;
		
		child = SpGetNextComponent(child);
	    }

	    spAdjustComponentSize(component);
	    SpComponentPart(selected_tab).visible_flag = SP_FALSE;
	    
	    SpComponentPart(selected_tab).geometry_flag = SP_TRUE;
	    spMapComponent(selected_tab);
	    SpComponentPart(selected_tab).geometry_flag = SP_FALSE;

	    /* The size adjustment in spMapComonent will not work
	       because geometry_flag is SP_TRUE. So, adjust size again. */
	    spAdjustComponentSize(selected_tab);
	    
	    spDebug(30, "spMapTabItemDefault", "done: index = %d\n", index);
	    return index;
	}

	child = SpGetNextComponent(child);
    }
    
    return -1;
}

int spMapTabItem(spComponent component, int index)
{
    if (spIsTabBox(component) == SP_FALSE || index < 0
	|| spIsCreated(component) == SP_FALSE) return -1;

    spDebug(30, "spMapTabItem", "in: index = %d\n", index);

    return spMapTabItemArch(component, index);
}

int spGetSelectedTabIndex(spComponent component)
{
    if (spIsTabBox(component) == SP_FALSE) return -1;

    if (spIsCreated(component) == SP_TRUE) {
	SpTabBoxPart(component).selected_index = spGetSelectedTabIndexArch(component);
    }
    
    spDebug(10, "spGetSelectedTabIndex", "index = %d\n",
	    SpTabBoxPart(component).selected_index);
    
    return SpTabBoxPart(component).selected_index;
}

spComponent spGetSelectedTabItem(spComponent component)
{
    int index;
    spComponent child;
    
    if (spIsTabBox(component) == SP_FALSE 
	|| spIsCreated(component) == SP_FALSE) return NULL;

    if ((index = spGetSelectedTabIndex(component)) >= 0) {
	child = SpGetChild(component);
	
	while (child != NULL) {
	    if (SpPrimitivePart(child).index == index) {
		return child;
	    }

	    child = SpGetNextComponent(child);
	}
    }
    
    return NULL;
}

spComponent spCreateTabItemArg(spComponent parent, const char *name, spArg *args, int num_arg)
{
    int index;
    spComponent component;
    spComponent child;
    
    component = spCreateComponentArg(SpContainerClass, SpBox, name, parent, args, num_arg);

    if (component != NULL) {
        index = SpPrimitivePart(component).index;

        child = SpGetChild(parent);
        while (child != NULL) {
            if (child == component) {
            } else if (SpPrimitivePart(child).index >= index) {
                SpPrimitivePart(child).index++;
            }

            child = SpGetNextComponent(child);
        }
    
        spAddTabItemArch(component, index);
    }
    
    return component;
}

int spUpdateTabBoxToAdd(spComponent component, int index, spBool *visible) {
    if (spIsTabBox(component) == SP_FALSE || index > SpTabBoxPart(component).num_item) {
        return -1;
    }

    if (index < 0) {
	index = SpTabBoxPart(component).num_item;
    }
    SpTabBoxPart(component).num_item++;

    if (visible != NULL) {
        if (SpTabBoxPart(component).selected_index != index) {
            *visible = SP_FALSE;
        } else {
            *visible = SP_TRUE;
        }
    }
    
    return index;
}

spComponent spAddTabItem(spComponent parent, const char *name, int index, ...)
{
    int num_arg = 0;
    spArg args[SP_MAX_NUM_ARG];
    va_list argp;
    spBool visible = SP_FALSE;
    
    if (spIsCreated(parent) == SP_FALSE)
	return NULL;

    va_start(argp, index);
    spGetArgs(argp, args, num_arg);
    va_end(argp);

    if ((index = spUpdateTabBoxToAdd(parent, index, &visible)) < 0) {
        return NULL;
    }

    spDebug(50, "spAddTabItem", "index = %d, selected_index = %d, num_item = %d\n",
	    index, SpTabBoxPart(parent).selected_index, SpTabBoxPart(parent).num_item);
    
    spSetArg(args[num_arg], SppVisible, visible); num_arg++;
    spSetArg(args[num_arg], SppTabItemIndex, index); num_arg++;

    return spCreateTabItemArg(parent, name, args, num_arg);
}

spComponent spGetAncestorTabBox(spComponent component)
{
    spComponent parent;

    if (component == NULL) return NULL;
    
    parent = component;

    while (parent != NULL) {
	if (spIsTabBox(parent) == SP_TRUE) {
	    return parent;
	}
	parent = SpGetParent(parent);
    }

    return NULL;
}

spBool spIsTabBoxChild(spComponent component)
{
    if (spGetAncestorTabBox(component) != NULL) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spComponent spGetAncestorTabItem(spComponent component)
{
    spComponent parent;

    if (component == NULL) return NULL;
    
    parent = component;

    while (parent != NULL) {
	if (spIsTabItem(parent) == SP_TRUE) {
	    return parent;
	}
	parent = SpGetParent(parent);
    }

    return NULL;
}

