/*
 *	spToolBar_Gtk.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spGraphicsP.h>
#include <sp/spToolBarP.h>

extern spTopLevel sp_toplevel;

void spToolBarPartInitArch(spComponent component)
{
#if GTK_CHECK_VERSION(3,0,0)
    SpToolBarArch(component).pixbuf = NULL;
#else
    SpToolBarArch(component).bitmap = NULL;
#endif
    return;
}

void spToolBarPartFreeArch(spComponent component)
{
    return;
}

#if !GTK_CHECK_VERSION(3,0,0)
static void toolBarRealizeCB(GtkWidget *widget, gpointer data)
{
    spComponent component = (spComponent)data;
    
    if (SpToolBarPart(component).bitmap_data != NULL) {
	GdkColor color;

	spGetColorValueGtk("#c0c0c0", &color);
	
	spDebug(50, "toolBarRealizeCB", "create pixmap\n");
    
	SpPrimitiveArch(component).pixmap = 
	    gdk_pixmap_create_from_xpm_d(SpPrimitiveArch(component).widget->window,
					 &SpToolBarArch(component).bitmap, &color,
					 SpToolBarPart(component).bitmap_data);
    }
    
    return;
}
#endif

void spToolBarCreateArch(spComponent component)
{
    GtkWidget *parent_widget;
    
#if GTK_CHECK_VERSION(2,0,0)
    SpPrimitiveArch(component).widget =	gtk_toolbar_new();
    gtk_toolbar_set_style(GTK_TOOLBAR(SpPrimitiveArch(component).widget), GTK_TOOLBAR_ICONS);
#else
    SpPrimitiveArch(component).widget =
	gtk_toolbar_new(GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_ICONS);
    gtk_container_border_width(GTK_CONTAINER(SpPrimitiveArch(component).widget) ,0);
    gtk_toolbar_set_space_size(GTK_TOOLBAR(SpPrimitiveArch(component).widget), 0);
#endif

#if GTK_CHECK_VERSION(3,0,0)
    gtk_toolbar_set_icon_size(GTK_TOOLBAR(SpPrimitiveArch(component).widget), GTK_ICON_SIZE_MENU);
    parent_widget = SpParentPrimitiveArch(component).top_widget;
    gtk_box_pack_start(GTK_BOX(parent_widget), SpPrimitiveArch(component).widget, FALSE, FALSE, 0);
    
    if (SpToolBarPart(component).bitmap_data != NULL) {
	SpToolBarArch(component).pixbuf = 
	    gdk_pixbuf_new_from_xpm_data((const char **)SpToolBarPart(component).bitmap_data);
	if (SpToolBarArch(component).pixbuf == NULL) {
	    spDebug(50, "toolBarRealizeCB", "can't load xpm data\n");
	}
    }
#else
    parent_widget = SpParentPrimitiveArch(component).widget;
    gtk_container_add(GTK_CONTAINER(parent_widget), SpPrimitiveArch(component).widget);
    
    spSignalConnectCallbackGtk(SpPrimitiveArch(component).widget,
			       "realize", toolBarRealizeCB,
			       (gpointer)component);
#endif
    
    if (spIsVisible(component) == SP_TRUE) {
	gtk_widget_show(SpPrimitiveArch(component).widget);
    }
    
    spDebug(30, "spToolBarCreateArch", "create done\n");

    return;
}

void spToolBarSetParamsArch(spComponent component)
{
    return;
}
    
void spToolBarDestroyArch(spComponent component)
{
#if GTK_CHECK_VERSION(3,0,0)
    if (SpToolBarArch(component).pixbuf != NULL) {
	g_object_unref(SpToolBarArch(component).pixbuf);
    }
#else
    if (SpToolBarArch(component).bitmap != NULL) {
	gdk_bitmap_unref(SpToolBarArch(component).bitmap);
    }
#endif
    
    return;
}
